#!/usr/bin/perl -w

use strict;
use DBI;

require 'localutils.pl';

# Finds matching beacon packets in all tables whose names
# end with the suffix given as the command line argumen

###############################
# usage()
###############################

sub usage {
  print "
Usage: $0 [-v] -tnames <num inputs> <input packet tables> <output beacon table name>
        -v       Verbose (print each SQL query as it's issued).
        -create  Only create the output table and then exit

Finds all matching beacon frames that are found in the input tables
and inserts them into the output table. 

Example: \"$0 -tnames 2 raw_foo raw_bar beacons_foo_bar\" will produce
beacons_foo_bar containing beacons common to the two raw tables.
"; }

###############################
# Okay, real work starts here
###############################

my $tableName;  # assigned in the mainline
my $verbose;
my $LOGFILE;
$|=1;

my @srcTables;
my $tnamesCmdLine = 0;
my $CREATEONLY = 0;

while (defined($ARGV[0]) and $ARGV[0] =~ /^-/) {
    if ( $ARGV[0] eq '-v' ) {
	$verbose = 1;
	setVerbose($verbose);
	shift;
    }
    elsif ( $ARGV[0] eq '-r' ) {
	shift;
	setUser("root");
    }
    elsif ( $ARGV[0] eq '-create' ) {
	shift;
	$CREATEONLY = 1;
    }
    elsif ($ARGV[0] eq "-tnames") {
	shift @ARGV;
	$tnamesCmdLine = 1;
	my $numTables = shift @ARGV;
	map {$srcTables[$_] = shift @ARGV} (0..$numTables-1);
	$tableName = shift @ARGV;
    }
    else {
	usage();
	exit 1;
    }
}

if ( $#ARGV < 1 and !$tnamesCmdLine) {
    usage();
    exit 1;
}

my ($query, $sth);
dbconnect();

if (!$tnamesCmdLine) {
    my $tableSuffix = shift @ARGV;
    my $machines = join("|", @ARGV);

    $query = "SHOW TABLES LIKE 'raw_%$tableSuffix'";
    my $sth = doQuery( \$query );

    while ( my $tbl = $sth->fetchrow_array ) {
	push @srcTables,($tbl) if ($tbl =~ m/$machines/);
    }

# create the main table's name
    $tableName = 'beacons_' . join('_',@srcTables);
    $tableName =~ s/_$tableSuffix//g;
    $tableName =~ s/raw_//g;
    $tableName .=  "_$tableSuffix";
}

# open the log file
open LOGFILE, ">$tableName.log" || die "Couldn't open log file $tableName.log";
$LOGFILE = *LOGFILE;
setLogFile($LOGFILE);

mirrorPrint("\nStart: ", now(), "\n input tables = @srcTables\n output table = $tableName\n logging to $tableName.log\n");

# create the table index table

my $indexTableName = $tableName . '_TblIndex';
$query = "DROP TABLE IF EXISTS $indexTableName";
doQuery( \$query );

$query = "CREATE TABLE $indexTableName ( id int UNSIGNED NOT NULL,
                                          tablename char(48),
                                         PRIMARY KEY( id ))";
doQuery( \$query );

$query = "INSERT INTO $indexTableName ( id, tablename ) VALUES ";
foreach (0..$#srcTables) {
    $query .= "($_, '$srcTables[$_]'),";
}
$query =~ s/,$//;
doQuery( \$query );

$query = "DROP TABLE IF EXISTS $tableName";
doQuery( \$query );

#id int UNSIGNED NOT NULL AUTO_INCREMENT,

$query = "CREATE TABLE $tableName ( id INT UNSIGNED AUTO_INCREMENT,
                                    dBeaconTimestamp bigint UNSIGNED NOT NULL,";
map {$query .= "id_$_ int UNSIGNED, ftime_$_ decimal(18,7), mtime_$_ int UNSIGNED, \n"} (0..$#srcTables);
$query .= "                         PRIMARY KEY(id))";
doQuery( \$query );

exit if ($CREATEONLY);


##create exceptions for the duplicate beacon phenomenon
my @except = ("","");
foreach my $tbl (0..$#srcTables) {
    #if we have input from mojave on channel 1
    $except[$tbl] .= " AND value != '00:0e:d7:87:e4:10' " 
	if ($srcTables[$tbl]  =~ m/moj/ and $srcTables[$tbl] =~ m/i0_c1/);

    #poor beacon sources on channel 11
    $except[$tbl] .= " AND value != '00:05:3c:03:65:87' 
                       AND value != '00:05:3c:04:4f:4e' 
                       AND value != '00:06:25:2b:35:f1'
		       AND value != '00:02:2d:85:0c:c3' 
                       AND value != '00:05:3c:03:da:7c' 
                       AND value != '00:20:e0:8b:80:0a' "
	if ($srcTables[$tbl] =~ m/i2_c11/);

    #if we have input from sonoran on channel 11
#    $except[$tbl] .= " AND value != '00:05:3c:03:da:7c' 
#                       AND value != '00:05:3c:03:65:87' " 
#	if ($srcTables[$tbl]  =~ m/son/ and $srcTables[$tbl] =~ m/i2_c11/);

#    $except[$tbl] .= " AND value != '00:20:e0:8b:80:0a' 
#                       AND value != '00:06:25:2b:35:f1' 
#                       AND value != '00:05:3c:04:4f:4e' "
#	if ($srcTables[$tbl]  =~ m/kal/ and $srcTables[$tbl] =~ m/i2_c11/)
}

my $index;
my @tmptables;
for ($index=0; $index<=$#srcTables; $index++) {
    $tmptables[$index] = "tmp$index"."_$$";

    $query = "CREATE TEMPORARY TABLE $tmptables[$index] (dBeaconTimeStamp bigint UNSIGNED NOT NULL,
                cSeqno smallint UNSIGNED, src char(18),
                id_$index int UNSIGNED, ftime_$index decimal(18,7), mtime_$index int UNSIGNED, 
                PRIMARY KEY( dBeaconTimeStamp, cSeqno, src))";
    doQuery(\$query);

    $query = "INSERT INTO $tmptables[$index]
                 SELECT dBeaconTimestamp, cSeqno, macindex.value, packet.id, aFrameTime, bMACTime 
                 FROM   $srcTables[$index] packet, $srcTables[$index]_MACIndex macindex
                 WHERE  cSubtype=8 AND zErr is NULL AND packet.cA2_Src = macindex.id  $except[$index]
                      ";
    doQueryUse( \$query );
    mirrorPrint("done-$index ", now(), "\n");
}

my $dstColnames = join(",", "dBeacontimeStamp", map { "id_$_, ftime_$_, mtime_$_" } (0..$#srcTables));
my $srcColnames = join(",", "$tmptables[0].dBeacontimeStamp",
                            map{ "$tmptables[$_].id_$_, $tmptables[$_].ftime_$_, $tmptables[$_].mtime_$_" } (0..$#srcTables));

$query = "INSERT INTO $tableName ($dstColnames)
             SELECT $srcColnames FROM ".
    (join(" natural join ", (map {$tmptables[$_]} (0..$#srcTables)) )) .
    " ORDER BY id_0";  ## not necessary i think
#    " ORDER BY dBeaconTimestamp";  ## not necessary i think
doQueryUse(\$query);

dbdisconnect();

mirrorPrint("\nTerminated normally(?)\n");
mirrorPrint(now(), "\n");    
