#!/usr/bin/perl

use strict;
use Time::Local;
use Data::Dumper;

use DBI;

require 'localutils.pl';

# Creates tables:
#     indexIndex -- a description of the fields that are stored in indexed fashion
#     typeIndex  -- the type index table
#     subtypeIndex -- the subtype index table
# The last two are different from the other indexed tables in that they are not
# generated on the fly, as we see values in the data stream.  Instead, their
# id's correspond to the ones defined by 802.11.

my $elem;

# list of fields that are inverted.  For each, we have a name, a type, and an indication
# of whether or not it is dynamically created -- entries added as we see new
# values in the input stream -- or statically created (by this program) -- type and
# subtype index tables.

my %dbIndexedFields = (
		_zBSSID => { indexedType => 'char(18)',         # type in the index table
			     indexTableName => '[ThisDataset]_MACIndex',
			     tableType   => 'int UNSIGNED', # type in the primary table
			     dyn         =>  1},
		_zClient => { indexedType => 'char(18)',         # type in the index table
			     indexTableName => '[ThisDataset]_MACIndex',
			     tableType   => 'int UNSIGNED', # type in the primary table
			     dyn         =>  1},
		_zPartner => { indexedType => 'char(18)',         # type in the index table
			     indexTableName => '[ThisDataset]_MACIndex',
			     tableType   => 'int UNSIGNED', # type in the primary table
			     dyn         =>  1},
		_zDir =>   { indexedType => 'char(8)',          # type in the index table
			     indexTableName => 'dirIndex',
			     tableType   => 'tinyint UNSIGNED', # type in the primary table
			     dyn         =>  0},
		_zParseSymbol =>   { indexedType => 'char(48)',          # type in the index table
			     indexTableName => '_zParseSymbolIndex',
			     tableType   => 'tinyint UNSIGNED', # type in the primary table
			     dyn         =>  1},
		zErr =>    { indexedType => 'char(18)',
			     indexTableName => 'rawErrorCodesIndex',
			     tableType   => 'tinyint UNSIGNED',
			     dyn         => 0},
		cA1_Dest =>{ indexedType => 'char(18)',
			     indexTableName => '[ThisDataset]_MACIndex',
			     tableType   => 'int UNSIGNED',
			     dyn         => 1},
		cA2_Src => { indexedType => 'char(18)',
			     indexTableName => '[ThisDataset]_MACIndex',
			     tableType   => 'int UNSIGNED',
			     dyn         => 1},
		cA3 =>     { indexedType => 'char(18)',
			     indexTableName => '[ThisDataset]_MACIndex',
			     tableType   => 'int UNSIGNED',
			     dyn         => 1},
		cA4 =>     { indexedType => 'char(18)',
			     indexTableName => '[ThisDataset]_MACIndex',
			     tableType   => 'int UNSIGNED',
			     dyn         => 1},
		cSubtype =>{ indexedType => 'char(48)',
			     indexTableName => 'subtypeIndex',
			     tableType   => 'tinyint UNSIGNED',
			     dyn         =>  0},
		cType =>   { indexedType => 'char(16)',
			     indexTableName => 'typeIndex',
			     tableType   => 'tinyint UNSIGNED',
			     dyn         =>  0},
    );


###############################
# Okay, real work starts here
###############################


if ( $#ARGV != -1 ) {
    usage();
    exit 1;
}

dbconnect();

# create the indexIndex table

my $query = 'DROP TABLE IF EXISTS indexIndex';
doQuery(\$query);

$query = 'CREATE TABLE indexIndex ( id int UNSIGNED NOT NULL AUTO_INCREMENT,
                                    field           char(16),
                                    dyn             tinyint,
                                    tabletype       char(24), 
                                    indexedtype     char(24),
                                    indextablename  char(24), 
                                    PRIMARY KEY ( id ))';
doQuery(\$query);

$query = 'INSERT INTO indexIndex (field, dyn, tabletype, indexedtype, indextablename) VALUES ';
foreach $elem (sort keys %dbIndexedFields) {
    $query .= "( '$elem'," . 
	         $dbIndexedFields{$elem}->{dyn} . "," . 
		 "'" . $dbIndexedFields{$elem}->{tableType} . "'," .
		 "'" . $dbIndexedFields{$elem}->{indexedType} . "'," . 
		 "'" . $dbIndexedFields{$elem}->{indexTableName} . "'" . 
	      "),";
}
$query =~ s/,$//;
doQuery(\$query);
print "(Re)created indexIndex table\n";

# Create the typeIndex table

my $query = 'DROP TABLE IF EXISTS typeIndex';
doQuery(\$query);

$query = 'CREATE TABLE typeIndex ( id int UNSIGNED NOT NULL,
                                   value ' . $dbIndexedFields{cType}->{indexedType} . ')';
doQuery(\$query);

$query = "INSERT INTO typeIndex (id, value) VALUES
               (0, 'Management'),
               (1, 'Control'),
               (2, 'Data'),
               (3, 'Reserved')";
doQuery(\$query);
print "(Re)created typeIndex table\n";

# Create the subtypeIndex table

my $query = 'DROP TABLE IF EXISTS subtypeIndex';
doQuery(\$query);

$query = 'CREATE TABLE subtypeIndex ( id int UNSIGNED NOT NULL,
                                      value ' . $dbIndexedFields{cSubtype}->{indexedType} . ',
                                      length tinyint UNSIGNED,
                                      lengthOp char(2))';
doQuery(\$query);

# codes from http://standards.ieee.org/getieee802/download/802.11-1999.pdf

$query = "INSERT INTO subtypeIndex (id, value, length, lengthOp) VALUES
           (0x00, 'Association request', 28, '>='),
           (0x01, 'Association response', 28, '>='),
           (0x02, 'Reassociation request', 28, '>='),
           (0x03, 'Reassociation response', 28, '>='),
           (0x04, 'Probe request', 28, '>='),
           (0x05, 'Probe response', 28, '>='),
           (0x06, 'Reserved', null, null),
           (0x07, 'Reserved', null, null),
           (0x08, 'Beacon', 28, '>='),
           (0x09, 'Announcement traffic indication message (ATIM)', 28, '>='),
           (0x0A, 'Disassociation', 28, '>='),
           (0x0B, 'Authentication', 28, '>='),
           (0x0C, 'Deauthentication', 28, '>='),
           (0x0D, 'Reserved', null, null),
           (0x0E, 'Reserved', null, null),
           (0x0F, 'Reserved', null, null),
           (0x10, 'Reserved', null, null),
           (0x11, 'Reserved', null, null),
           (0x12, 'Reserved', null, null),
           (0x13, 'Reserved', null, null),
           (0x14, 'Reserved', null, null),
           (0x15, 'Reserved', null, null),
           (0x16, 'Reserved', null, null),
           (0x17, 'Reserved', null, null),
           (0x18, 'Reserved', null, null),
           (0x19, 'Reserved', null, null),
           (0x1A, 'Power Save (PS)-Poll', 20, '=='),
           (0x1B, 'Request To Send (RTS)', 20, '=='),
           (0x1C, 'Clear To Send (CTS)', 14, '=='),
           (0x1D, 'Acknowledgment (ACK)', 14, '=='),
           (0x1E, 'Contention-Free (CF)-End', 20, '=='),
           (0x1F, 'CF-End + CF-Ack', 20, '=='),
           (0x20, 'Data', 28, '>='),
           (0x21, 'Data + CF-Ack', 28, '>='),
           (0x22, 'Data + CF-Poll', 28, '>='),
           (0x23, 'Data + CF-Ack + CF-Poll', 28, '>='),
           (0x24, 'Null function (no data)', 28, '>='),
           (0x25, 'CF-Ack (no data)', 28, '>='),
           (0x26, 'CF-Poll (no data)', 28, '>='),
           (0x27, 'CF-Ack + CF-Poll (no data)', 28, '>='),
           (0x28, 'Reserved', null, null),
           (0x29, 'Reserved', null, null),
           (0x2A, 'Reserved', null, null),
           (0x2B, 'Reserved', null, null),
           (0x2C, 'Reserved', null, null),
           (0x2D, 'Reserved', null, null),
           (0x2E, 'Reserved', null, null),
           (0x2F, 'Reserved', null, null),
           (0x30, 'Reserved', null, null),
           (0x31, 'Reserved', null, null),
           (0x32, 'Reserved', null, null),
           (0x33, 'Reserved', null, null),
           (0x34, 'Reserved', null, null),
           (0x35, 'Reserved', null, null),
           (0x36, 'Reserved', null, null),
           (0x37, 'Reserved', null, null),
           (0x38, 'Reserved', null, null),
           (0x39, 'Reserved', null, null),
           (0x3A, 'Reserved', null, null),
           (0x3B, 'Reserved', null, null),
           (0x3C, 'Reserved', null, null),
           (0x3D, 'Reserved', null, null),
           (0x3E, 'Reserved', null, null),
           (0x3F, 'Reserved', null, null)";
doQuery(\$query);
print "(Re)created subtypeIndex table\n";

# the direction index table
my $query = 'DROP TABLE IF EXISTS dirIndex';
doQuery(\$query);

$query = 'CREATE TABLE dirIndex ( id int UNSIGNED NOT NULL,
                                      value ' . $dbIndexedFields{_zDir}->{indexedType} . ')';
doQuery(\$query);

$query = "INSERT INTO dirIndex (id, value) VALUES (0,'up'),(1,'down'),(2,'ad hoc'),(3,'ap-ap')";
doQuery(\$query);

print "(Re)created dirIndex table\n";

dbdisconnect();

