#!/usr/bin/perl

use strict;
use Time::Local;
use Data::Dumper;

require 'localutils.pl';
$|=1;  ##do not buffer output

# Script to process the sigcomm trace data and stuff it in a db.
# Parses each packet and presents it to logPacket, allowing some action
#  to be taken.

my $TETHEREAL = "$ENV{HOME}/usr/local/bin/tethereal";
my $tableName;  # assigned in the mainline

my $logger;         # Machine doing the logging
my $interface;      # interface being logged
my $currentChannel; # initial (anticipated) channel

my $dbh;
my $sth;
my $rv;

my $verbose;
my $noPrompt = 0;

my $LOGFILE;

# THESE ARE NOT ALL THE DB FIELDS!
# Indexed fields will be taken from the indexIndex table and inserted
#    with type regsitered in that table.
# %frameControlFields will also be added, with type tinyint UNSIGNED.
# id field will be added once all other fields have been added.

my %dbFields = (
# fields from Frame section of tethereal output
    aFrameTime => 'decimal(18,7)',
    aTimeDelta => 'int UNSIGNED',
# fields from Prism Header section of tethereal output
    bChannel  => 'tinyint UNSIGNED',
    bHostTime => 'int UNSIGNED',
    bMACTime  => 'int UNSIGNED',
    bRate     => 'float',
    bRssi     => 'smallint',
    bSize     => 'smallint UNSIGNED',
# fields from 802.11 section of tethereal output
    # frame control fields are managed using %frameControlFields
    # address fields are managed using @addrName
    cDur      => 'smallint UNSIGNED',
    cFragno   => 'tinyint UNSIGNED',
    cSeqno    => 'smallint UNSIGNED',
# fields from any subsequent sections of tethereal output
    dBeaconTimestamp   => 'bigint UNSIGNED',
# synthetic fields
    zErr      => 'tinyint UNSIGNED',
    zErrDriver=> 'tinyint UNSIGNED',
    zErrPhy   => 'tinyint UNSIGNED',
    );
my @dbFieldsKeys;  # the keys, extracted for speed(?)

# elements are  (field name, mask, shift amount)
my @frameControlFields = (
			  ['cProtver',  0x00000003,  0],
			  ['cType',     0x0000000c,  2],
			  ['cSubtype',  0x000000f0,  4],
			  ['cDSTo',     0x00000100,  8],
			  ['cDSFrom',   0x00000200,  9],
			  ['cMorefrag', 0x00000400, 10],
			  ['cRetry',    0x00000800, 11],
			  ['cPwrmgt',   0x00001000, 12],
			  ['cMoredata', 0x00002000, 13],
			  ['cWEP',      0x00004000, 14],
			  ['cOrdered',  0x00008000, 15],
		       );

# Address fields 1-4 are given in order in tethereal output
my $addrIndex;
my @addrName = ("cA1_Dest", "cA2_Src", "cA3", "cA4");
my %macIndex = ();
my $MAX_MACINDEX_VALUE = 65535;
my $macindex_overflow_detected = 0;

my %errorCode = (
#    none => ,
    malformedFrame => 1,
    malformedPrism => 2,
    malformed802 => 3,
    malformedOther => 4,
    driverReported => 5,
    unexpectedSection => 6,
    badLength => 7,
    badTypeSubtype => 8,
  );

my $frameControl;
my $fcEntry;

my @subtypeInfo;  # indexed 1-63.  Each element is an array ref -> [length, lengthOp]
my $SUBTYPE_LENGTH_FIELD = 0;
my $SUBTYPE_OP_FIELD = 1;
my $subtype;

my $data;
my $pktdata;
my $expected;
my $gotErr;

my %tableData = (
		 sqlInsertRoot => "",
		 querystring => "",
		 nextID => 1,
	        );
my $beaconType;
my $SQLINSERTPERIOD = 40;
my $totalErrorPackets = 0;  # we keep track only of the number of error packets,
                            # since all fields we thought we found with tethereal
                            # are bogus

# Prevent corrupted RSSI values.  (Actual max is 60.)
my $RSSI_THRESHOLD = 100;

my $VERBOSE = 1;

# If this is defined, only this many packets are processed (for debugging)
my $PACKET_LIMIT;
#my $PACKET_LIMIT = 100;

my $totalPkts;
my $currDate = "start";
my $startClockTime = "";

####################################################################
# myHex -- perl's hex() assumes
#   numbers are positive (unsigned)
####################################################################

sub myHex {
    return hex($_[0]) - 0x100000000 if hex($_[0])  > 0x7fffffff;
    return hex($_[0]);
}

###parseMDYTime

my %localMonthsHash = (
  Jan => 0,
  Feb => 1,
  Mar => 2,
  Apr => 3,
  May => 4,
  Jun => 5,
  Jul => 6,
  Aug => 7,
  Sep => 8,
  Oct => 9,
  Nov => 10,
  Dec => 11
);

sub parseMDYTime {
  my ($dateStamp) = @_;
  if ( $dateStamp =~ m/([A-Za-z]+) +(\d+), (\d+) (\d+):(\d+):([\d\.]+)/ ) {
    my ($month, $day, $year, $hour, $min, $sec) = ($1, $2, $3, $4, $5, $6);
    $year = $year - 1900;
    my $m = $localMonthsHash{$month};
    ## timelocal doesn't convert fractional seconds
    my $epochSecs = $sec + timelocal(0, $min, $hour, $day, $m, $year);
    return $epochSecs;
  }
  else {
    die("Couldn't decode date: $dateStamp\n");
  }
}

####################################################################
# processPackets()
####################################################################

sub processPackets {
  my ($file) = @_;

  my %pkt;

  my $regex_hex = '0x[0-9a-f]+';
  my $regex_number = '-?\d+(?:\.\d+)?';
  #my $regex_mac = '[0-9a-f]{2}:[0-9a-f]{2}:[0-9a-f]{2}:[0-9a-f]{2}:[0-9a-f]{2}:[0-9a-f]{2}';
  my $regex_mac = '(..:){5}..';

  # Read packets from tethereal verbose output
  die "$TETHEREAL not found\n" if (!-x $TETHEREAL);
  my $command = "$TETHEREAL -V -r $file -o wlan.check_fcs:true";
  mirrorPrint("\n", now(), "\nProcessing command: $command\n");
  open(STREAM, "$command |") or die("Couldn't open packet stream");

  while ( 1 ) {

    # Check if we've reached the limit (for debugging)
    #if (defined $PACKET_LIMIT && $totalPkts >= $PACKET_LIMIT) {
    #  print "[Reached debugging packet limit of $PACKET_LIMIT]\n";
    #  return;
    #}

    # We expect to see three sections:
    #    Frame ...
    #    Prism Monitoring Header...
    #    IEEE 802.11
    # followed by sections we don't care about.
    # The frame info ends with a blank line.

    # discard the rest of the input up to the next Frame... line
    # (which indicates the start of a new frame).
    # The lines are probably from the tail of the last packet already
    # processed

    %pkt = ();
    $gotErr = 0;
    $addrIndex = 0;

    # wait until we recognize the beginning of a new Frame
    while ( <STREAM> ) {
	if (m/^Frame (\d+) /o) {
	    $pkt{num} = $1;
	    last;
	}
    }
    # quit if done
    last if ! defined $_;

    # I'm using a loop here, but only to avoid try...catch, which I'm worried will have
    # a runtime overhead.  (I'm using 'last' as 'throw'.)
    # This "loop" executes exactly once.

    do {

	#------------ Frame header ---------------------------------------------------------------------------------
	# Expected fields:
	#    Arrival Time: Aug 30, 2004 08:03:54.782480000
	#    Time delta from previous packet: 0.000000000 seconds
	#    Time since reference or first frame: 0.000000000 seconds
	#    Frame Number: 1
	#    Packet Length: 288 bytes
	#    Capture Length: 288 bytes
	#    Protocols in frame: prism:wlan

	$_ = <STREAM>;
	if ( s/^\s+Arrival Time:\s+// ) {$pkt{aFrameTime} = sprintf("%.7f", &parseMDYTime($_)); }
#	  if ( s/^\s+Arrival Time:\s+// ) {$pkt{aFrameTime} = &parseMDYTime($_); }
	  else { $expected = "Arrival Time:"; $gotErr = 1; last; }
	$_ = <STREAM>;
	  # convert time delta from fractional seconds to integer # usecs.
	  if ( m/^\s+Time delta from previous packet: ([^ ]+) /o ) { $pkt{aTimeDelta} = int($1 * 1000000) if $pkt{num}>1; }
	  else { $expected = "Time delta..."; $gotErr = 1; last; }
	$_ = <STREAM>;  if ( !m/^\s/o ) { $expected = "Time since..."; $gotErr = 1; last; }
	$_ = <STREAM>;	if ( !m/^\s/o ) { $expected = "Frame Number..."; $gotErr = 1; last; }
	$_ = <STREAM>;	if ( !m/^\s/o ) { $expected = "Packet Length..."; $gotErr = 1; last; }
	$_ = <STREAM>;	if ( !m/^\s/o ) { $expected = "Capture Length..."; $gotErr = 1; last; }
	$_ = <STREAM>;	if ( !m/^\s/o ) { $expected = "Protocols in frame:"; $gotErr = 1; last; }
    } while ( 0 );

    if ( $gotErr > 0 ) {
	#mirrorPrint("Packet $pkt{num}: got '$_' but expected '$expected'\n");
	$pkt{zErr} = $errorCode{malformedFrame};
    }

    # another loop, for its throw...catch qualities

    do {
	#------------ Prism header ---------------------------------------------------------------------------------
	# Expected fields:
	#Prism Monitoring Header
	#    Message Code: 68
	#    Message Length: 144
        #    Device: ath0
	#    Host Time: 0x31f7d (DID 0x10044, Status 0x0, Length 0x4)
        #    MAC Time: 0x1b6a1 (DID 0x20044, Status 0x0, Length 0x4)
        #    Channel: 0x1 (DID 0x30044, Status 0x0, Length 0x4)
        #    RSSI: 0x0 (DID 0x40044, Status 0x0, Length 0x4)
        #    SQ: 0x0 (DID 0x0, Status 0x0, Length 0x0)
        #    Signal: 0x2 (DID 0x60044, Status 0x0, Length 0x4)
        #    Noise: 0x2 (DID 0x70044, Status 0x0, Length 0x4)
	#    Data Rate: 1.0 Mb/s
        #    IsTX: 0x0 (DID 0x90044, Status 0x0, Length 0x4)
	#    Frame Length: 0x90 (DID 0xa0044, Status 0x0, Length 0x4)

	$_ = <STREAM>;	if ( !m/^Prism Mon/o )  { $expected = "Prism Monitoring Header"; $gotErr = 1; last; }
	$_ = <STREAM>;	if ( !m/^\s/o ) { $expected = "Message Code..."; $gotErr = 1; last; }
	$_ = <STREAM>;	if ( !m/^\s/o ) { $expected = "Message Length"; $gotErr = 1; last; }
	$_ = <STREAM>;	if ( !m/^\s/o ) { $expected = "Device:..."; $gotErr = 1; last; }
	$_ = <STREAM>;	
	  if ( m/^\s+Host Time: (0x[^ ]+)/o ) { $pkt{bHostTime} = hex($1); }
	  else { $expected = "Host Time:..."; $gotErr = 1; last; }
	$_ = <STREAM>;
	  if ( m/^\s+MAC Time: (0x[^ ]+)/o ) { $pkt{bMACTime} = hex($1); }
	  else { $expected = "Mac Time:..."; $gotErr = 1; last; }
	$_ = <STREAM>;
	  if ( m/^\s+Channel: (0x[^ ]+)/o ) { $pkt{bChannel} = hex($1); }
	  else { $expected = "Channel:..."; $gotErr = 1; last; }
	$_ = <STREAM>;
	  if ( m/^\s+RSSI: (0x[^ ]+)/o ) { $pkt{zErrPhy} = hex($1); }
	  else { $expected = "RSSI:..."; $gotErr = 1; last; }
	$_ = <STREAM>;	if ( !m/^\s/o ) { $expected = "SQ:..."; $gotErr = 1; last; }
	$_ = <STREAM>;
	  if ( m/^\s+Signal: (0x[^ ]+)/o ) { $pkt{bRssi} = myHex($1); }
	  else { $expected = "Signal:..."; $gotErr = 1; last; }
	$_ = <STREAM>;
	  if ( m/^\s+Noise: 0x([^ ]+)/o ) { $data = myHex($1); $pkt{zErrDriver} = $data if $data > 0; }
	  else { $expected = "Noise:..."; $gotErr = 1; last; }
	$_ = <STREAM>;
	  if (m/^\s+Data Rate: ([^ ]+) Mb/o ) { $pkt{bRate} = $1; }
	  else { $expected = "Data Rate:..."; $gotErr = 1; last; }
	$_ = <STREAM>;	if ( !m/^\s/o ) { $expected = "IsTX:..."; $gotErr = 1; last; }
	$_ = <STREAM>;
	  if ( m/^\s+Frame Length: (0x[^ ]+)/o ) { $pkt{bSize} = hex($1); }
	  else { $expected = "Frame Length:..."; $gotErr = 1; last; }

    } while ( 0 );  # we want "last" statements from loop to come here...

    if ( $gotErr > 0 && $pkt{zErr} == $errorCode{none} ) {
	#mirrorPrint("Packet $pkt{num}: got '$_' but expected '$expected'\n");
	$pkt{zErr} = $errorCode{malformedPrism};
    }


    #------------ 802.11 header ---------------------------------------------------------------------------------

    if ( $pkt{zErr}==$errorCode{none} ) {
	$_ = <STREAM>;
	if ( m/^IEEE 802.11$/o ) {
	    while (<STREAM>) {
		last if m/^[^ ]/o or m/^\s*$/o;  # quit if we see an unindented or blank line

		s/^\s+//o;  # get rid of leading whitespace, to speed up following comparisons(?)
		if (m/^Frame Control: 0x([^ ]+)/o ) {
		    $frameControl = hex($1);
		    foreach $fcEntry (@frameControlFields) {
			$pkt{$fcEntry->[0]} = ($frameControl & $fcEntry->[1]) >> $fcEntry->[2];
		    }
		    # adjust cSubtype field to match id's in subtypeIndex table
		    $pkt{cSubtype} += 16*$pkt{cType} if (defined $pkt{cSubtype}) && (defined $pkt{cType});
		}
		elsif (m/^Duration: (\d+)/o ) {
		    $pkt{cDur} = $1;
		}
		elsif (m/^BSS Id:.*($regex_mac)/o ) {
		    $pkt{$addrName[$addrIndex++]} = $1;
		}
		elsif (m/^(Source|Transmitter|Destination|Receiver) address:.*($regex_mac)/o ) {
		    $pkt{$addrName[$addrIndex++]} = $2;
		}
		elsif (m/^Fragment number: (\d+)/o ) {
		    $pkt{cFragno} = $1;
		}
		elsif (m/^Sequence number: (\d+)/o ) {
		    $pkt{cSeqno} = $1;
		}
	    }
	    if (m/^\[Malformed Packet/o ) {
		# Note: it seems this message can sometimes be spurious, so
		# we look for it only up to here.
		$pkt{zErr} = $errorCode{malformed802};
	    }
	}
	else {
	    # something wrong --
	    #mirrorPrint "Frame $totalPkts/", $pkt{num}, ": Expected 'IEEE 802.11...' but got '$_'\n";
	    $pkt{zErr} = $errorCode{unexpectedSection};
	}
    }

    #------------ Other header ---------------------------------------------------------------------------------

    # At this point the only interesting data still to come is for beacon frames --
    # we save their timestamps.

    if ( $pkt{zErr}==$errorCode{none} && $pkt{cSubtype}==8 # 'Beacon'
         && m/^IEEE 802.11 wireless LAN management frame$/o ) {
	# we want the timestamp field
	while (<STREAM>) {
	    last if m/^[^ ]/o or m/^\s*$/o;  # quit if we see an unindented or blank line
	    if ( m/^\s+Timestamp: (0x[^\s]+)/ ) {
		$pkt{dBeaconTimestamp} = $1;
		last;
	    }
	    if (m/^\[Malformed Packet/o ) {
	      $pkt{zErr} = $errorCode{malformedOther};
	    }
	}
    }

    #------------ look for mailformed  ----------------------------------------------------------------------

    if ( ! m/^\s*$/o ) {
      while ( <STREAM> ) {
	if (m/^\[Malformed/o) {
	  $pkt{zErr} = $errorCode{malformedOther};
	  last;
	}
	last if (m/^\s*$/o);
      }
    }

    #------------ finish up ---------------------------------------------------------------------------------
    # check for, and indicate, driver-reported error (passed up through Noise and RSSI fields)

    $pkt{zErr} = $errorCode{driverReported} if defined $pkt{zErrDriver};
    undef $pkt{zErrPhy} unless ($pkt{zErrDriver} & 8) != 0;

    # now check the packet for "semantic errors" (impossible field values not flagged by tethereal)

    if ( $pkt{zErr}==$errorCode{none} ) {
	$subtype = $pkt{cSubtype};
	if ( $subtype>63 || !defined $subtypeInfo[$subtype][0] ) {
	    $pkt{zErr} = $errorCode{badTypeSubtype};
	}
	elsif ( $subtypeInfo[$subtype][$SUBTYPE_OP_FIELD] eq '==' ) {
	    $pkt{zErr} = $errorCode{badLength} if $pkt{bSize} != $subtypeInfo[$subtype][$SUBTYPE_LENGTH_FIELD];
	}
	elsif ( $subtypeInfo[$subtype][$SUBTYPE_OP_FIELD] eq '>=' ) {
	    $pkt{zErr} = $errorCode{badLength} if $pkt{bSize} < $subtypeInfo[$subtype][$SUBTYPE_LENGTH_FIELD];
	}
    }

    # we've processed everything we're interested in.

    $totalPkts++;
    mirrorPrint(commify($totalPkts), " packets\n") if $totalPkts % 100000 == 0;
    &logPacket( \%pkt );

    last if ! defined $_;
  }
}

##################################
# dumpPktRef()
##################################

sub dumpPktRef {
    my $pkt = @_[0];

    mirrorPrint("Packet dump:\n");
    foreach my $field (sort keys %{$pkt}) {
	mirrorPrint("\t$field = ", $$pkt{$field}, "\n");
    }
}

##################################
# logPacket()
#   Dumps packet data to db
##################################

sub logPacket {
  my ($pkt) = @_;
  
  # First check to see if we need to switch database tables.
  # (Table name is [monitor host]_i[interface #]_c[channel #],
  # so if channel changes, table changes.)

  if ( $pkt->{zErr} == $errorCode{none} &&
       $pkt->{bChannel} != $currentChannel ) {
	   mirrorPrint(now(), "\nSwitching channels: $currentChannel ->", $pkt->{bChannel}, "\n");
	   $currentChannel = $pkt->{bChannel};
	   setTableName();
	   createTable();
  }

  # Convert indexed fields to their indexed values.
  # We should actually look to see what $pkt{} values will
  # actually be inserted -- error packets don't have any, for instance --
  # and write only those.  This implementation is a hack -- either 
  # a packet has them all, or none.

  foreach my $elem ( @addrName ) {
      $pktdata = $pkt->{$elem};
      # never index NULL
      next if !defined($pktdata);
   
      # try to grab index out of local hash
      $data = $macIndex{$pktdata};
      # we don't enter new macs into index if the packet is corrupted
      if ( (! defined $data) && ($pkt->{zErr}==$errorCode{none}) ) {
	  # need to create index in db
	  my $value = $pktdata ? $pktdata : 'NULL';
	  doStrQuery( 'INSERT INTO ' . $tableName . "_MACIndex (value) VALUES ('$value')" );

	  # retrieve the index
	  doStrQuery( "SELECT LAST_INSERT_ID()" );
	  $data = $sth->fetchrow_array || die "Couldn't read last id from $elem" . "Index";

	  $macIndex{$pktdata} = $data;

	  if ( $data > $MAX_MACINDEX_VALUE && $macindex_overflow_detected == 0 ) {
	      $macindex_overflow_detected = 1;
	      mirrorPrint("\n********* ERROR: MAC index overflowed: packet $pkt->{num}\n\n");
	  }
      }
      $pkt->{$elem} = $data;
  }

  # assign a UID to the packet
  $pkt->{id} = $tableData{nextID}++;

  # okay, at last, queue the data for insertion

  $tableData{querystring} .= '(';
  foreach my $elem ( @dbFieldsKeys ) {
      $pkt->{$elem} = "NULL" if ! defined $pkt->{$elem};
      $tableData{querystring} .= "$pkt->{$elem},";
  }
  $tableData{querystring} =~ s/,$//;
  $tableData{querystring} .= "),";
  $tableData{count}++;  # increment buffering count

  flushInserts( $SQLINSERTPERIOD );
}

#####################################
# resetQueryString()
#####################################

sub resetQueryString {
    $tableData{querystring} = "INSERT INTO " . $tableName . $tableData{sqlInsertRoot};
}


#####################################
# flushInserts()
#   since we buffer inserts into
#   primary tables, we have to
#   make sure to flush the buffers
#   before terminating
#####################################

sub flushInserts {
    my $limit = @_[0];
    if ( $tableData{count} > $limit ) {
	$tableData{querystring} =~ s/,$//;
	# now execute the query
	doStrQuery( $tableData{querystring} );
	$tableData{count} = 0;
	resetQueryString();
    }
}

################################################################
# initFieldData()
#   Called once, from the mainline
################################################################

sub initFieldData {

    # insert additional fields into master hash list: frame control and address
    foreach my $elem (@frameControlFields) {
	$dbFields{$elem->[0]} = 'tinyint UNSIGNED';
    }
    foreach my $elem ( @addrName ) {
	$dbFields{$elem} = 'smallint UNSIGNED';
    }

    # set the dbFields supporting data
    @dbFieldsKeys = sort keys %dbFields;
    # add in the id field only now, because we want it first (out of alphabetical order)
    $dbFields{id} = 'int NOT NULL UNIQUE';
    @dbFieldsKeys = ('id', @dbFieldsKeys);
    
}

################################################################
# dropExistingTables()
#   Drops tables that this run MIGHT write into.  Done once,
#   at the beginning of the run, so that if we encounter a
#   single interface switching channels, and then switching
#   back, we will append to the table created at the beggining
#   of the run.
#
# Relies on index table names starting with table name!!!!
#
################################################################

sub dropExistingTables {
    my $tableroot = $tableName;
    $tableroot =~ s/_c[0-9]+$/_c/ ||
	die "Expected _c.+ at end of table name: $tableName";
    doStrQuery( "SHOW TABLES LIKE '$tableroot%'" );
    my @dropTables;
    while ( my $tbl = $sth->fetchrow_array ) {
	push @dropTables,($tbl);
    }

    if ( ! $noPrompt ) {
	mirrorPrint("The following tables will be deleted:\n\t",
	      join( "\n\t", @dropTables),
	      "\nContinue (y or n)? ");
	my $resp = <>;
	exit 2 if not $resp =~ /^(y|Y)/;
    }
    if ( $#dropTables >= 0 ) {
	doStrQuery( "DROP TABLE " . join(',', @dropTables ) );
	mirrorPrint("Dropped tables ", join(',', @dropTables ), "\n");
    }
}


################################################################
# createTable
#   Creates the main table that represent the
#   packet data, if it doesn't already exist.
#   Ditto for the index table, which holds MAC addresses.
################################################################

sub createTable {

    # now deal with the primary table
    my $createTable = "CREATE TABLE IF NOT EXISTS $tableName ( ";
    foreach my $elem ( @dbFieldsKeys ) {
	$createTable .= $elem . " " . $dbFields{$elem} . ", ";
    }
    $createTable .= "PRIMARY KEY ( id ))";
    doStrQuery( $createTable );
    mirrorPrint("(Re-)created trace table $tableName\n");

    # create the dynamically generated MAC index table

    $createTable = "CREATE TABLE IF NOT EXISTS $tableName" . "_MACIndex
                           ( id int UNSIGNED NOT NULL AUTO_INCREMENT, 
                             value char(18),
                             PRIMARY KEY ( id ))";
    doStrQuery( $createTable );

    # reset id counter
    $tableData{nextID} = 1;

    mirrorPrint("(Re-)created index table $tableName", "_MACIndex\n");
}

###############################
# storeErrorCodes()
###############################

sub storeErrorCodes {
    doStrQuery( "DROP TABLE IF EXISTS rawErrorCodesIndex");
    doStrQuery( "CREATE TABLE IF NOT EXISTS rawErrorCodesIndex (id int UNSIGNED NOT NULL, value char(18), PRIMARY KEY( id ))" );

    my @orderedKeys = sort { $errorCode{$a} <=> $errorCode{$b} } keys %errorCode;
    foreach my $key (@orderedKeys) {
	doStrQuery( "INSERT INTO rawErrorCodesIndex (id, value) VALUES ($errorCode{$key}, '$key')" );
    }
}

###############################
# fetchSubtypeInfo()
###############################

sub fetchSubtypeInfo {
    my @row;
    doStrQuery("SELECT id,length,lengthOp FROM subtypeIndex");
    while ( @row = $sth->fetchrow_array ) {
	$subtypeInfo[$row[0]] = [$row[1], $row[2]];
    }
}

###############################
# doStrQuery()
###############################

sub doStrQuery {
    my $q = @_[0];
    $sth = doQuery(\$q);
}

###############################
# setTableName()
###############################

sub setTableName {
    $tableName = "raw_$logger" . "_i$interface" . "_c" . $currentChannel;
}

###############################
# usage()
###############################

sub usage {
  print "
Usage: ./createRawDatabase.pl [-b] [-d dir] [-v] [logging machine] [interface number] [channel]
        -b       Background mode.  Do NOT require confirmation before
                 deleting existing tables.
        -d  dir  Specify the parent directory of the data files.
                 [Default: .]
        -v       Verbose (print each SQL query as it's issued).

The script will read all data files generated by the logging machine named
for the interface named.  The output table is named raw_[logger]_i[int#]_c[chn#].
The channel number you give is the initial channel number -- if the interface 
switches channels, data is directed to a new (appropriately named) db table.

Example: ./createRawDatabase.pl -b sahara 0 11
";
}

###############################
# Okay, real work starts here
###############################

my $rootDir = ".";
while ($ARGV[0] =~ /^-/) {
    if ($ARGV[0] eq '-b') {
	$noPrompt = 1 ;
    }
    elsif ( $ARGV[0] eq '-d' ) {
	$rootDir = $ARGV[1];
	shift;
    }
    elsif ( $ARGV[0] eq '-v' ) {
	$verbose = 1;
	setVerbose(1);
    }
    else {
	usage();
	exit 1;
    }
    shift;
}

if ( $#ARGV != 2 ) {
    usage();
    exit 1;
}

$logger = shift;         # Machine doing the logging
$interface = shift;      # interface being logged
$currentChannel = shift; # initial (anticipated) channel

## temporary hack!
#die "skipping sonoran\n" if ($logger =~ m/sonoran/);


setTableName();

# open the log file
open LOGFILE, ">$tableName.log" || die "Couldn't open log file $tableName.log";
$LOGFILE = *LOGFILE;
setLogFile($LOGFILE);

mirrorPrint("\nStart: ", now(), "\n");

$dbh = dbconnect();

# save the error codes being used to a db table
storeErrorCodes();

# cache the type/subtype info (from db table)
fetchSubtypeInfo();

# finish the setup of the field data
initFieldData();

# drop the main table and index tables (relies on index table names starting with table name!)
dropExistingTables();

# create the first (possibly only) table and index table
createTable();

$tableData{sqlInsertRoot} = "(";
foreach my $elem ( @dbFieldsKeys ) {
    $tableData{sqlInsertRoot} .= $elem . ',';
}
$tableData{sqlInsertRoot} =~ s/,$//;
$tableData{sqlInsertRoot} .= ") VALUES";

resetQueryString();

$rootDir =~ s/\/+$//;
my @fileList = split(' ', `ls $rootDir/$logger/*-ath$interface.tcpdump*`);
foreach my $f (sort @fileList) {
    processPackets($f);
    flushInserts( 0 );
}

dbdisconnect();

mirrorPrint("\nTerminated normally(?)\n");
mirrorPrint(now(), "\n");
