#!/usr/bin/perl

# looks for packets that must have been sent
# but not heard by the monitor.
# Also splits input table into two tables: processed_xxx and processed_xxx_Extras

package LostPackets;

use strict;
use DBI;
use Time::HiRes qw( gettimeofday );
use lostPacketsCommon;

require './lostPacketsFSA.sub';
require 'localutils.pl';

my $basicInsertQuery;
my $basicInsertQueryRoot;

my $extrasInsertQuery;
my $extrasInsertQueryRoot;

my $synInsertQuery;
my $synInsertQueryRoot;
my $synInsertQueryCount = 0;
my $synInsertNextId = 1;

my $lockQuery;

my $insertQueryCount = 0;
my $nextInsertID = 1;

my $testPacketID = 1;

my @pendingPacketList;

my $CURRENT_PACKET_NUM = 0;

# these next two are hashes of cached states.  There are two because
# we want to limit memory usage, and use a second chance-like scheme
# to probabilistically select the most referenced states to keep around.
my $currentCachedStateSet;
my $oldCachedStateSet;
my $cssNewHit = 0;
my $cssOldHit = 0;
my $cssMiss = 0;
my $numCurrentCachedStates = 0;
my $numOldCachedStates = 0;
my $cssGeneration = 0;
my $cssTimestamp;
my $cssPacketNum;
my $STATE_CACHE_MAX = 6000;


my $READ_GROUPSIZE = 1000;
my $WRITE_GROUPSIZE = 1000;
my $DUMP_INTERVAL = 1000000;
my $PRINT_INTERVAL = 10;

my $wantDistinctPaths;
my $wantAPWarnings = 1;

my $logFile;

my @subtypeMap;   # subtype number -> name

my @fieldType;  # filled with sql types of the fields in the source table

# indices into packet array
my $ID_FIELD;
my $RETRY_FIELD;
my $SEQNO_FIELD;
my $MOREFRAG_FIELD;
my $FRAGNO_FIELD;
my $A1_DST_FIELD;
my $A2_SRC_FIELD;
my $A3_FIELD;
my $A4_FIELD;
my $DSFROM_FIELD;
my $DSTO_FIELD;
my $DUR_FIELD;
my $RATE_FIELD;
my $SIZE_FIELD;
my $SUBTYPE_FIELD;
my $TYPE_FIELD;
my $TIME_FIELD;

# these fields are appended to the packets read from the source db table
my $SRCID_FIELD;        # row id from the src table
my $BSSID_FIELD;        # (possibly inferred) BSSID
my $CONVID_FIELD;       # Each sequence of messages gets an id...
my $PATHID_FIELD;       # The path id (type of path) this conversation belongs to
my $DIR_FIELD;          # (possibly inferred) direction
my $CLIENT_FIELD;       # who do I think the client of this packet is?
my $PARTNER_FIELD;      # partner of this packet
my $PARSESYMBOL_FIELD;  # Artificial - parse symbol for reg exp
my $WASRCVD_FIELD;      # Artificial: 0 or 1, indicating whether the dst received the packet
my $SYN_FIELD;          # Artificial: 0 or 1, indicating whether there are packets in the syn table immediately before or after this one
# the following are not stored to any table
my $PENDING_FIELD;      # indicates whether we're done processing the packet; not written to DB
my $PATHDIR_FIELD;      # DIR_UP or DIR_DOWN, in the sense of the conversation.  (Special rules apply for ad hoc communications).

my $NUM_ALL_FIELDS;     # Index of the last field in the inflated packet array

# synthetic packet fields
my $SYNID_FIELD        = 0;
my $SYNSUBTYPE_FIELD   = 1;
my $SYNA1_DST_FIELD    = 2;
my $SYNA2_SRC_FIELD    = 3;
my $SYNRETRY_FIELD     = 4;
my $SYNSEQNO_FIELD     = 5;
my $SYNCONVID_FIELD    = 6;
my $SYNDIR_FIELD       = 7;
my $SYNWASRCVD_FIELD   = 8;
my $SYNPRIORID_FIELD   = 9;
my $SYNPRIORTIME_FIELD = 10;
my $SYNLATERID_FIELD   = 11;
my $SYNLATERTIME_FIELD = 12;
my $SYNPARSESYMBOL_FIELD = 13;

my $CONVERSATION_ID = 0;

use constant {
  SYNTH_AFTER =>  1,
  SYNTH_BEFORE => 2,
};

#----------------------------------------------------------------------------------------------------------
# FSA-related...

# symbol definitions

use constant { MAX_FRAGNO => 6 };

my %SYM;           # sym name -> symbol id
my @symWeight;     # sym id -> weight  (higher weights are more expensive)
my @symNameMap;    # sym id -> name
my @symSubtypeMap; # sym id -> subtype (or undef)
# the values correspond to the names of the $SUBTYPE_XXX variables
# we use them to construct @symSubtypeMap map
my @symList = qw( IGNORED EPSILON WILDCARD MARKER
                  BEACON
                  ACK UNMATCHED_ACK
                  RTS CTS UNMATCHED_CTS
                  DATA_0 DATA_0R
                  PS_POLL
                  PROBE_REQ PROBE_RESP PROBE_RESPR
                  AUTH AUTHR
                  DEAUTH DEAUTHR
                  ASSOC_REQ  ASSOC_REQR
                  ASSOC_RESP ASSOC_RESPR
                  REASSOC_REQ  REASSOC_REQR
                  REASSOC_RESP REASSOC_RESPR
                  DISASSOC  DISASSOCR
               );
#                  FRAG_0 FRAG_0R DATA_1 DATA_1R
#                  FRAG_1 FRAG_1R DATA_2 DATA_2R
my $max_fragno;

my %symNameToSubtypeMap = ( IGNORED => undef,
	        EPSILON => undef,
		WILDCARD => undef,
                MARKER => undef,
	        BEACON =>'BEACON',
		ACK => 'ACK',
                UNMATCHED_ACK => 'ACK',
		RTS => 'RTS',
		CTS => 'CTS',
		UNMATCHED_CTS => 'CTS',
		PS_POLL => 'PS_POLL',
		PROBE_REQ => 'PROBE_REQ',
		PROBE_RESP => 'PROBE_RESP',
		PROBE_RESPR => 'PROBE_RESP',
		DATA_0 => 'DATA',
		DATA_0R => 'DATA',
		AUTH => 'AUTH',
		AUTHR => 'AUTH',
		DEAUTH => 'AUTH',
		DEAUTHR => 'AUTH',
		ASSOC_REQ =>'ASSOC_REQ',
		ASSOC_REQR => 'ASSOC_REQ',
                ASSOC_RESP => 'ASSOC_RESP',
                ASSOC_RESPR => 'ASSOC_RESP',
		REASSOC_REQ =>'REASSOC_REQ',
		REASSOC_REQR => 'REASSOC_REQ',
                REASSOC_RESP => 'REASSOC_RESP',
                REASSOC_RESPR => 'REASSOC_RESP',
		DISASSOC =>'DISASSOC',
		DISASSOCR => 'DISASSOC',
	       );
#		FRAG_0 => 'DATA',
#               FRAG_0R => 'DATA',
#		DATA_1 => 'DATA',
#		DATA_1R => 'DATA',
#		FRAG_1 => 'DATA',
#		FRAG_1R => 'DATA',
#		DATA_2 => 'DATA',
#		DATA_2R => 'DATA',

# symbol value constants
my $SYM_IGNORED;
my $SYM_EPSILON;
my $SYM_WILDCARD;
my $SYM_MARKER;
my $SYM_BEACON;
my $SYM_ACK;
my $SYM_RTS;
my $SYM_CTS;
my $SYM_PS_POLL;
my $SYM_PROBE_REQ;
my $SYM_PROBE_RESP;
my $SYM_PROBE_RESPR;
my $SYM_DATA_0;
my $SYM_DATA_0R;
my $SYM_FRAG_0;
my $SYM_FRAG_0R;
my $SYM_DATA_1;
my $SYM_DATA_1R;
my $SYM_FRAG_1;
my $SYM_FRAG_1R;
my $SYM_DATA_2;
my $SYM_DATA_2R;
my $SYM_FRAG_2;
my $SYM_FRAG_2R;
my $SYM_DATA_3;
my $SYM_DATA_3R;
my $SYM_FRAG_3;
my $SYM_FRAG_3R;
my $SYM_DATA_4;
my $SYM_DATA_4R;
my $SYM_FRAG_4;
my $SYM_FRAG_4R;
my $SYM_DATA_5;
my $SYM_DATA_5R;
my $SYM_FRAG_5;
my $SYM_FRAG_5R;
my $SYM_DATA_6;
my $SYM_DATA_6R;
my $SYM_FRAG_6;
my $SYM_FRAG_6R;
my $SYM_DATA_7;
my $SYM_DATA_7R;
my $SYM_AUTH;
my $SYM_AUTHR;
my $SYM_DEAUTH;
my $SYM_DEAUTHR;
my $SYM_ASSOC_REQ;
my $SYM_ASSOC_REQR;
my $SYM_ASSOC_RESP;
my $SYM_ASSOC_RESPR;
my $SYM_REASSOC_REQ;
my $SYM_REASSOC_REQR;
my $SYM_REASSOC_RESP;
my $SYM_REASSOC_RESPR;
my $SYM_DISASSOC;
my $SYM_DISASSOCR;

my @isRetrySym;

my $FSA;
my $START_STATE;
my $ACCEPT_STATE;
my $WILDCARD_STATE;

#                  (   [DATA_0:0 ACK:1,$,-] ([RTS:0 CTS:1,-,<]? [DATA_0R:0 ACK:1,$,-])+
#                    | [FRAG_0:0 ACK:1,-,([RTS:0 CTS:1,-,<]? [FRAG_0R:0 ACK:1,$,-])+ ] [RTS:0 CTS:1,-,<]?
#                         (  [DATA_1:0 ACK:1, $, -] ([RTS:0 CTS:1,-,<]? [DATA_1R:0 ACK:1,$,-])+
#                          | [FRAG_1:0 ACK:1, -, ([RTS:0 CTS:1,-,<]? [FRAG_1R:0 ACK:1,$,-])+ ] [RTS:0 CTS:1,-,<]?
#                                [DATA_2:0 ACK:1,$,-] ([RTS:0 CTS:1,-,<]? [DATA_2R:0 ACK:1,$,-])+
#                         )
#                   )
my $FSA_INPUT = 
'
  [RTS:0 CTS:1,-,$]?
                  DATAPATTERN:0
| [PS_POLL:0,-,$]
                  DATAPATTERN:1
| [PS_POLL:0 ACK:1,$,$]
| [PROBE_REQ:0, -, $] ([PROBE_RESP:1 ACK:0, $, -] ([PROBE_RESPR:1 ACK:0, $, -])*)* [EPSILON:0, $, $]
| [BEACON:0, $, <]
| [AUTH:0 ACK:1,-,-][AUTHR:0 ACK:1,-,-]*
       ([AUTH:1 ACK:0,$,-] ([AUTHR:1 ACK:0, $,-]+))*
| [DEAUTH:0 ACK:1,-,-][DEAUTHR:0 ACK:1,-,-]*
| [ASSOC_REQ:0 ACK:1,-,-][ASSOC_REQR:0 ACK:1,$,-]*
      ([ASSOC_RESP:1 ACK:0,$,-][ASSOC_RESPR:1 ACK:0,$,-]+)*
| [REASSOC_REQ:0 ACK:1,-,-][REASSOC_REQR:0 ACK:1,$,-]*
      ([REASSOC_RESP:1 ACK:0,$,-][REASSOC_RESPR:1 ACK:0,$,-]+)*
| [DISASSOC:0 ACK:1,-,-][DISASSOCR:0 ACK:1,$,-]*
';

#                  (   [DATA_0:1 ACK:0,$,-] ([RTS:1 CTS:0,-,<]? [DATA_0R:1 ACK:0,$,-])+
#                    | [FRAG_0:1 ACK:0,-,([RTS:1 CTS:0,-,<]? [FRAG_0R:1 ACK:0,$,-])+ ] [RTS:1 CTS:0,-,<]?
#                         (  [DATA_1:1 ACK:0, $, -] ([RTS:1 CTS:0,-,<]? [DATA_1R:1 ACK:0,$,-])+
#                          | [FRAG_1:1 ACK:0, -, ([RTS:1 CTS:0,-,<]? [FRAG_1R:1 ACK:0,$,-])+ ] [RTS:1 CTS:0,-,<]?
#                                [DATA_2:1 ACK:0,$,-] ([RTS:1 CTS:0,-,<]? [DATA_2R:1 ACK:0,$,-])+
#                         )
#                   )

#my $FSA_INPUT = '  [RTS:0 CTS:1,-,$]?
#                                   (   [DATA_0:0 ACK:1,$,-] ([RTS:0 CTS:1,-,<]? [DATA_0R:0 ACK:1,$,-])+
#                                     | [FRAG_0:0 ACK:1,-,([RTS:0 CTS:1,-,<]? [FRAG_0R:0 ACK:1,$,-])+ ] [RTS:0 CTS:1,-,<]?
#                                          (  [DATA_1:0 ACK:1, $, -] ([RTS:0 CTS:1,-,<]? [DATA_1R:0 ACK:1,$,-])+
#                                           | [FRAG_1:0 ACK:1, -, ([RTS:0 CTS:1,-,<]? [FRAG_1R:0 ACK:1,$,-])+ ] [RTS:0 CTS:1,-,<]?
#                                                 [DATA_2:0 ACK:1,$,-] ([RTS:0 CTS:1,-,<]? [DATA_2R:0 ACK:1,$,-])+
#                                          )
#                                    )
#                 | [PS_POLL:0,-,$]
#                                   (   [DATA_0:1 ACK:0,$,-] ([RTS:1 CTS:0,-,<]? [DATA_0R:1 ACK:0,$,-])+
#                                     | [FRAG_0:1 ACK:0,-,([RTS:1 CTS:0,-,<]? [FRAG_0R:1 ACK:0,$,-])+ ] [RTS:1 CTS:0,-,<]?
#                                          (  [DATA_1:1 ACK:0, $, -] ([RTS:1 CTS:0,-,<]? [DATA_1R:1 ACK:0,$,-])+
#                                           | [FRAG_1:1 ACK:0, -, ([RTS:1 CTS:0,-,<]? [FRAG_1R:1 ACK:0,$,-])+ ] [RTS:1 CTS:0,-,<]?
#                                                 [DATA_2:1 ACK:0,$,-] ([RTS:1 CTS:0,-,<]? [DATA_2R:1 ACK:0,$,-])+
#                                          )
#                                    )
#                 | [PS_POLL:0 ACK:1,$,$]
#                 | [PROBE_REQ:0, -, $] ([PROBE_RESP:1 ACK:0, $, -] ([PROBE_RESPR:1 ACK:0, $, -])*)* [EPSILON:0, $, $]
#                 | [BEACON:0, $, <]
#                 ';

#my $FSA_INPUT = ' [DATA_0:0 ACK:1, $, $]';

# "end of phrase" marker values
my $MARKER_NONE = -1;  # must be a value impossible to get from parsesymbolIndex
my $MARKER_SYM;        # init'ed in code (to correspond to symbol from @symList)

#----------------------------------------------------------------------------------------------------------
# time constants

my $MARKER_TIMEOUTVALUE = 0.25;      # if we don't hear any packets in a conversation for 0.25 secs, it's over...
my $AP_TIMEOUT          = 20.0;      # if we don't see a BSSID used for 20 secs., retire it
my $SIFS                = 10.0;      # usec.

#----------------------------------------------------------------------------------------------------------
# client data

my @clientData;           # has the following fields...

use constant {
  CLIENT_NAME                    => 0,  # the MAC
  CLIENT_AP_TYPE                 => 1,  # holds a XXX_AP value, reflecting confidence that this mac is an AP
  CLIENT_LAST_USE_AS_AP          => 2,  # last time we noticed this node was used as a BSSID
  CLIENT_CURRENT_CONV            => 3,  # holds a uid for the client's current conversation
  CLIENT_CURRENT_CONV_PARTNER    => 4,  # who is client talking with?
  CLIENT_CONV_LAST_ACTIVITY_TIME => 5,  # timestamp of last packet attributed to current conversation 
  CLIENT_CONV_LAST_SEQNO         => 6,  # for DATA exchanges, the sequence number of the current conversation
                                        # This is an array, indexed by $DIR_UP/DIR_DOWN
  CLIENT_CONV_LAST_FRAGNO        => 7,  # for DATA exchanges, the sequence number of the current conversation
                                        # Also an array, indexed by dir
  CLIENT_LAST_KNOWN_BSSID        => 8,  # We keep track of client's BSSID at times it seems reliable
  CLIENT_LAST_KNOWN_PARTNER      => 9,  # Keep track of the last partner the client talked with (a best guess, at times)
  CLIENT_PACKET_LIST             => 10, # array of packet refs that led to the state(s) we're in
  CLIENT_STATESET                => 11, # ... is undef if client is not currently involved in a conversation.
                                        # Otherwise, it's [ state0, state1, ...]
                                        #    where each state is [ stateID, pathCost, [ \edge, \edge, ... ] ]
  CLIENT_PROBEREQ_DATA           => 12, # ->[ probeReq dest addr, { AP->last seqno} ]
  CLIENT_RTS_DURVEC              => 13, # if current conversation involves RTS messages, keep track of dur fields by rate
  CLIENT_NUM_FIELDS              => 13,
};

# CLIENT_AP_TYPE values
use constant {
  AP_NONE           => 0,
  AP_ADHOC          => 1,
  AP_INFRASTRUCTURE => 2,
};

# elements of each state descriptor in clientData[$client][$CLIENT_STATESET]
use constant {
  STATE_ID          => 0,
  STATE_CONV_PARITY => 1,
  STATE_PATHCOST    => 2,
  STATE_PATHLIST    => 3,
};

# elements of each CLIENT_PROBEREQ_DATA vector
use constant {
  PROBEREQD_DEST    => 0,
  PROBEREQD_SEQHASH => 1,
};

#----------------------------------------------------------------------------------------------------------
# pathsequence data

# path sequence data
my $NEXT_PATHSEQUENCE_ID = 0;
my $pathMap = {};      # ->{sym}->{sym}->...->{sym} according to path.
                       #      ->{id} at any node gives an index into the following arrays
my @pathSequence;          # [id] is an arrayRef:
use constant {
  PATHSEQ_PATH      => 0, #    {path} -> [sym, sym, ..., sym] (i.e, the path)
  PATHSEQ_COUNT     => 1, #    {count}->[dir]-> the number of times the sequence was seen
  PATHSEQ_HEARD_CNT => 2, #    {heard}->[dir]->[] where [j] is the number of jth messages in path that were heard by the monitor
  PATHSEQ_SYNTH_CNT => 3, #    {synthesized}->[dir]-> [] where [j] is the number of jth messages in the path that were cons'ed up
  PATHSEQ_RCVD_FLAG => 4, #    {received}->[] where [j] is 1 if the edge traversed is marked as receiver correctly receives, 0 otherwise
  PATHSEQ_ID        => 5, #    {pathid}->int  the id of this path (the index in @pathSequence
};

#----------------------------------------------------------------------------------------------------------
# by-symbol data

my @symCount;    # $symCount[$sym]->[$dir] is total occurrences of symbol in that dir across all paths
my @symHeard;    # $symHeard[$sym]->[$dir]    -- total number heard by the monitor (synthesized is calculated from this plus the total)
my @symReceived; # $symReceived[$sym]->[$dir]->[notheard|heard|noinfo] -- total number received by the client

#----------------------------------------------------------------------------------------------------------

# WARNING! These constants must be kept consistent with those in
# lostPacketsFSA.pl  (don't know how to make them shared in perl...)
use constant{
    edge_fromState   => 0,
    edge_nextState   => 1,
    edge_path        => 2,
    edge_symbol      => 3,
    edge_parity      => 4,
    edge_wasReceived => 5,
    edge_weight      => 6,
    edge_pathWeight  => 7,        # weight used in resolving multiple paths to
	                          # accept state, given actual trace data
};

# subtype values
my $SUBTYPE_DATA;
my $SUBTYPE_ACK;
my $SUBTYPE_RTS;
my $SUBTYPE_CTS;
my $SUBTYPE_PS_POLL;
my $SUBTYPE_PROBE_REQ;
my $SUBTYPE_PROBE_RESP;
my $SUBTYPE_AUTH;
my $SUBTYPE_ASSOC_REQ;
my $SUBTYPE_ASSOC_RESP;
my $SUBTYPE_BEACON;

# These items must be kept consistent with lostPacketsFSA.sub!!!
use constant {
  PARITY_ANY   => -1,
  PARITY_UP    =>  0,
  PARITY_DOWN  =>  1,
};

# not fully used as of 1/6/2006
my $SUBTYPE_CFEND_ACK;
my $SUBTYPE_CFEND;
my $SUBTYPE_REASSOC_REQ;
my $SUBTYPE_REASSOC_RESP;
my $SUBTYPE_DISASSOC;
my $SUBTYPE_AUTH;
my $SUBTYPE_DEAUTH;

# type values
use constant {
  TYPE_MGMT    => 0,
  TYPE_CONTROL => 1,
  TYPE_DATA    => 2,
};

# dir field values
# The values of DIR_UP and DIR_DOWN are important to the implementation, in two ways:
#   - We assume DIR_X + 1 % 2 = DIR_Y (for understanding the direction of each edge in a path)
#   - We use the packet direction as an index:
#            * in generateMarker(): $clientData[$client][CLIENT_CONV_LAST_SEQNO][$pkt->[$DIR_FIELD]]
#            * in terminateConverstation:
use constant {
  DIR_UP       => 0,   # values of DIR_UP and DIR_DOWN are significant!  (See above)
  DIR_DOWN     => 1,   # values of ....
  DIR_ADHOC    => 2,
  DIR_APAP     => 3,
};

# Protocol violation IDs
my $PV_SEQNO = 0;
my $totalProtocolViolations = 0;
my @protocolViolation = ( { printString   => '{ "Sequence number $pkt->[$SEQNO_FIELD] repeated but retry is 0" }',
			    summaryString => 'Sequence number is repeated but retry is 0',
			    byNode        => {},
			    bySymbol      => {},
			  },
			 );

my $BROADCAST_NODE;
my $UNKNOWN_NODE;        # used (somtimes) when we can't figure out who the client is
my $NOTASSOCIATED_NODE;  # 44:44:44:44:44:44 -- used by some cards to indicate not associated

use constant { ALWAYS => 0, DEBUG_WARN => 1, DEBUG_FULL => 2 };
my $debug = 0;

use constant { NO_RECEIVED_INFO => 2, WAS_RECEIVED => 1, WASNT_RECEIVED => undef };

my $limitCount = 1000*1000*1000;
my $startID = 1;

my $currentTime = 0;
my $lastScanTime = 0;
my $SCAN_TIMEOUT = 2;  # maximum elapsed time between scanPendingPackets() calls.

#----------------------------------------------------------------------------
# splitting-related vars
my @usefulFields = qw(id aFrameTime bRate bSize cA1_Dest cA2_Src
                      cDSFrom cDSTo cDur cFragno cMorefrag cRetry cSeqno
                      cSubtype zErr
                     );
my %isUsefulField;

foreach (@usefulFields) { $isUsefulField{$_} = 1; }  # register the useful field types

my @usefulFieldNames;
my @usefulFieldTypes;
my @usefulIndices;

my @otherFieldNames;
my @otherFieldTypes;
my @otherIndices;

#----------------------------------------------------------------------------
# Globals allocated out here for performance reasons.  This is stupid because:
#  (a) it's always dangerous to distort the code for performance reasons, and
#  (b) there's no clear info on what perl is doing, and no real idea if this helps!
my @candidateStateSet;
my @newStatePathCost;

my $USE_RTS_CHECK = 0;


############################################
# initLog()
############################################

sub initLog {
    my ( $basicOutputTable, $dataFile ) = @_;

    my $mainTable = $dataFile;  # used if -dat specified (because $basicOutputTable should be undef)
    $mainTable = $basicOutputTable if !defined($mainTable);

    my $fileName = $mainTable . '.log';
    open OUTFILE, ">$fileName" || die "Couldn't open log file $fileName";
    $logFile = *OUTFILE;
    setLogFile($logFile);
    #$| = 1;

    mirrorPrint( "Logging to $fileName\n" );
}

sub Log {
    my $level = shift;
    if ($level <= $debug ) {
      my $str;
      mirrorPrint($str) while ( defined ($str = shift) );
    }
}

sub Warn {
  my ($pktNum, $warnNum) = @_;
  mirrorPrint("WARNING($pktNum): $LostPackets::warningInfo[$warnNum]->[LPWARN_MSG]\n");
}

sub flagProtocolViolation {
    my ($pvID, $client, $pkt) = @_;

    Log ALWAYS, "PROTOCOL VIOLATION: client $client at packet $pkt->[$ID_FIELD] (",
		           $symNameMap[$pkt->[$PARSESYMBOL_FIELD]], ") [", stateIDList($client), "]\n",
		 "\t", eval $protocolViolation[$pvID]->{printString}, "\n";

    $protocolViolation[$pvID]->{byNode}->{$client}++;
    $protocolViolation[$pvID]->{bySymbol}->{$pkt->[$PARSESYMBOL_FIELD]}++;
    $totalProtocolViolations++;
}

############################################
# initSymTable()
############################################

sub initSymTable {
  my ($wtFile, $wtTable) = @_;

  my $uid = 0;

  # make sure the parse symbol index table is current
  my $query = 'DROP TABLE IF EXISTS _zParseSymbolIndex';
  doQuery(\$query);
  $query = 'CREATE TABLE _zParseSymbolIndex ( id int UNSIGNED NOT NULL,
                                              value char(48))';
  doQuery(\$query);
  $query = 'INSERT INTO _zParseSymbolIndex (id, value) VALUES ';

  foreach (@symList) {
    $query .= "($uid, '$_'),";
    $symNameMap[$uid] = $_;
    eval( "\$SYM_$_ = $uid" );
    $SYM{$_} = $uid;

    eval( "\$symSubtypeMap[$uid] = \$SUBTYPE_$symNameToSubtypeMap{$_}" );

    # keep flag about whether or not the symbol is a retry symbol.
    # One is iff it ends with 'R' and is not 'MARKER'
    $isRetrySym[$uid] = ( m/R$/ && $_ ne 'MARKER' ) ? 1 : 0;
    $uid++;
  }

  foreach (@symList) {
    die "No value assigned to symSubtypeMap for $_" if defined($symNameToSubtypeMap{$_}) && !defined($symSubtypeMap[$SYM{$_}]);
  }

  chop $query;
  doQuery(\$query);
  Log ALWAYS, "(Re)created _zParseSymbolIndex table\n";

  $MARKER_SYM = $SYM_MARKER;

  # read a file of sym weights, if there is one
  # file format:  symbol  weight  
  # weights should be in range 0...1, normally

  if ( $wtFile && open(WTFILE, "<$$wtFile") ) {
    Log ALWAYS, "Using symbol weights from file $$wtFile\n";
    while ( <WTFILE> ) {
      next if /^\s*#/;
      next if /^\s*$/;
      my @field = split;
      $symWeight[$SYM{$field[0]}] = $field[1];
    }
  }
  else {
    Log ALWAYS, "Assigning equal weight to each symbol (fewest hops results).\n";
    foreach (@symList) {
      $symWeight[$SYM{$_}] = 1;
    }
  }

  foreach (@symList) {
    die "No symbol weight assigned for $_" if !defined($symWeight[$SYM{$_}]);
  }

  # write weights to a table, to document how the other output tables were constructed

  if ( $$wtTable ) {
    my $query = "DROP TABLE IF EXISTS $$wtTable";
    doQuery( \$query );

    # we redudantly write the id so that we can have both an 'id' field (for consistency)
    # and a specially named field that can be marked as invertible for query.cgi

    $query = "CREATE TABLE $$wtTable ( id int NOT NULL UNIQUE,
                                       _zParseSymbol int UNSIGNED,
                                       weight float,
                                       PRIMARY KEY( id ) )";
    doQuery( \$query );

    $query = "INSERT INTO $$wtTable VALUES ";
    for ( $uid=0; $uid<=$#symWeight; $uid++ ) {
      $query .= "($uid,$uid,$symWeight[$uid]),";
    }
    chop $query;
    doQuery( \$query );
  }
}

############################################
# fetchTestPacket()
############################################

my $TESTPKT_TIMEDELTA = 0;
my $TESTPKT_SUBTYPE   = 1;
my $TESTPKT_DSFROM    = 2;
my $TESTPKT_DSTO      = 3;
my $TESTPKT_MOREFRAG  = 4;
my $TESTPKT_FRAGNO    = 5;
my $TESTPKT_RETRY     = 6;
my $TESTPKT_SEQNO     = 7;
my $TESTPKT_SRC       = 8;
my $TESTPKT_DEST      = 9;

my $lastTestPktTime = 0;

sub fetchTestPacket {
  my $fh = shift;
  my $pkt;
  my $line;

  while ( ($line = <$fh>) && ( ($line =~ m/^\s*#/) || ($line =~ m/^\s*$/) ) ) {}
  return undef if !defined($line);

  my @field = split(' ', $line);

  $lastTestPktTime += $field[$TESTPKT_TIMEDELTA] / 1000000;

  $pkt = [];
  $pkt->[$TIME_FIELD]    = $lastTestPktTime;
  $pkt->[$SUBTYPE_FIELD] = $field[$TESTPKT_SUBTYPE];
  $pkt->[$DSFROM_FIELD]  = $field[$TESTPKT_DSFROM];
  $pkt->[$DSTO_FIELD]    = $field[$TESTPKT_DSTO];
  $pkt->[$MOREFRAG_FIELD]= $field[$TESTPKT_MOREFRAG];
  $pkt->[$RETRY_FIELD ]  = $field[$TESTPKT_RETRY];
  $pkt->[$SEQNO_FIELD ]  = $field[$TESTPKT_SEQNO];
  $pkt->[$FRAGNO_FIELD]  = $field[$TESTPKT_FRAGNO];
  $pkt->[$A2_SRC_FIELD ] = $field[$TESTPKT_SRC];
  $pkt->[$A1_DST_FIELD ] = $field[$TESTPKT_DEST];

  # fixup/cons some fields

  $pkt->[$ID_FIELD] = $testPacketID++;
  $pkt->[$TYPE_FIELD]    = int($field[$TESTPKT_SUBTYPE] / 16);
  if ( $pkt->[$SUBTYPE_FIELD] == $SUBTYPE_ACK || $pkt->[$SUBTYPE_FIELD] == $SUBTYPE_CTS ) {
    $pkt->[$SIZE_FIELD] = 14;
    $pkt->[$RATE_FIELD] = 1;
  }
  $pkt->[$A3_FIELD] = $pkt->[$A2_SRC_FIELD] if $pkt->[$SUBTYPE_FIELD] == $SUBTYPE_BEACON;

  for ( my $i=0; $i<=$#{$pkt}; $i++ ) {
    undef $pkt->[$i] if $pkt->[$i] eq '-';
  }

  return $pkt;
}

############################################
# expandDataRegExp()
#    Expands macros in reg expressions string
#    to automate deeply nested data/frag
#    expression.
############################################

my $dataPattern = '(   [DATA_NUM:UP ACK:DOWN,$,-] ([RTS:UP CTS:DOWN,-,<]? [DATA_NUMR:UP ACK:DOWN,$,-])+
                    | [FRAG_NUM:UP ACK:DOWN,-,([RTS:UP CTS:DOWN,-,<]? [FRAG_NUMR:UP ACK:DOWN,$,-])+ ] [RTS:UP CTS:DOWN,-,<]?';
#my $dataTailPattern = '[DATA_NUM:UP ACK:DOWN,$,-] ([RTS:UP CTS:DOWN,-,<]? [DATA_NUMR:UP ACK:DOWN,$,-])+';
my $dataTailPattern = '([DATA_NUM:UP ACK:DOWN,-,-] ([RTS:UP CTS:DOWN,-,<]? [DATA_NUMR:UP ACK:DOWN,-,-])+)*';

sub expandDataRegExp {
  my ( $regexp, $maxFragno) = @_;
  while ( $$regexp =~ /DATAPATTERN:(\d)/ ) {
    my $result;
    my $up = $1;
    my $down = ($up+1)%2;

    my $next;
    my $frag;
    for ( $frag=0; $frag<=$maxFragno; $frag++ ) {
      $next = $dataPattern;
      $next =~ s/_NUM/_$frag/g;
      $next =~ s/:UP/:$up/g;
      $next =~ s/:DOWN/:$down/g;
      $result .= $next . "\n";;
    }
    $frag = $maxFragno+1;
    $next = $dataTailPattern;
    # modified to allow any number of successive fragments beyond the max fragno,
    # without keeping track of their individual identities...
#    $next =~ s/_NUM/_$frag/g;
    $next =~ s/_NUM/_MAX/g;
    $next =~ s/:UP/:$up/g;
    $next =~ s/:DOWN/:$down/g;
    $result .= "$next\n";
    for ( my $frag=0; $frag<=$maxFragno; $frag++ ) {
      $result .= ")\n";
    }
    $$regexp =~ s/DATAPATTERN:\d/$result/;
  }
}


############################################
# pushOutput()
#   Pushed buffered insert to db
############################################

sub pushOutput {

  if ( $insertQueryCount || $synInsertQueryCount ) {

    doQuery( \$lockQuery );

    if ( $insertQueryCount ) {
      chop $basicInsertQuery;
      doQuery( \$basicInsertQuery );
      $basicInsertQuery = $basicInsertQueryRoot;

      chop $extrasInsertQuery;
      doQuery( \$extrasInsertQuery );
      $extrasInsertQuery = $extrasInsertQueryRoot;

      $insertQueryCount = 0;
    }

    if ( $synInsertQueryCount ) {
      chop $synInsertQuery;
      doQuery( \$synInsertQuery );
      $synInsertQuery = $synInsertQueryRoot;
      $synInsertQueryCount = 0;
    }

    my $query = 'UNLOCK TABLES';
    doQuery( \$query );
  }
}

########################################################################################
# scanPendingPackets()
#   Look at packet pending fields, and prepare any prefix of them that is ready.
#   We have to output packets in order,  but we don't have all the field values
#   for a packet until some late time.  So, we remember a queue of them here.
#   Their pending field tells us when they're ready to go, which scanPendingPackets()
#   handles.
########################################################################################

sub scanPendingPackets {

  my $client;
  my $packet;

  $lastScanTime = $currentTime;

  while ( $#pendingPacketList>=0 ) {

    $packet = $pendingPacketList[0];
    if ( $debug >= DEBUG_FULL ) {
      mirrorPrint('scanning packet (', join(', ', @{$packet}), ")\n");
    }

    if ( $packet->[$PENDING_FIELD] != 0 ) {
      $client = $pendingPacketList[0]->[$CLIENT_FIELD];
      if ( !defined($client) ) {
	$packet->[$PENDING_FIELD] = 0;
      } elsif ( timedOut($client) ) {
	takeFSAStep( $client, $MARKER_SYM, undef );  # force termination of this conversation (hopefully)
	terminateConversation( $client, -1);
      }
    }

    last if ( $packet->[$PENDING_FIELD] != 0 );

    $packet = shift @pendingPacketList;  # now actually remove the packet

    if ( $debug >= DEBUG_WARN ) {
      mirrorPrint("Inserting packet $packet->[$ID_FIELD] as $nextInsertID: src = $packet->[$A2_SRC_FIELD] ",
		  "dst = $packet->[$A1_DST_FIELD]\n");
    }
    $packet->[$SRCID_FIELD] = $packet->[$ID_FIELD];

    # first buffer the basic fields
    #$basicInsertQuery .= "($nextInsertID," . join(',', map{ defined($packet->[$_])?$packet->[$_]:"NULL" } @usefulIndices ) . '),';
    $basicInsertQuery .= "($nextInsertID,";
    foreach (@usefulIndices) {
      $basicInsertQuery .= defined($packet->[$_]) ? $packet->[$_] . ',' : 'NULL,';
    }
    chop $basicInsertQuery;
    $basicInsertQuery .= '),';

    # now the extra fields
    #$extrasInsertQuery .= "($nextInsertID, $packet->[$ID_FIELD]," .
    #                      join(',', map{ defined($packet->[$_])?$packet->[$_]:"NULL" } @otherIndices ) .
    #                      '),';
    $extrasInsertQuery .= "($nextInsertID, $packet->[$ID_FIELD],";
    foreach (@otherIndices) {
      $extrasInsertQuery .= defined($packet->[$_])?$packet->[$_] . ',':'NULL,';
    }
    chop $extrasInsertQuery;
    $extrasInsertQuery .= '),';

    $nextInsertID++;

    pushOutput if ( ++$insertQueryCount >= $WRITE_GROUPSIZE );

  }
}

####################################################################
# initFieldData()
####################################################################

sub initFieldData {

  my ($srcTable) = (@_);

    # get field names from source table

    my $query = "SHOW FIELDS FROM $srcTable";
    my $sth = doQuery ( \$query );
    my $fieldRef = $sth->fetchall_arrayref();

    $NUM_ALL_FIELDS = $#{$fieldRef};

    for ( my $field=0; $field <= $NUM_ALL_FIELDS; $field++ ) {

      # remember the sql type info for the source fields
      push @fieldType, ("$fieldRef->[$field]->[1] $fieldRef->[$field]->[5]");

      # these fields should appear in every table
      if ($fieldRef->[$field]->[0] eq 'id') {
	$ID_FIELD      = $field;
	next; # we don't push the id field into the useful or other structures
      }
      $A1_DST_FIELD  = $field   if $fieldRef->[$field]->[0] eq 'cA1_Dest';
      $A2_SRC_FIELD  = $field   if $fieldRef->[$field]->[0] eq 'cA2_Src';
      $A3_FIELD      = $field   if $fieldRef->[$field]->[0] eq 'cA3';
      $A4_FIELD      = $field   if $fieldRef->[$field]->[0] eq 'cA4';
      $DSFROM_FIELD  = $field   if $fieldRef->[$field]->[0] eq 'cDSFrom';
      $DSTO_FIELD    = $field   if $fieldRef->[$field]->[0] eq 'cDSTo';
      $DUR_FIELD     = $field   if $fieldRef->[$field]->[0] eq 'cDur';
      $RETRY_FIELD   = $field   if $fieldRef->[$field]->[0] eq 'cRetry';
      $SEQNO_FIELD   = $field   if $fieldRef->[$field]->[0] eq 'cSeqno';
      $FRAGNO_FIELD  = $field   if $fieldRef->[$field]->[0] eq 'cFragno';
      $MOREFRAG_FIELD = $field  if $fieldRef->[$field]->[0] eq 'cMorefrag';
      $TIME_FIELD    = $field   if $fieldRef->[$field]->[0] eq 'aFrameTime';
      $SIZE_FIELD    = $field   if $fieldRef->[$field]->[0] eq 'bSize';
      $RATE_FIELD    = $field   if $fieldRef->[$field]->[0] eq 'bRate';

      $SUBTYPE_FIELD = $field   if $fieldRef->[$field]->[0] eq 'cSubtype';
      $TYPE_FIELD    = $field   if $fieldRef->[$field]->[0] eq 'cType';

      # _z... fields appear only in tables created by this program.
      # We shouldn't be processing them...

      die "Source table $srcTable is a processed table.  Not allowed." if $fieldRef->[$field]->[0] =~ '_z';

      if ( defined($isUsefulField{$fieldRef->[$field]->[0]}) ){
	push @usefulFieldNames, ($fieldRef->[$field]->[0]);
	push @usefulFieldTypes, ("$fieldRef->[$field]->[1] $fieldRef->[$field]->[5]");
	push @usefulIndices, ($field);
      } else {
	push @otherFieldNames, ($fieldRef->[$field]->[0]);
	push @otherFieldTypes, ("$fieldRef->[$field]->[1] $fieldRef->[$field]->[5]");
	push @otherIndices, ($field);
      }
    }

    # new, useful fields

    $BSSID_FIELD       = ++$NUM_ALL_FIELDS;
    push @usefulFieldNames, ('_zBSSID');
    push @usefulFieldTypes,"$fieldRef->[$A4_FIELD]->[1] $fieldRef->[$A4_FIELD]->[5]";
    push @fieldType,"$fieldRef->[$A4_FIELD]->[1] $fieldRef->[$A4_FIELD]->[5]";
    push @usefulIndices, ($BSSID_FIELD);

    $CONVID_FIELD      = ++$NUM_ALL_FIELDS;
    push @usefulFieldNames, ('_zConvId');
    push @usefulFieldTypes, 'int UNSIGNED';
    push @fieldType,'int UNSIGNED';
    push @usefulIndices, ($CONVID_FIELD);

    $DIR_FIELD         = ++$NUM_ALL_FIELDS;
    push @usefulFieldNames, ('_zDir');
    push @usefulFieldTypes, 'tinyint UNSIGNED';
    push @fieldType,'tinyint UNSIGNED';
    push @usefulIndices, ($DIR_FIELD);

    $WASRCVD_FIELD    = ++$NUM_ALL_FIELDS;
    push @usefulFieldNames, ('_zWasrcvd');
    push @usefulFieldTypes, 'tinyint UNSIGNED';
    push @fieldType,'tinyint UNSIGNED';
    push @usefulIndices, ($WASRCVD_FIELD);

    $SYN_FIELD      = ++$NUM_ALL_FIELDS;
    push @usefulFieldNames, ('_zSyn');
    push @usefulFieldTypes, 'tinyint UNSIGNED';
    push @fieldType,'tinyint UNSIGNED';
    push @usefulIndices, ($SYN_FIELD);

    # We also create a separate 'extras' table.  To make things easier as we proceed,
    # the fields that go in that table are appended to packets we read from the source table

    $SRCID_FIELD       = ++$NUM_ALL_FIELDS;
    push @otherFieldNames, ('_zSrcID');
    push @otherFieldTypes, 'int UNSIGNED';
    push @otherIndices, ($SRCID_FIELD);

    $PATHID_FIELD      = ++$NUM_ALL_FIELDS;
    push @otherFieldNames, ('_zPathId');
    push @otherFieldTypes, 'int UNSIGNED';
    push @otherIndices, ($PATHID_FIELD);


    $PARSESYMBOL_FIELD = ++$NUM_ALL_FIELDS;
    push @otherFieldNames, ('_zParseSymbol');
    push @otherFieldTypes, 'tinyint UNSIGNED';
    push @otherIndices, ($PARSESYMBOL_FIELD);

    $CLIENT_FIELD      = ++$NUM_ALL_FIELDS;
    push @otherFieldNames, ('_zClient');
    push @otherFieldTypes, "$fieldRef->[$A4_FIELD]->[1] $fieldRef->[$A4_FIELD]->[5]";
    push @otherIndices, ($CLIENT_FIELD);

    $PARTNER_FIELD     = ++$NUM_ALL_FIELDS;
    push @otherFieldNames, ('_zPartner');
    push @otherFieldTypes, "$fieldRef->[$A4_FIELD]->[1] $fieldRef->[$A4_FIELD]->[5]";
    push @otherIndices, ($PARTNER_FIELD);

}

####################################################################
# initOutputTables()
#   We output two main tables: a copy of the "useful fields" from
#   the src table plus critical _z fields,
#   and a table with src table's other ("extra") fields plus other _z fields.
#   We also output an index table (relating intput and output table
#   names) and create two  copies of the MACIndex table from the src.
####################################################################

sub initOutputTables {
    my ($srcTable, $basicOutputTable, $extrasOutputTable, $synTable ) = @_;

    my $query = "DROP TABLE IF EXISTS $basicOutputTable, $extrasOutputTable, $synTable, $basicOutputTable" . "_PTblIndex, " .
                " $basicOutputTable" . "_MACIndex";
    my $sth = doQuery( \$query );


    # recreate the basic tables
    $query = "CREATE TABLE $basicOutputTable (id int NOT NULL UNIQUE,";
    map {$query .= "$usefulFieldNames[$_] $usefulFieldTypes[$_],"} (0..$#usefulFieldNames);
    $query .= " PRIMARY KEY( id ), KEY( aFrameTime) )";
    doQuery( \$query );

    # remember a useful string and initialize the current query string
    $basicInsertQueryRoot = "INSERT INTO $basicOutputTable VALUES ";
    $basicInsertQuery = $basicInsertQueryRoot;


    # recreate the extras tables
    $query = "CREATE TABLE $extrasOutputTable ( id int NOT NULL UNIQUE, _zOrigID int, ";
    map {$query .= "$otherFieldNames[$_] $otherFieldTypes[$_],"} (0..$#otherFieldNames);
    $query .= " PRIMARY KEY( id ) )";
    doQuery(\$query);

    # we add a few fields to the packet for our use, but they aren't written to any table
    $PATHDIR_FIELD     = ++$NUM_ALL_FIELDS;
    $PENDING_FIELD     = ++$NUM_ALL_FIELDS;

    $extrasInsertQueryRoot = "INSERT INTO $extrasOutputTable VALUES ";
    $extrasInsertQuery = $extrasInsertQueryRoot;

    # recreate the syn table
    # MUST CHANGE THE $SYNXXX_FIELD definitions if this changes!
    $query = "CREATE TABLE $synTable ( id int NOT NULL UNIQUE, 
                                         cSubtype       $fieldType[$SUBTYPE_FIELD],
                                         cA1_Dest       $fieldType[$A1_DST_FIELD],
                                         cA2_Src        $fieldType[$A2_SRC_FIELD],
                                         cRetry         $fieldType[$RETRY_FIELD],
                                         cSeqno         $fieldType[$SEQNO_FIELD],
                                         _zConvId       $fieldType[$CONVID_FIELD],
                                         _zDir          $fieldType[$DIR_FIELD],
                                         _zWasrcvd      $fieldType[$WASRCVD_FIELD],
                                         priorPktId     int,
                                         priorPktTime   $fieldType[$TIME_FIELD],
                                         laterPktId     int,
                                         laterPktTime   $fieldType[$TIME_FIELD],
                                         _zParseSymbol  tinyint UNSIGNED,
                                         PRIMARY KEY( id ),
                                         KEY (_zConvId),
                                         KEY (laterPktId) )";
    doQuery(\$query);

    $synInsertQueryRoot = "INSERT INTO $synTable VALUES ";
    $synInsertQuery = $synInsertQueryRoot;

    # now create the index table

    $query = "CREATE TABLE $basicOutputTable" . "_PTblIndex 
              (id int NOT NULL AUTO_INCREMENT, tablename char(48), function char(48), primary key(id))";
    doQuery(\$query);

    $query = "INSERT INTO  $basicOutputTable" . "_PTblIndex (tablename, function) VALUES
              ('$srcTable', 'source data'),
              ('$basicOutputTable', 'output primary fields'),
              ('$extrasOutputTable', 'output extra fields')";
    doQuery(\$query);

    # copy the MACIndex table

    my $macTable = $basicOutputTable . '_MACIndex';
    $query = "CREATE TABLE $macTable LIKE $srcTable" . '_MACIndex';
    doQuery(\$query);
    $query = "INSERT INTO $macTable SELECT * FROM $srcTable" . '_MACIndex';
    doQuery(\$query);
    $query = "INSERT INTO $macTable VALUES ($UNKNOWN_NODE, 'UNKNOWN')";
    doQuery( \$query );
}

############################################
# initSubtypeMap()
############################################

sub initSubtypeMap {

    # get map of subtype names 

    my $query = "SELECT id,value FROM subtypeIndex";
    my $sth = doQuery( \$query );
    my @row;
    while ( @row = $sth->fetchrow_array() ) {
	$subtypeMap[$row[0]] = $row[1];

	$SUBTYPE_DATA       = $row[0] if $row[1] eq 'Data';
	$SUBTYPE_ACK        = $row[0] if $row[1] eq 'Acknowledgment (ACK)';
	$SUBTYPE_RTS        = $row[0] if $row[1] eq 'Request To Send (RTS)';
	$SUBTYPE_CTS        = $row[0] if $row[1] eq 'Clear To Send (CTS)';
	$SUBTYPE_PS_POLL    = $row[0] if $row[1] eq 'Power Save (PS)-Poll';
	$SUBTYPE_PROBE_REQ  = $row[0] if $row[1] eq 'Probe request';
	$SUBTYPE_PROBE_RESP = $row[0] if $row[1] eq 'Probe response';
	$SUBTYPE_AUTH       = $row[0] if $row[1] eq 'Authentication';
	$SUBTYPE_ASSOC_REQ  = $row[0] if $row[1] eq 'Association request';
	$SUBTYPE_ASSOC_RESP = $row[0] if $row[1] eq 'Association response';

	# no parse symbols for these, but included for completeness...
	$SUBTYPE_CFEND_ACK  = $row[0] if $row[1] eq 'CF-End + CF-Ack';
	$SUBTYPE_CFEND      = $row[0] if $row[1] eq 'Contention-Free (CF)-End';
	$SUBTYPE_BEACON     = $row[0] if $row[1] eq 'Beacon';
	$SUBTYPE_REASSOC_REQ  = $row[0] if $row[1] eq 'Reassociation request';
	$SUBTYPE_REASSOC_RESP = $row[0] if $row[1] eq 'Reassociation response';
	$SUBTYPE_DISASSOC         = $row[0] if $row[1] eq 'Disassociation';
	$SUBTYPE_DEAUTH = $row[0] if $row[1] eq 'Deauthentication';
    }
}

############################################
# initNodeMap()
############################################

sub initNodeMap {
    # Establish the node naming functions

    my ($srcTable) = @_;

    # now fetch the node names

    $UNKNOWN_NODE = -1;
    my $query = "SELECT id,value FROM $srcTable" . "_MACIndex";
    my $sth = doQuery ( \$query );
    while ( my ($node, $mac) = $sth->fetchrow_array() ) {
      $clientData[$node]->[CLIENT_NAME] = $mac;
      $BROADCAST_NODE = $node if $mac eq 'ff:ff:ff:ff:ff:ff';
      $NOTASSOCIATED_NODE = $node if $mac eq '44:44:44:44:44:44';
      $UNKNOWN_NODE = $node+1 if $node >= $UNKNOWN_NODE;
    }
    $clientData[$UNKNOWN_NODE]->[CLIENT_NAME] = 'Unknown';
}

############################################
# initAPs()
#
# 6/14/2006 - This routine no longer used.
#  Instead, we dynamically add and timeout 
#  APs.
############################################

sub initAPs {
  # Use the existing AP list table, if it exists.
  # Otherwise, create it (and use it).

  my ($srcTable) = @_;

  # is there an existing table?

  my $apTable = $srcTable. '_APList';
  my $query = "SHOW TABLES LIKE '$apTable'";
  my $sth = doQuery( \$query );
  if ( my $tbl = $sth->fetchrow_array ) {
    # if we get here, there's a table.  It might be an out of date format, though...
    Log ALWAYS, "AP table exists: $apTable\n";
    $sth->finish;

    $query = "SHOW FIELDS FROM $apTable";
    $sth = doQuery ( \$query );
    my $fieldRef = $sth->fetchall_hashref('Field');

    if ( defined($fieldRef->{bssType}) ) {
      # APList table exists and looks good.  Read it.
      $query = "SELECT apID,bssType FROM $apTable";
      $sth = doQuery( \$query );
      while ( my ($node,$type) = $sth->fetchrow_array ) {
	next unless defined $node;
	next if $node == $BROADCAST_NODE;
	$clientData[$node]->[CLIENT_AP_TYPE] = $type;
      }
      return;   # LOOK OUT -- RETURNING HERE<<<<<<<<<<<<<<<<<<<<<<<<<<<<
    }

    Log ALWAYS, "AP table is in out of date format.  Deleting it.\n";
    $query = "DROP TABLE $apTable";
    $sth = doQuery( \$query );
  }

  # table doesn't exist -- use a heuristic to find APs and create it

  Log ALWAYS, "Creating cached table of APs: $apTable\n";

  $query = "CREATE TABLE $apTable ( id      int NOT NULL UNIQUE,
                                    apID    $fieldType[$A1_DST_FIELD],
                                    bssType smallint,
                                    macStr  char(18),
                                    PRIMARY KEY(id),
                                    UNIQUE KEY(apID))";
  doQuery( \$query );

  # This will take a while...
  #$query = "SELECT cA2_Src,COUNT(*) FROM $srcTable where cSubtype=$SUBTYPE_ASSOC_RESP and zErr is null group by 1";
  # Updated 6/7/2006.  This still uses only a subset of the plausible information available, but it
  # seems very likely to catch everything.
  $query = "SELECT cA3,COUNT(*) FROM $srcTable where cSubtype=$SUBTYPE_BEACON and zErr is null group by 1";
  $sth = doQuery( \$query );

  my $type;
  my $id = 1;
  $query = '';
#  if ( defined $NOTASSOCIATED_NODE ) {
#    $type = AP_ADHOC;
#    $query = "(1, $NOTASSOCIATED_NODE, $type, '$clientData[$NOTASSOCIATED_NODE]->[CLIENT_NAME]'),";
#    $id++;
#  }
  while ( my ($node, $count) = $sth->fetchrow_array ) {
    next if $node == $BROADCAST_NODE;
    if ( $node == $NOTASSOCIATED_NODE ) {
      $type = AP_ADHOC;
    }
    else {
      $type = AP_INFRASTRUCTURE;
      my $digit = substr($clientData[$node]->[CLIENT_NAME],1,1);
      $type = AP_ADHOC if '2367abef' =~ m/$digit/i;
    }
    $clientData[$node]->[CLIENT_AP_TYPE] = $type;
    $query .= "($id, $node, $type, '$clientData[$node]->[CLIENT_NAME]'),";
    $id++;
    if ( $id % 100 == 0 ) {
      chop $query;
      $query = "INSERT INTO $apTable VALUES " . $query;
      doQuery( \$query );
      $query = '';
    }
  }
  if ( $query ) {
    chop $query;
    $query = "INSERT INTO $apTable VALUES " . $query;
    doQuery( \$query );
  }
}

############################################
# initClientData()
############################################

sub initClientData {
  # CLIENT_NAME has already been set (I hope)

  for ( my $c =0; $c<=$#clientData; $c++ ) {
    $clientData[$c]->[CLIENT_CURRENT_CONV] = -1;
    $clientData[$c]->[CLIENT_CURRENT_CONV_PARTNER] = -1;
    $clientData[$c]->[CLIENT_CONV_LAST_ACTIVITY_TIME] = -1;
    $clientData[$c]->[CLIENT_CONV_LAST_FRAGNO] = -1;
    $clientData[$c]->[CLIENT_LAST_KNOWN_BSSID] = -1;
    $clientData[$c]->[CLIENT_LAST_KNOWN_PARTNER] = -1;
    $clientData[$c]->[CLIENT_AP_TYPE] = AP_NONE;
  }
}

############################################
# formatCacheStats()
############################################

sub formatCacheStats {

  my $tpp;
  $tpp = (gettimeofday - $cssTimestamp) * 1000000.0 / ($CURRENT_PACKET_NUM - $cssPacketNum)
    if $CURRENT_PACKET_NUM - $cssPacketNum > 0;

  my $pid = $$;

  my $str = "Cached state statistics [Generation=$cssGeneration, packet=$CURRENT_PACKET_NUM]:\n";
  $str .= "\tSize: $numCurrentCachedStates / $numOldCachedStates\n";
  $str .= "\t" . sprintf("%d (%d / %d) ", $cssNewHit+$cssOldHit, $cssNewHit, $cssOldHit) . " hits (new/old)\n";
  $str .= "\t$cssMiss misses\n";
  $str .= "\t" . sprintf("%4.1f", ($cssNewHit+$cssOldHit)/($cssNewHit+$cssOldHit+$cssMiss)*100.0) . "% hit rate\n"
    if $cssNewHit+$cssOldHit+$cssMiss > 0;
  $str .= "\t$tpp usecs. / packet\n" if defined($tpp);
  $str .= "\t" . `cat /proc/$pid/status | grep VmSize`;

  $cssPacketNum = $CURRENT_PACKET_NUM;
  $cssTimestamp = gettimeofday;

  return $str;
}

############################################
# dumpResults()
############################################

sub dumpResults {
  my ($count, $useDB, $doPrint ) = @_;

  Log ALWAYS, "\n", now(), "\n";
  Log ALWAYS, "Processed ", commify($count), " packets\n";
  Log ALWAYS, "Pending packet queue size: $#pendingPacketList\n";
  Log ALWAYS, "$NEXT_PATHSEQUENCE_ID paths\n";
  Log ALWAYS, "$totalProtocolViolations protocol violations\n";

  printResults() if $doPrint;
}


############################################
# printResults()
############################################

sub printResults {
    my $heard;
    my $synthesized;
    my $received;
    my $noInfo;
    my $dir;

    my $pct;
    my $sym;

    ########################################################
    # path information
    ########################################################

    if ( $wantDistinctPaths ) {
      Log ALWAYS, "\nPath data --   (downstream instances)  (upstream instances)\n";

      foreach my $pathData (sort {$b->[PATHSEQ_COUNT]->[DIR_UP]+$b->[PATHSEQ_COUNT]->[DIR_DOWN] <=>
				    $a->[PATHSEQ_COUNT]->[DIR_UP]+$a->[PATHSEQ_COUNT]->[DIR_DOWN]} @pathSequence) {

	# print the number of occurrences
	Log ALWAYS, "\nPath $pathData->[PATHSEQ_ID]:  ", commify($pathData->[PATHSEQ_COUNT][DIR_UP]+$pathData->[PATHSEQ_COUNT][DIR_DOWN]),
	             " instances  (", commify($pathData->[PATHSEQ_COUNT][DIR_DOWN]), ' / ',
                                      commify($pathData->[PATHSEQ_COUNT][DIR_UP]),
                                  ")\n";

	# print the path
	for ( my $j=0; $j<=$#{$pathData->[PATHSEQ_PATH]}; $j++ ) {
	  $sym = $pathData->[PATHSEQ_PATH]->[$j];
	  Log ALWAYS, sprintf("    %12s", $symNameMap[$sym]);
	  foreach $dir (DIR_DOWN, DIR_UP) {
	    $heard = $pathData->[PATHSEQ_HEARD_CNT]->[$dir][$j];
	    $synthesized = $pathData->[PATHSEQ_SYNTH_CNT]->[$dir][$j];
	    $received = $pathData->[PATHSEQ_RCVD_FLAG]->[$j] ? $heard+$synthesized : 0;

	    if ( $heard+$synthesized > 0 ) {
	      $pct = (int(1000.0*$heard/($heard+$synthesized)))/10;
	      Log ALWAYS, sprintf("     [$heard / $synthesized] %5.1f%", $pct);
	    } else {
	      Log ALWAYS, sprintf("     [0 / 0]      ", $pct);
	    }
	  }
	  Log ALWAYS, "\n";
	}
      }
    }

    ########################################################
    # symbol information
    ########################################################

    Log ALWAYS, "\n\n---- Summary ----\n\n";

    my $DIR_TOTAL = 2;  # hack!  1 larger than max(DIR_UP,DIR_DOWN), which MUST use values 0 and 1

    my $grandTotal = 0;
    foreach (@symCount) { $grandTotal += $_->[DIR_UP] + $_->[DIR_DOWN]; }

    my $printWidth = length commify($grandTotal);
    $printWidth = 9 if $printWidth < 9;
    #              SYM     Total             #Heard            #Syn           %Heard    #Recv'ed    %Rcv'ed
    my $format = "%13s  %$printWidth"."s  %$printWidth"."s  %$printWidth"."s  %7s  %$printWidth"."s  %7s  %$printWidth"."s  %7s\n";
    Log ALWAYS, sprintf($format, "Packet Type", "Total", "#Heard", "#Inferred", "%Heard", "#Rcv'ed", "%Rcv'ed", "NoInfo", "%NoInfo" );
    $format =~ s/\%7s/%6.1f%%/g;
    my $pctReceived;
    my $pctNoInfo;

    my @totalHeard;
    my @totalReceived;
    my @totalNoInfo;
    my @totalCount;
    for ($sym=0; $sym<=$#symCount; $sym++ ) {
      $symCount[$sym]->[$DIR_TOTAL] = $symCount[$sym]->[DIR_UP] + $symCount[$sym]->[DIR_DOWN];
      $symHeard[$sym]->[$DIR_TOTAL] = $symHeard[$sym]->[DIR_UP] + $symHeard[$sym]->[DIR_DOWN];
      foreach my $wasReceived (WASNT_RECEIVED, WAS_RECEIVED, NO_RECEIVED_INFO ) {
	$symReceived[$sym]->[$DIR_TOTAL]->[$wasReceived] =
	  $symReceived[$sym]->[DIR_UP]->[$wasReceived] + $symReceived[$sym]->[DIR_DOWN]->[$wasReceived];
      }

      foreach my $dir ($DIR_TOTAL, DIR_DOWN, DIR_UP) {
	$totalHeard[$dir] += $symHeard[$sym]->[$dir];
	$totalReceived[$dir] += $symReceived[$sym]->[$dir]->[WAS_RECEIVED];
	$totalNoInfo[$dir] += $symReceived[$sym]->[$dir]->[NO_RECEIVED_INFO];
	$totalCount[$dir] += $symCount[$sym]->[$dir];
      }
    }

    for ($sym=0; $sym<=$#symCount; $sym++ ) {
      next if $symCount[$sym]->[DIR_UP] + $symCount[$sym]->[DIR_DOWN] == 0;
      Log ALWAYS, "\n";
      my @label = ('Upstream', 'Downstream', substr($symNameMap[$sym].'           ',0,13));
      foreach my $dir ($DIR_TOTAL, DIR_DOWN, DIR_UP) {
	$heard = $symHeard[$sym]->[$dir];
	$synthesized = $symCount[$sym]->[$dir] - $heard;
	$received = $symReceived[$sym]->[$dir]->[WAS_RECEIVED];
	$noInfo  = $symReceived[$sym]->[$dir]->[NO_RECEIVED_INFO];

	if ( $heard+$synthesized > 0 ) {
	  $pct = (int(1000.0*$heard/($heard+$synthesized)))/10;
	  $pctReceived = ( int(1000.0*$received/($heard+$synthesized)) ) / 10;
	  $pctNoInfo = ( int(1000.0*$noInfo/($heard+$synthesized)) ) / 10;
	} else {
	  $pct = 0;
	  $pctReceived = 0;
	  $pctNoInfo = 0;
	}
	Log ALWAYS, sprintf($format, $label[$dir], commify($heard+$synthesized),
                          commify($heard), commify($synthesized), $pct,
			  commify($received), $pctReceived,
			  commify($noInfo), $pctNoInfo);
      }
    }

    # totals
    Log ALWAYS, "\n";

    my @label = ('Upstream', 'Downstream', 'Total');
    foreach my $dir ($DIR_TOTAL, DIR_DOWN, DIR_UP) {
      if ( $totalCount[$dir] != 0 ) {
	$pct = (int(1000*$totalHeard[$dir]/$totalCount[$dir]))/10;
	$pctReceived = (int(1000*$totalReceived[$dir]/$totalCount[$dir]))/10;
	$pctNoInfo = (int(1000*$totalNoInfo[$dir]/$totalCount[$dir]))/10;
      } else {
	$pct = "    ";
	$pctReceived = $pct;
	$pctNoInfo = $pct;
      }

      my $totalSyn = $totalCount[$dir] - $totalHeard[$dir];

      Log ALWAYS, sprintf($format, $label[$dir], commify($totalCount[$dir]), commify($totalHeard[$dir]),
			   commify($totalSyn), $pct,  commify($totalReceived[$dir]), $pctReceived,
			   commify($totalNoInfo[$dir]), $pctNoInfo);
    }

    ########################################################
    # protocol violation information
    ########################################################

    if ( $totalProtocolViolations > 0 ) {
	Log ALWAYS, "\n", commify($totalProtocolViolations), " Protocol Violations:\n";
	foreach my $pv (@protocolViolation) {
	    Log ALWAYS, "\t$pv->{summaryString}:\n";
	    my $total = 0;
	    foreach my $node (keys %{$pv->{byNode}} ) {
		Log ALWAYS, "\t\t", $clientData[$node]->[CLIENT_NAME], "  ", commify($pv->{byNode}->{$node}), "\n";
		$total += $pv->{byNode}->{$node};
	    }
	    Log ALWAYS, "\n";
	    foreach my $sym (keys %{$pv->{bySymbol}} ) {
		Log ALWAYS, "\t\t", sprintf( "%13s  %$printWidth"."s\n", $symNameMap[$sym], commify($pv->{bySymbol}->{$sym}));
	    }
	    Log ALWAYS, "\t$total Total\n";
	}
    }
}

##############################################################
# setDirection
#   packet is passed with src and dest fields as filled
#   in as possible.  DIR_FIELD may also have a guess
#   about the direction (DATA packets only).
#   bssid could be null.
#   if there is a bssid, there is also a src/dst.
#   if not, there can be a src/dst or just a dst.
#
# Must set DIR_FIELD, PATHDIR_FIELD, CLIENT_FIELD, and PARTNER_FIELD
# in $pkt
##############################################################

sub setDirection {
  my ($pkt) = @_;

  my $bssid = $pkt->[$BSSID_FIELD];
  my $dst = $pkt->[$A1_DST_FIELD];
  my $src = $pkt->[$A2_SRC_FIELD];

  # Watch out for broadcast address.  (Waddawedo if it's an AP transmission to the
  # broadcast address?)

  #################################################################
  # if there's no source (ACK or CTS), see if we can guess one.
  # (If the packet being ack/cts'ed was captured in the trace, code
  # in classifyPacket() will have supplied the source, so if there's no
  # source at this point there's no guessing based on previous packets)
  #################################################################

  unless( defined($src) ) {
    if ( $clientData[$dst]->[CLIENT_AP_TYPE] == AP_NONE ) {
      $src = $clientData[$dst]->[CLIENT_LAST_KNOWN_PARTNER] if $clientData[$dst]->[CLIENT_LAST_KNOWN_PARTNER] != -1;
    }
    unless ( defined($src) ) {
      Warn( $pkt->[$ID_FIELD], LPWNoSrc);
      $src = $UNKNOWN_NODE;
    }
    $pkt->[$A2_SRC_FIELD] = $src;
  }

  #################################################################
  # check if the src or dest has timed out as an AP
  #################################################################

  if ($clientData[$src]->[CLIENT_AP_TYPE] != AP_NONE && ($currentTime - $clientData[$src]->[CLIENT_LAST_USE_AS_AP] > $AP_TIMEOUT)) {
    $clientData[$src]->[CLIENT_AP_TYPE] = AP_NONE;
  }
  if ($clientData[$dst]->[CLIENT_AP_TYPE] != AP_NONE && ($currentTime - $clientData[$dst]->[CLIENT_LAST_USE_AS_AP] > $AP_TIMEOUT)) {
    $clientData[$dst]->[CLIENT_AP_TYPE] = AP_NONE;
  }


  #################################################################
  # register supplied bssid, if it isn't already
  #################################################################

  if ( defined($bssid) && $bssid != $BROADCAST_NODE && $bssid != $NOTASSOCIATED_NODE ) {

    if ( $clientData[$bssid]->[CLIENT_AP_TYPE] == AP_NONE ) {
      my $type = AP_INFRASTRUCTURE;
      my $digit = substr($clientData[$bssid]->[CLIENT_NAME],1,1);
      $type = AP_ADHOC if '2367abef' =~ m/$digit/i;
      $clientData[$bssid]->[CLIENT_AP_TYPE] = $type;
    }

    $clientData[$bssid]->[CLIENT_LAST_USE_AS_AP] = $currentTime;

    # we update last known bssid only if we're sure of the bssid --
    # i.e., it was supplied by classifyPacket()

    $clientData[$src]->[CLIENT_LAST_KNOWN_BSSID] = $bssid;
    $clientData[$dst]->[CLIENT_LAST_KNOWN_BSSID] = $bssid;
  }

  #################################################################
  # If we haven't been supplied with a bssid, try to figure one out
  #################################################################

  unless ( defined($bssid) ) {
    # $src is at least UNKNOWN_NODE at this point...
    # if one of src/dst is an infrastructure AP and the other isn't, pick it
    if ( $clientData[$src]->[CLIENT_AP_TYPE] > $clientData[$dst]->[CLIENT_AP_TYPE] ) {
      $bssid = $src;
    }
    elsif ( $clientData[$src]->[CLIENT_AP_TYPE] < $clientData[$dst]->[CLIENT_AP_TYPE] ) {
      $bssid = $dst;
    }
    # otherwise, pick the last known bssid, starting with the dst (which is reliable, vs. the src)
    elsif ( $clientData[$dst]->[CLIENT_LAST_KNOWN_BSSID] != -1 ) {
      $bssid = $clientData[$dst]->[CLIENT_LAST_KNOWN_BSSID];
    }
    elsif ( $clientData[$src]->[CLIENT_LAST_KNOWN_BSSID] != -1 ) {
      $bssid = $clientData[$src]->[CLIENT_LAST_KNOWN_BSSID];
    }

    $pkt->[$BSSID_FIELD] = $bssid;

    Warn( $pkt->[$ID_FIELD], LPWNoBSSID ) unless defined($bssid);
  }

  #################################################################
  # Now try to figure out client, partner, dir, and pathdir
  #################################################################

  if ($clientData[$bssid]->[CLIENT_AP_TYPE] == AP_INFRASTRUCTURE ) {
    if ( $src == $bssid && $dst != $BROADCAST_NODE ) {
      $pkt->[$DIR_FIELD] = DIR_DOWN;
      $pkt->[$PATHDIR_FIELD] = DIR_DOWN;
      $pkt->[$CLIENT_FIELD] = $dst;
      $pkt->[$PARTNER_FIELD] = $src;
    }
    else {
      $pkt->[$DIR_FIELD] = DIR_UP;
      $pkt->[$PATHDIR_FIELD] = DIR_UP;
      $pkt->[$CLIENT_FIELD] = $src;
      $pkt->[$PARTNER_FIELD] = $dst;
    }
  }

  elsif ( $clientData[$bssid]->[CLIENT_AP_TYPE] == AP_ADHOC ) {
    $pkt->[$DIR_FIELD] = DIR_ADHOC;
    if ( ($dst == $BROADCAST_NODE) || ($src < $dst && $src != $BROADCAST_NODE && $src != $NOTASSOCIATED_NODE) ) {
      # smaller numbered node is defined to be the client
      $pkt->[$CLIENT_FIELD] = $src;
      $pkt->[$PARTNER_FIELD] = $dst;
      $pkt->[$PATHDIR_FIELD] = DIR_UP;
    }
    else {
      $pkt->[$CLIENT_FIELD] = $dst;
      $pkt->[$PARTNER_FIELD] = $src;
      $pkt->[$PATHDIR_FIELD] = DIR_DOWN;
    }
  }

  else {
    Warn( $pkt->[$ID_FIELD], LPWUnrecogBSSID)
      unless $bssid == $BROADCAST_NODE ||
	     !defined($bssid);

    # data packets will have set the dir field, based on the DS bits
    if ( defined($pkt->[$DIR_FIELD]) && $pkt->[$DIR_FIELD] == DIR_UP) {
      $pkt->[$CLIENT_FIELD] = $src;
      $pkt->[$PARTNER_FIELD]= $dst;
      $pkt->[$PATHDIR_FIELD] = DIR_UP;
    }
    elsif ( defined($pkt->[$DIR_FIELD]) && $pkt->[$DIR_FIELD] == DIR_DOWN) {
      $pkt->[$CLIENT_FIELD] = $dst;
      $pkt->[$PARTNER_FIELD]= $src;
      $pkt->[$PATHDIR_FIELD] = DIR_DOWN;
    }
    # for everything else, we act as though it's an ad hoc packet
    else {
      $pkt->[$DIR_FIELD] = DIR_ADHOC;

      if ( $dst == $bssid ) {
	$pkt->[$CLIENT_FIELD] = $src;
	$pkt->[$PARTNER_FIELD]= $dst;
	$pkt->[$PATHDIR_FIELD] = DIR_UP;
      } 
      elsif ( $src == $bssid )  {
	$pkt->[$CLIENT_FIELD] = $dst;
	$pkt->[$PARTNER_FIELD]= $src;
	$pkt->[$PATHDIR_FIELD] = DIR_DOWN;
      }
      elsif ( $clientData[$dst]->[CLIENT_AP_TYPE] > $clientData[$src]->[CLIENT_AP_TYPE] ) {
	$pkt->[$CLIENT_FIELD] = $src;
	$pkt->[$PARTNER_FIELD]= $dst;
	$pkt->[$PATHDIR_FIELD] = DIR_UP;
      }
      elsif ( $clientData[$src]->[CLIENT_AP_TYPE] > $clientData[$dst]->[CLIENT_AP_TYPE] ) {
	$pkt->[$CLIENT_FIELD] = $dst;
	$pkt->[$PARTNER_FIELD]= $src;
	$pkt->[$PATHDIR_FIELD] = DIR_DOWN;
      }
      elsif ( $src < $dst ) {
	$pkt->[$CLIENT_FIELD] = $src;
	$pkt->[$PARTNER_FIELD]= $dst;
	$pkt->[$PATHDIR_FIELD] = DIR_UP;
      }
      else {
	$pkt->[$CLIENT_FIELD] = $dst;
	$pkt->[$PARTNER_FIELD]= $src;
	$pkt->[$PATHDIR_FIELD] = DIR_DOWN;
      }
    }
  }

  if ( $dst != $BROADCAST_NODE && $src != $UNKNOWN_NODE ) {
    $clientData[$src]->[CLIENT_LAST_KNOWN_PARTNER] = $dst;
    $clientData[$dst]->[CLIENT_LAST_KNOWN_PARTNER] = $src;
  }
}

##############################################################
# classifyPacket
#   Returns the client this packet is associated with.
#   Fills in _zBSSID, _zDir in the packet.
#   May also write the packet's src address.
##############################################################

sub classifyPacket {
    my ($pkt, $lastpacketByDest ) = @_;

    # translate packet type into a symbol of our artificial language.
    # (E.g., not all data packets are the same -- rexmit 

    $pkt->[$PARSESYMBOL_FIELD] = $SYM_IGNORED;


    #############################################
    ### DATA packets
    #############################################

    if ( $pkt->[$TYPE_FIELD] == TYPE_DATA ) {
      # DATA packets:
      #    seqno,  bssid, useful ds, src, dst
      # We ignore broadcast data packets for now (they aren't ack'ed)

      # set BSSID and direction
      if ( $pkt->[$DSTO_FIELD] == 0 ) {
	if ( $pkt->[$DSFROM_FIELD] == 0 ) {
	  # technically this is a sideways packet... bssid field should indicate nodes are in ad hoc mode
	  # Experiments show either A1 or A2 is almost always equal to A3, unless the dest is broadcast mac,
	  # so pick the other as the client

	  $pkt->[$BSSID_FIELD] = $pkt->[$A3_FIELD];
	  $pkt->[$DIR_FIELD] = DIR_ADHOC;

	  unless ( $clientData[$pkt->[$A3_FIELD]]->[CLIENT_AP_TYPE] == AP_ADHOC ) {
	    Warn( $pkt->[$ID_FIELD],
	                $clientData[$pkt->[$A3_FIELD]]->[CLIENT_AP_TYPE] == AP_INFRASTRUCTURE ?
		               LPWAdHocInfrasBSS : LPWAdHocUnrecogBSS );
	  }
	} else {
	  $pkt->[$BSSID_FIELD] = $pkt->[$A2_SRC_FIELD];
	  $pkt->[$DIR_FIELD] = DIR_DOWN;
	}
      } else {
	if ( $pkt->[$DSFROM_FIELD] == 0 ) {
	  $pkt->[$BSSID_FIELD] = $pkt->[$A1_DST_FIELD];
	  $pkt->[$DIR_FIELD] = DIR_UP;
	} else {
	  Warn( $pkt->[$ID_FIELD], LPWDS11 );
	}
      }

      # we see only subtype Data and Null function (no data) among data packets (so we don't
      # worry here about the CFxxx varieties!

      # The kind of data packet this is depends on the morefrag and retry bits and the fragno field
      my $typeName;
      if ( $pkt->[$FRAGNO_FIELD] <= $max_fragno ) {
	$typeName = $pkt->[$MOREFRAG_FIELD] ? 'FRAG_' : 'DATA_';
	$typeName .= $pkt->[$FRAGNO_FIELD];
      } else {
	$typeName = 'DATA_MAX';
      }
      $typeName .= 'R' if $pkt->[$RETRY_FIELD] != 0;
      $pkt->[$PARSESYMBOL_FIELD] = $SYM{$typeName};
      die "Constructed undefined data packet symbol $typeName" if !defined($pkt->[$PARSESYMBOL_FIELD]);

    }

    #############################################
    ### CONTROL packets
    #############################################

    elsif ( $pkt->[$TYPE_FIELD] == TYPE_CONTROL ) {

      # ACK (control) packets:
      #    no seqno, no bssid, no useful ds, no src, have dst
      # CTS (control) packets:
      #    no seqno, no bssid, no useful ds, no src, have dst

      if ( ($pkt->[$SUBTYPE_FIELD] == $SUBTYPE_ACK) or ($pkt->[$SUBTYPE_FIELD] == $SUBTYPE_CTS) ) {

	# set packet type
	if ( $pkt->[$SUBTYPE_FIELD] == $SUBTYPE_ACK ) {
	  $pkt->[$PARSESYMBOL_FIELD] = $SYM_ACK;
	} else {
	  $pkt->[$PARSESYMBOL_FIELD] = $SYM_CTS;
	}

	# We have to rely on the last packet sent to tell us who is doing the ack'ing.
	my $timeDelta = ($pkt->[$TIME_FIELD] - $lastpacketByDest->[$TIME_FIELD]) * 1000000.0;
	my $bound = 2*$SIFS + 192 + $pkt->[$SIZE_FIELD]*8.0/$pkt->[$RATE_FIELD];
	if (  ($timeDelta <= $bound) and ($pkt->[$A1_DST_FIELD] == $lastpacketByDest->[$A2_SRC_FIELD]) ) {

	  # set this packet's source
	  $pkt->[$A2_SRC_FIELD] = $lastpacketByDest->[$A1_DST_FIELD];

	  # and its BSSID
	  $pkt->[$BSSID_FIELD] = $lastpacketByDest->[$BSSID_FIELD];

	  # direction of this packet is opposite that of the previous one
#	  $pkt->[$DIR_FIELD] = ($lastpacketByDest->[$DIR_FIELD] == DIR_UP) ? DIR_DOWN : DIR_UP;
	}
      }


      # RTS (control) packets: 
      #    no seqno, no bssid, no useful ds, have src and dst

      elsif ( $pkt->[$SUBTYPE_FIELD] == $SUBTYPE_RTS ) {
	$pkt->[$PARSESYMBOL_FIELD] = $SYM_RTS;
      }

      # POWER SAVE POLL (control) packets:
      #    no seqno, has bssid, no useful ds, has src, has dst

      elsif ( $pkt->[$SUBTYPE_FIELD] == $SUBTYPE_PS_POLL ) {
	$pkt->[$BSSID_FIELD] = $pkt->[$A1_DST_FIELD];
#	$pkt->[$DIR_FIELD] = DIR_UP;
	$pkt->[$PARSESYMBOL_FIELD] = $SYM_PS_POLL;
      }

      # CF-END and CF-END-ACk
      # We don't really parse using these.  Just set their BSSID and DIR fields.

      elsif ( $pkt->[$SUBTYPE_FIELD] == $SUBTYPE_CFEND_ACK ) {
	$pkt->[$BSSID_FIELD] = $pkt->[$A2_SRC_FIELD];
#	$pkt->[$DIR_FIELD] = DIR_DOWN;
      }

      elsif ( $pkt->[$SUBTYPE_FIELD] == $SUBTYPE_CFEND ) {
	$pkt->[$BSSID_FIELD] = $pkt->[$A2_SRC_FIELD];
#	$pkt->[$DIR_FIELD] = DIR_DOWN;
      }
    }

    #############################################
    ### MANAGEMENT packets
    #############################################

    else {

      # BSSID is always in A3 field of mgt frames, but may be broadcast addr
      $pkt->[$BSSID_FIELD] = $pkt->[$A3_FIELD];

      # PROBE REQUEST (mgt) packets:
      #    seqno,  bssid, no useful ds, src, dst (may be broadcast)

      if ( $pkt->[$SUBTYPE_FIELD] == $SUBTYPE_PROBE_REQ ) {
#	$pkt->[$DIR_FIELD] = DIR_UP;
	$pkt->[$PARSESYMBOL_FIELD] = $SYM_PROBE_REQ;
      }

      # PROBE RESPONSE (mgt) packets:
      #    seqno,  bssid, no useful ds, src, dst
      elsif ( $pkt->[$SUBTYPE_FIELD] == $SUBTYPE_PROBE_RESP ) {
#	$pkt->[$DIR_FIELD] = DIR_DOWN;
	$pkt->[$PARSESYMBOL_FIELD] = $pkt->[$RETRY_FIELD] ?  $SYM_PROBE_RESPR :  $SYM_PROBE_RESP;
      }

      # AUTH:
      #    seqno, bssid, no useful ds, src, dst
      elsif ($pkt->[$SUBTYPE_FIELD] == $SUBTYPE_AUTH ) {
	$pkt->[$PARSESYMBOL_FIELD] = $pkt->[$RETRY_FIELD] ? $SYM_AUTHR : $SYM_AUTH;
      }

      # ASSOCIATION REQUEST
      elsif ($pkt->[$SUBTYPE_FIELD] == $SUBTYPE_ASSOC_REQ ) {
#	$pkt->[$DIR_FIELD] = DIR_UP;
	$pkt->[$PARSESYMBOL_FIELD] = $pkt->[$RETRY_FIELD] ? $SYM_ASSOC_REQR : $SYM_ASSOC_REQ;
      }

      # ASSOCIATION RESPONSE
      elsif ($pkt->[$SUBTYPE_FIELD] == $SUBTYPE_ASSOC_RESP ) {
#	$pkt->[$DIR_FIELD] = DIR_DOWN;
	$pkt->[$PARSESYMBOL_FIELD] = $pkt->[$RETRY_FIELD] ? $SYM_ASSOC_RESPR : $SYM_ASSOC_RESP;
      }

      # BEACON
      elsif ($pkt->[$SUBTYPE_FIELD] == $SUBTYPE_BEACON ) {
#	$pkt->[$DIR_FIELD] = DIR_DOWN;
	$pkt->[$PARSESYMBOL_FIELD] = $SYM_BEACON;
      }

      # REASSOCIATION REQUEST
      elsif ($pkt->[$SUBTYPE_FIELD] == $SUBTYPE_REASSOC_REQ ) {
#	$pkt->[$DIR_FIELD] = DIR_UP;
	$pkt->[$PARSESYMBOL_FIELD] = $pkt->[$RETRY_FIELD] ? $SYM_REASSOC_REQR : $SYM_REASSOC_REQ;
      }

      # REASSOCIATION RESPONSE
      elsif ($pkt->[$SUBTYPE_FIELD] == $SUBTYPE_REASSOC_RESP ) {
#	$pkt->[$DIR_FIELD] = DIR_DOWN;
	$pkt->[$PARSESYMBOL_FIELD] = $pkt->[$RETRY_FIELD] ? $SYM_REASSOC_RESPR : $SYM_REASSOC_RESP;
      }

      # DISASSOCIATION
      elsif ($pkt->[$SUBTYPE_FIELD] == $SUBTYPE_DISASSOC ) {
#	if ( $pkt->[$A2_SRC_FIELD] == $pkt->[$A3_FIELD] ) {
#	  $pkt->[$DIR_FIELD] = DIR_DOWN;
#	} else {
#	  $pkt->[$DIR_FIELD] = DIR_UP;
#	}
	$pkt->[$PARSESYMBOL_FIELD] = $pkt->[$RETRY_FIELD] ? $SYM_DISASSOCR : $SYM_DISASSOC;
      }

      # DEAUTHENTICATION
      elsif ($pkt->[$SUBTYPE_FIELD] == $SUBTYPE_DEAUTH ) {
	$pkt->[$PARSESYMBOL_FIELD] = $pkt->[$RETRY_FIELD] ? $SYM_DEAUTHR : $SYM_DEAUTH;
      }
    }

    setDirection( $pkt );

    # figure out who the client is, if this was recognized packet
    #if ( $pkt->[$DIR_FIELD]==DIR_UP ) {
    #  $pkt->[$CLIENT_FIELD] = $pkt->[$A2_SRC_FIELD];
    #  $pkt->[$PARTNER_FIELD] = $pkt->[$A1_DST_FIELD];
    #}
    #elsif ( $pkt->[$DIR_FIELD]==DIR_DOWN ) {
    #  $pkt->[$CLIENT_FIELD] = $pkt->[$A1_DST_FIELD];
    #  $pkt->[$PARTNER_FIELD] = $pkt->[$A2_SRC_FIELD];
    #}

    Warn( $pkt->[$ID_FIELD], LPWNoDir )
      unless defined($pkt->[$DIR_FIELD]);

    Warn( $pkt->[$ID_FIELD], LPWNoPathDir )
      unless defined($pkt->[$PATHDIR_FIELD]);

    Warn( $pkt->[$ID_FIELD], LPWBCastClient )
      if $pkt->[$CLIENT_FIELD] == $BROADCAST_NODE;

    Warn( $pkt->[$ID_FIELD], LPWAPasClient )
      if $clientData[$pkt->[$CLIENT_FIELD]]->[CLIENT_AP_TYPE] &&
	 $pkt->[$PARTNER_FIELD] != $BROADCAST_NODE;

    Warn( $pkt->[$ID_FIELD], LPWNonAPasPartner )
      unless !$wantAPWarnings
	|| $clientData[$pkt->[$PARTNER_FIELD]]->[CLIENT_AP_TYPE]
	|| $pkt->[$PARTNER_FIELD] == $BROADCAST_NODE
        || $pkt->[$DIR_FIELD] == DIR_ADHOC
	|| $pkt->[$PARTNER_FIELD] == $UNKNOWN_NODE;

    return $pkt->[$CLIENT_FIELD];
}

############################################
# timedOut()
############################################

sub timedOut {
  my ($client) = @_;

  return 1 if defined($clientData[$client]->[CLIENT_STATESET]) &&
              ($currentTime - $clientData[$client]->[CLIENT_CONV_LAST_ACTIVITY_TIME] > $MARKER_TIMEOUTVALUE);
  return 0;
}

############################################
# generateMarker()
############################################

sub generateMarker {
    my ($pkt) = @_;

    my $client = $pkt->[$CLIENT_FIELD];
    my $partner = $pkt->[$PARTNER_FIELD];
    my $marker = $MARKER_NONE;

    # in some circumstances we can convince ourselves this packet is part of the
    # current conversation.  In those (that) case, skip the tests (esp. avoiding
    # the timeout test.

    if ( defined($pkt->[$SEQNO_FIELD]) &&
	 defined($clientData[$client]->[CLIENT_CONV_LAST_SEQNO]->[$pkt->[$DIR_FIELD]]) &&
	 $pkt->[$SEQNO_FIELD] == $clientData[$client]->[CLIENT_CONV_LAST_SEQNO]->[$pkt->[$DIR_FIELD]]
       ) {
      # nothing -- just skip the remaining tests
    }
    else {

      # We're checking if the current packet indicates the existing conversation for
      # the client has ended.  There are three kinds of checks:
      #   - conversation has timed out
      #   - seqno based checks
      #   - partner checks
      # (Additionally, the FSA may find that the type of the packet indicates the current
      #  conversation must have ended, or that the packet's direction is inappropriate,
      #  but that doesn't happen here.)

      # timeout
      if ( timedOut($client)) {
	if ( $debug >= DEBUG_FULL ) {
	  mirrorPrint( "generateMarker: timeout  $pkt->[$TIME_FIELD] vs. ",
		       "$clientData[$client]->[CLIENT_CONV_LAST_ACTIVITY_TIME]\n");
	}
	$marker = $MARKER_SYM;
      }

      # ignored packet type check
      elsif ( $pkt->[$PARSESYMBOL_FIELD] == $SYM_IGNORED ) {
	if ( $debug >= DEBUG_FULL ) {
	  mirrorPrint("generateMarker: IGNORED packet\n");
	}
	$marker = $MARKER_SYM;
      }

      # Both the partner and seqno checks depend on knowing about the PROBE_REQ, if the
      # current packet might be responding to it.  But, the PROBE_REQ might not have
      # been heard by the monitors....
      if ( ($marker == $MARKER_NONE) &&
	   ($pkt->[$SUBTYPE_FIELD] == $SUBTYPE_PROBE_RESP) &&
	   !defined($clientData[$client]->[CLIENT_PROBEREQ_DATA]) )  {
	# return marker sym.  Caller will terminate any ongoing conversation and then
	# create probereq_data.  (Can't do it here -- it'll be destroyed by terminate conversation.)
	# cons up the PROBE_REQ data.  Assume it was to a broadcast address...
	#$clientData[$client]->[CLIENT_PROBEREQ_DATA] = [ $BROADCAST_NODE, undef ];
	$marker = $MARKER_SYM;
      }


      # partner check.
      # (We exclude conversations starting with broadcast PROBE_REQs from this check because
      # they may have multiple responders.)
      if ( ($marker == $MARKER_NONE) &&
	   ( !defined($clientData[$client]->[CLIENT_PROBEREQ_DATA]) ||
	     $clientData[$client]->[CLIENT_PROBEREQ_DATA]->[PROBEREQD_DEST] != $BROADCAST_NODE )
	 ) {

	if ( $clientData[$client]->[CLIENT_CURRENT_CONV_PARTNER] == -1 ) {
	  $clientData[$client]->[CLIENT_CURRENT_CONV_PARTNER] = $partner;
	} elsif ( ($clientData[$client]->[CLIENT_CURRENT_CONV_PARTNER] != $partner) ) {
	  mirrorPrint("generateMarker: new partner on packet $pkt->[$ID_FIELD]: $partner vs. ",
		      "$clientData[$client]->[CLIENT_CURRENT_CONV_PARTNER]\n")
	    if $debug >= DEBUG_FULL;
	  $marker = $MARKER_SYM;
	}
      }

      # general sequence number check. PROBE_REQs have their own elaborate
      # data,which PROBE_RESPs use.

      if ( $marker == $MARKER_NONE && defined($pkt->[$SEQNO_FIELD]) ) {
	if ( defined($clientData[$client]->[CLIENT_PROBEREQ_DATA]) ) {  # is the client in a PROBE_REQ conversation?
	  if ( exists($clientData[$client]->[CLIENT_PROBEREQ_DATA]->[PROBEREQD_SEQHASH]->{$partner}) ) {
	    if ( $clientData[$client]->[CLIENT_PROBEREQ_DATA]->[PROBEREQD_SEQHASH]->{$partner} != $pkt->[$SEQNO_FIELD] ) {
	      if ( $debug >= DEBUG_FULL ) {
		mirrorPrint("generateMarker: new seqno on PROBE_RESP from $partner at packet $pkt->[$ID_FIELD]\n");
	      }
	      $marker = $MARKER_SYM;
	    }
	  } else {
	    $clientData[$client]->[CLIENT_PROBEREQ_DATA]->[PROBEREQD_SEQHASH]->{$partner} = $pkt->[$SEQNO_FIELD];
	  }
	}
	else {
	  if ( defined($clientData[$client]->[CLIENT_CONV_LAST_SEQNO]->[$pkt->[$DIR_FIELD]]) ) {
	    if ( $clientData[$client]->[CLIENT_CONV_LAST_SEQNO]->[$pkt->[$DIR_FIELD]] != $pkt->[$SEQNO_FIELD] ) {
	      if ( $debug >= DEBUG_FULL ) {
		mirrorPrint( "generateMarker: new seqno on packet: $pkt->[$SEQNO_FIELD] vs. ",
			     $clientData[$client]->[CLIENT_CONV_LAST_SEQNO]->[$pkt->[$DIR_FIELD]], "\n");
	      }
	      $marker = $MARKER_SYM;
	    }
	    # this check looks for out of order packets, which may occur through protocol or merging bugs
	    elsif ($pkt->[$RETRY_FIELD]==0 && 
		   $pkt->[$FRAGNO_FIELD]==$clientData[$client]->[CLIENT_CONV_LAST_FRAGNO] &&
		   $pkt->[$SUBTYPE_FIELD]!=$SUBTYPE_PROBE_REQ ) {
	      # many clients seem to reuse probe_req sequence numbers  (do APs ignore the dups with retry == 0?)
	      flagProtocolViolation( 0, $client, $pkt );
	      if ( $debug >= DEBUG_FULL ) {
		mirrorPrint("generateMarker: protocol violation\n");
	      }
	      $marker = $MARKER_SYM;
	    }
	  }
	  else {
	    $clientData[$client]->[CLIENT_CONV_LAST_SEQNO]->[$pkt->[$DIR_FIELD]] = $pkt->[$SEQNO_FIELD];
	    $clientData[$client]->[CLIENT_CONV_LAST_FRAGNO] = $pkt->[$FRAGNO_FIELD];
	  }
	}
      }

      # specialized RTS check.  Look for a cDur field that differs from an earlier one in this conversation
      # (which would imply a data packet of a new length)

      if ( $USE_RTS_CHECK && ($marker == $MARKER_NONE) && ($pkt->[$SUBTYPE_FIELD] == $SUBTYPE_RTS) ) {
	# casting rate to an int causes no ambiguities (rates are at least 1Mbps apart)
	my $intRate = int($pkt->[$RATE_FIELD]);
	if ( defined($clientData[$client]->[CLIENT_RTS_DURVEC]->[$intRate]) ) {
	  $marker = $MARKER_SYM if $pkt->[$DUR_FIELD] != $clientData[$client]->[CLIENT_RTS_DURVEC]->[$intRate];
	} else {
	  $clientData[$client]->[CLIENT_RTS_DURVEC]->[$intRate] = $pkt->[$DUR_FIELD];
	}
      }
    }

    # all done

    $clientData[$client]->[CLIENT_CONV_LAST_ACTIVITY_TIME] = $pkt->[$TIME_FIELD];
    #Log DEBUG_FULL, "generateMarker done: $client $marker $clientData[$client]->[CLIENT_CONV_LAST_ACTIVITY_TIME]\n";
    return $marker;
}

############################################
# terminateConversation()
############################################

sub terminateConversation {
    my ($client, $pktID ) = @_;

    # all states should be the accept state

    my $stateVec = $clientData[$client]->[CLIENT_STATESET];
    my $pathDir = $stateVec->[0]->[STATE_CONV_PARITY];

    # The stateVec used to contain a STATE_PATHLIST that was a ref to an array
    # of edgeref's.  As an attempt to improve running time, this was converted
    # to [ ref-to-preceding-state's-STATE_PATHLIST, edge] -- i.e., the full
    # path is found by chaining back.

    # Here's the old use...
    #my $edgeList = $stateVec->[0]->[STATE_PATHLIST];

    # Here's the new use.  (A ref just to minimize code changes.)
    my $edgeList = [];
    for ( my $pathSeg = $stateVec->[0]->[STATE_PATHLIST]; $pathSeg->[1]; $pathSeg = $pathSeg->[0] ) {
      push @{$edgeList}, ($pathSeg->[1]);
    }
    @{$edgeList} = reverse @{$edgeList};

    #mirrorPrint("\npktID = $pktID: [" . $#{$edgeList} . "]\n\t(" . join(', ',@{$edgeList}) . ")\n");

    if ( $debug > 0 ) {
      my $dieMsg;
      $dieMsg = "isn't in any state"      if (!@{$stateVec}) || ($#{$stateVec} < 0);
      $dieMsg = 'in multiple states'      if $#{$stateVec} > 0;
      $dieMsg = "isn't in accept state"   if $stateVec->[0]->[STATE_ID] != $ACCEPT_STATE;
      Log ALWAYS, "INTERNAL ERROR: packet ID = $pktID\n\tClient $client " . $dieMsg .
	' at close of conversation: [' . stateIDList($client) . "]\n"
	  if $dieMsg;
    }

    #################################################################
    # now prepare to tally some statistics
    #################################################################

    if ( $debug >= DEBUG_WARN ) {
      mirrorPrint("Terminating conversation for $client:  Cost = $stateVec->[0]->[STATE_PATHCOST]\n");
    }
    my $pathElement = $pathMap;
    my $sym;
    my @heardSym = ();
    my $edgeDir;
    my $edge;
    my $index;
    my $pkt;
    my $lastPkt;
    my $usefulPkt;
    my @retryCnt = (0,0);
    my @seqno = (undef,undef);

    for ($index=0; $index<=$#{$edgeList}; $index++ ) {
     $edge = $edgeList->[$index];
     $lastPkt = $pkt;  # remember it in case we have to gen synth packets
     $pkt = $clientData[$client]->[CLIENT_PACKET_LIST]->[$index];

      if ( !$edge->[edge_path] ) {
	$sym = $edge->[edge_symbol];
	next if $sym == $MARKER_SYM;

	$sym = $pkt->[$PARSESYMBOL_FIELD];  # take the packet's symbol to avoid WILDCARD that may be on the edge

	$edgeDir = ($pathDir + $edge->[edge_parity]) % 2;

	$seqno[$edgeDir] = $pkt->[$SEQNO_FIELD] if defined($pkt->[$SEQNO_FIELD]);

	$retryCnt[$edgeDir] = $isRetrySym[$sym] ? $retryCnt[$edgeDir]+1 : 0;
	$pkt->[$RETRY_FIELD] = $retryCnt[$edgeDir] if $pkt->[$RETRY_FIELD];

	$symCount[$sym]->[$edgeDir]++;
	$symHeard[$sym]->[$edgeDir]++;
	$pkt->[$WASRCVD_FIELD] = $edge->[edge_wasReceived];
	$symReceived[$sym]->[$edgeDir]->[$edge->[edge_wasReceived]]++;

	if ( $wantDistinctPaths ) {
	  $pathElement->{$sym} = {} if !defined $pathElement->{$sym};
	  $pathElement = $pathElement->{$sym};
	  push @heardSym, (1);
	  if ( $debug >= DEBUG_WARN ) {
	    mirrorPrint( sprintf("    %3d -> %3d  %8s", $edge->[edge_fromState], $edge->[edge_nextState],  $symNameMap[$sym]) );
	    mirrorPrint(" (H)") if $edge->[edge_wasReceived] == WAS_RECEIVED;
	    mirrorPrint("\n");
	  }
	}
      }
      else {
	$lastPkt->[$SYN_FIELD] += SYNTH_AFTER if defined($lastPkt);
	$pkt->[$SYN_FIELD]     += SYNTH_BEFORE if defined($pkt);

	$usefulPkt = $pkt ? $pkt : $lastPkt;

	$edgeDir = ($pathDir + $edge->[edge_path]->[-1]->[edge_parity]) % 2;
	$seqno[$edgeDir] = $pkt->[$SEQNO_FIELD] if defined($pkt->[$SEQNO_FIELD]);

	foreach (@{$edge->[edge_path]}) {
	  # wildcard edges can occur only if there is a matching packet
	  $sym = $_->[edge_symbol];
	  next if $sym == $MARKER_SYM;

	  if ( $debug >= DEBUG_WARN ) {
	    mirrorPrint( sprintf("    %3d -> %3d  %8s", $_->[edge_fromState], $_->[edge_nextState],  $symNameMap[$_->[edge_symbol]]) );
	    mirrorPrint(" (H)") if $_->[edge_wasReceived] == WAS_RECEIVED;
	  }

	  $edgeDir = ($pathDir + $_->[edge_parity]) % 2;

	  $retryCnt[$edgeDir] = $isRetrySym[$sym] ? $retryCnt[$edgeDir]+1 : 0;

	  $symCount[$sym]->[$edgeDir]++;
	  $symReceived[$sym]->[$edgeDir]->[$_->[edge_wasReceived]]++;

	  if ( $_ != $edge->[edge_path]->[-1] ) {

	    # push packet to synth table
	    my $synPkt = [];
	    $synPkt->[$SYNPARSESYMBOL_FIELD] = $sym;
	    $synPkt->[$SYNSUBTYPE_FIELD] = $symSubtypeMap[$sym];
	    $synPkt->[$SYNCONVID_FIELD] = $usefulPkt->[$CONVID_FIELD];
	    $synPkt->[$SYNDIR_FIELD] = $edgeDir;
	    $synPkt->[$SYNWASRCVD_FIELD] = $_->[edge_wasReceived];
	    $synPkt->[$SYNPRIORID_FIELD] = $lastPkt->[$ID_FIELD];
	    $synPkt->[$SYNPRIORTIME_FIELD] = $lastPkt->[$TIME_FIELD];
	    $synPkt->[$SYNLATERID_FIELD] = $pkt->[$ID_FIELD];
	    $synPkt->[$SYNLATERTIME_FIELD] = $pkt->[$TIME_FIELD];
	    $synPkt->[$SYNRETRY_FIELD] = $retryCnt[$edgeDir];
	    $synPkt->[$SYNID_FIELD] = $synInsertNextId++;
	    $synPkt->[$SYNSEQNO_FIELD]  = $seqno[$edgeDir];

	    if ($edgeDir == DIR_UP ) {
	      $synPkt->[$SYNA2_SRC_FIELD] = $usefulPkt->[$CLIENT_FIELD];
	      $synPkt->[$SYNA1_DST_FIELD] = $usefulPkt->[$PARTNER_FIELD];
	    } else {
	      $synPkt->[$SYNA2_SRC_FIELD] = $usefulPkt->[$PARTNER_FIELD];
	      $synPkt->[$SYNA1_DST_FIELD] = $usefulPkt->[$CLIENT_FIELD];
	    }

	    $synInsertQuery .= '(' . join(',',map{defined($_) ? $_ : 'null'}@{$synPkt}) . '),';
	    $synInsertQueryCount++;
	    pushOutput() if $synInsertQueryCount > $WRITE_GROUPSIZE;

	    if ( $wantDistinctPaths ) {
	      if ( $debug >= DEBUG_WARN ) {
		mirrorPrint("  !\n");
	      }
	      push @heardSym, (0);
	    }
	  } else {
	    if ( $wantDistinctPaths ) {
	      if ( $debug >= DEBUG_WARN ) { mirrorPrint("\n"); }
	      push @heardSym, (1);
	    }
	    $symHeard[$sym]->[$edgeDir]++;
	    $pkt->[$WASRCVD_FIELD] = $_->[edge_wasReceived];
	    $pkt->[$RETRY_FIELD] = $retryCnt[$edgeDir] if $pkt->[$RETRY_FIELD] && $retryCnt[$edgeDir];  # make sure never to turn it off or on
	  }

	  if ( $wantDistinctPaths ) {
	    $pathElement->{$sym} = {} if !defined $pathElement->{$sym};
	    $pathElement = $pathElement->{$sym};
	  }
	}
      }
    }

    my $id;
    if ( $wantDistinctPaths ) {
      $id = $pathElement->{id};
      if (!defined($id) ) {
	$id = $NEXT_PATHSEQUENCE_ID++ ;
	$pathElement->{id} = $id;

	$pathSequence[$id]->[PATHSEQ_ID] = $id;
	$pathSequence[$id]->[PATHSEQ_PATH] = [];
	$pathSequence[$id]->[PATHSEQ_COUNT] = [0,0];
	$pathSequence[$id]->[PATHSEQ_HEARD_CNT] = [[],[]];
	$pathSequence[$id]->[PATHSEQ_SYNTH_CNT] = [[],[]];

	#foreach my $edge (@{$edgeList}) {
	for ($index=0; $index<=$#{$edgeList}; $index++ ) {
	  $edge = $edgeList->[$index];
	  if ( !$edge->[edge_path] ) {
	    next if $edge->[edge_symbol] == $MARKER_SYM;
	    $pkt = $clientData[$client]->[CLIENT_PACKET_LIST]->[$index];
	    push @{$pathSequence[$id]->[PATHSEQ_PATH]}, ($pkt->[$PARSESYMBOL_FIELD]);
	    push @{$pathSequence[$id]->[PATHSEQ_RCVD_FLAG]}, ($edge->[edge_wasReceived]==2?1:0);
	    push @{$pathSequence[$id]->[PATHSEQ_HEARD_CNT]->[DIR_UP]}, (0);
	    push @{$pathSequence[$id]->[PATHSEQ_HEARD_CNT]->[DIR_DOWN]}, (0);
	    push @{$pathSequence[$id]->[PATHSEQ_SYNTH_CNT]->[DIR_UP]}, (0);
	    push @{$pathSequence[$id]->[PATHSEQ_SYNTH_CNT]->[DIR_DOWN]}, (0);
	  } else {
	    foreach (@{$edge->[edge_path]}) {
	      next if $_->[edge_symbol] == $MARKER_SYM;
	      push @{$pathSequence[$id]->[PATHSEQ_PATH]}, ($_->[edge_symbol]);
	      push @{$pathSequence[$id]->[PATHSEQ_RCVD_FLAG]}, ($_->[edge_wasReceived]==2?1:0);
	      push @{$pathSequence[$id]->[PATHSEQ_HEARD_CNT]->[DIR_UP]}, (0);
	      push @{$pathSequence[$id]->[PATHSEQ_HEARD_CNT]->[DIR_DOWN]}, (0);
	      push @{$pathSequence[$id]->[PATHSEQ_SYNTH_CNT]->[DIR_UP]}, (0);
	      push @{$pathSequence[$id]->[PATHSEQ_SYNTH_CNT]->[DIR_DOWN]}, (0);
	    }
	  }
	}
      }

      # at last... tally the statisitics

      my $pathData = $pathSequence[$id];
      $pathData->[PATHSEQ_COUNT]->[$pathDir]++;  # tag the entire path's direction using the first packet sent

      for ( my $i=0; $i<=$#heardSym; $i++ ) {
	if ( $heardSym[$i] == 0 ) {
	  $pathData->[PATHSEQ_SYNTH_CNT]->[$pathDir][$i]++;
	} else {
	  $pathData->[PATHSEQ_HEARD_CNT]->[$pathDir][$i]++;
	}
      }
    }

    #################################################################
    # reset some client state
    #################################################################

    undef $clientData[$client]->[CLIENT_STATESET];
    foreach (@{$clientData[$client]->[CLIENT_PACKET_LIST]}) {
      next if !defined($_);
      $_->[$PENDING_FIELD] = 0;
      $_->[$PATHID_FIELD] = $id;  # $id is undef is !$wantDistinctPaths
    }
    undef $clientData[$client]->[CLIENT_PACKET_LIST];
    undef $clientData[$client]->[CLIENT_PROBEREQ_DATA];
    $clientData[$client]->[CLIENT_CURRENT_CONV_PARTNER] = -1;
    undef $clientData[$client]->[CLIENT_CONV_LAST_SEQNO];
    undef $clientData[$client]->[CLIENT_RTS_DURVEC];
}

############################################
# swapCachedStates()
############################################

sub swapCachedStates {
  my ($pktNum) = @_;

  mirrorPrint( "Swapping old: " . formatCacheStats() ) ;
  $oldCachedStateSet = $currentCachedStateSet;
  $numOldCachedStates = $numCurrentCachedStates;
  undef $currentCachedStateSet;
  $currentCachedStateSet->{undef} = $oldCachedStateSet->{undef};
  $currentCachedStateSet->{wildcard} = $oldCachedStateSet->{wildcard};
  $numCurrentCachedStates = 2;
  $cssGeneration++;

  mirrorPrint( "New: " . formatCacheStats() );
}


############################################
# takeFSAStep()
############################################

sub takeFSAStep {
    my ($client, $sym, $pkt) = @_;

    # these vars now declared globally, to avoid memory allocation overhead(????)
    #my @candidateStateSet = ();
    #my @newStatePathCost = ();
    @candidateStateSet = ();
    @newStatePathCost = ();

    my $dir = DIR_UP;  # if we haven't been passed a pkt, we don't care about direction????
    $dir = $pkt->[$DIR_FIELD] if defined($pkt);

    my $stateVec;
    my $newStateVec;

    my $edgeList;
    my $edge;
    my $nextState;
    my $newCost;

    # if the client has no current state set, this packet is starting a conversation...
    if ( !defined($clientData[$client]->[CLIENT_STATESET]) ) {
      # the initalization vector entries must match STATE_xxx values
      $clientData[$client]->[CLIENT_STATESET] = $currentCachedStateSet->{undef};
      $clientData[$client]->[CLIENT_CURRENT_CONV] = $CONVERSATION_ID++;
      # we keep special data for PROBE_REQ
      if ( defined($pkt) ) {
	if ( $pkt->[$SUBTYPE_FIELD] == $SUBTYPE_PROBE_REQ ) {
	  $clientData[$client]->[CLIENT_PROBEREQ_DATA] = [ $pkt->[$PARTNER_FIELD], undef ];
	}
	elsif ( $pkt->[$SUBTYPE_FIELD] == $SUBTYPE_PROBE_RESP &&
		!defined($clientData[$client]->[CLIENT_PROBEREQ_DATA])
	      ){
	  $clientData[$client]->[CLIENT_PROBEREQ_DATA] = [$BROADCAST_NODE, undef];
	  $clientData[$client]->[CLIENT_PROBEREQ_DATA]->[PROBEREQD_SEQHASH]->{$pkt->[$PARTNER_FIELD]} = $pkt->[$SEQNO_FIELD];
	}
      }
    }

    # try to find a cached state
    my $clientStateVec = $currentCachedStateSet->{($clientData[$client]->[CLIENT_STATESET], $sym, $dir)};
    if ( $clientStateVec ) {
      $cssNewHit++;
    }
    else {
      $clientStateVec = $oldCachedStateSet->{($clientData[$client]->[CLIENT_STATESET], $sym, $dir)};
      if ( $clientStateVec ) {
	$cssOldHit++;
	$currentCachedStateSet->{($clientData[$client]->[CLIENT_STATESET], $sym, $dir)} = $clientStateVec;
	$numCurrentCachedStates++;
	swapCachedStates($pkt->[$ID_FIELD]) if $numCurrentCachedStates > $STATE_CACHE_MAX;
      }
    }

    if ( $clientStateVec ) {
      return 0 if $#{$clientStateVec} < 0;
      $clientData[$client]->[CLIENT_STATESET] = $clientStateVec;
    }
    else {
      $cssMiss++;
      my $nextParity;
      my $edgeParity;
      foreach $stateVec (@{$clientData[$client]->[CLIENT_STATESET]}) {
	$edgeList = $FSA->[$stateVec->[STATE_ID]]->[$sym];
	# We should have a wildcard link iff there are no links for any other symbol
	# (Now passing $SYM_WILDCARD in one call that wants to take that transition)
	#$edgeList = $FSA->[$stateVec->[STATE_ID]]->[$SYM_WILDCARD] if !defined($edgeList);
	next if !defined($edgeList);
	foreach $edge (@{$edgeList}) {
	  $nextState = $edge->[edge_nextState];
	  # verify that the edge goes in the right direction for this transition
	  # $edge->[edge_parity] should be PARITY_ANY only for marker symbol edges,
	  # in which case we don't care about parity any more
	  if ( $edge->[edge_parity] != PARITY_ANY ) {
	    $edgeParity = ( $dir + $edge->[edge_parity]) % 2;
	    if ( $stateVec->[STATE_CONV_PARITY] != PARITY_ANY ) {
	      next if $edgeParity != $stateVec->[STATE_CONV_PARITY];
	      $nextParity = $stateVec->[STATE_CONV_PARITY];
	    } else {
	      $nextParity = $edgeParity;
	    }
	  }
	  else {
	    $nextParity = $stateVec->[STATE_CONV_PARITY];
	  }
	  # we don't have to worry about parity for costs -- all that matters is getting
	  # there, with packets going in the right directions for the state descriptor(s).
	  $newCost = $stateVec->[STATE_PATHCOST] + $edge->[edge_pathWeight];
	  if ( !defined($newStatePathCost[$nextState]) || $newCost < $newStatePathCost[$nextState] ) {
	    # we want to remember the conversation's parity (not the edge's)
	    # updated form in which we keep track of the full path to save time/space (hopefully)
	    # $newStateVec = [ $nextState, $nextParity, $newCost, [ @{$stateVec->[STATE_PATHLIST]}, $edge ] ];
	    $newStateVec = [ $nextState, $nextParity, $newCost, [ $stateVec->[STATE_PATHLIST], $edge ] ];
	    push @candidateStateSet,($newStateVec);
	    $newStatePathCost[$nextState] = $newCost;
	  }
	}
      }

      # the first occurrence of each destination state, going backwards through @candidateStateSet,
      # is the lowest cost one.

      $clientStateVec = [];
      for (my $i=$#candidateStateSet; $i>=0; $i-- ) {
	if ( $candidateStateSet[$i]->[STATE_PATHCOST] == $newStatePathCost[$candidateStateSet[$i]->[STATE_ID]] ) {
	  push @{$clientStateVec}, ($candidateStateSet[$i]);
	  $newStatePathCost[$candidateStateSet[$i]->[STATE_ID]] = -1;  # something impossible to match
	}
      }
      $currentCachedStateSet->{($clientData[$client]->[CLIENT_STATESET], $sym, $dir)} = $clientStateVec;
      $numCurrentCachedStates++;
      swapCachedStates($pkt->[$ID_FIELD]) if $numCurrentCachedStates > $STATE_CACHE_MAX;


      # If we just processed a symbol that shouldn't be in this conversation, there
      # will be no states in the new set.  That's a problem.
      # We handle it in the mainline by forcing an end of conversation, and then retrying this symbol
      # NOTE: We cannot destroy the current client state set until we know that there are some plausible
      # successor states -- the mainline will want to terminate this conversation, from where it is,
      # if that's the case.

      return 0 if $#{$clientStateVec} < 0;

      $clientData[$client]->[CLIENT_STATESET] = $clientStateVec;
    }

    if ( defined($pkt) ) {
      push @{$clientData[$client]->[CLIENT_PACKET_LIST]}, ($pkt);
      $pkt->[$PENDING_FIELD] = 1;
    }

    return 1;
}

############################################
# resetClientSeqnos
############################################

sub resetClientSeqnos {
  my ($pkt) = @_;
  return if !defined($pkt);
  $clientData[$pkt->[$CLIENT_FIELD]]->[CLIENT_CONV_LAST_SEQNO]->[$pkt->[$DIR_FIELD]] = $pkt->[$SEQNO_FIELD];
  $clientData[$pkt->[$CLIENT_FIELD]]->[CLIENT_CONV_LAST_FRAGNO] = $pkt->[$FRAGNO_FIELD];
}


############################################
# stateIDList()
############################################

sub stateIDList {
    my $client = shift;
    return ' ' if !defined($client);
    return ' ' if !defined($clientData[$client]->[CLIENT_STATESET]);
    join(',', map( (defined $_) && $_->[STATE_ID].'('.$_->[STATE_CONV_PARITY].')', 
		   sort {$a->[STATE_ID] <=> $b->[STATE_ID]} @{$clientData[$client]->[CLIENT_STATESET]}));
}


############################################
# usage()
############################################

sub usage {
  print "
Usage:  ./lostPackets.pl [-C cachsize] [-d level] [-dat datafile] [-di cnt] [-dp] [-e regexpFile] [-l packetLimit] [-m fragno] [-nw] [-pi cnt]
              [-rg cnt] [-s startID] [-t timeout] [-wg cnt] [-v] srcTable [basicOutputTable] [extrasOutputName]
Options:
         -C cachesize    Set limit on cached state size.
         -d level        Set debugging output level.
         -dat datafile   Input is taken from datafile.  Field information is
                         still read from srcTable.  (Used to debug the reg-exp.)
         -di cnt         Unlock/re-lock tables, every cnt packets.
                         [Default: $DUMP_INTERVAL]
         -dp             Keep statistics about distinct path sequences seen.
                         [Default: off]
         -e regexpFile   Take the regular expression from regexpFile, rather
                         than the hardcoded default.
         -l packetLimit  Set number of src table records to read.
         -m fragno       Set the maximum fragment number required in the reg-exp.
                         [Default: $max_fragno]
         -nw             Suppress warnings related to AP mismatches.
         -pi cnt         Set the number of -di intervals to go by between
                         printing current results.
                         [Default: $PRINT_INTERVAL]
         -R              Use RTS duration field info to terminate conversations. 
         -rg cnt         Read from src table in batches of cnt rows/query.
                         [Default: $READ_GROUPSIZE]
         -s startID      Set id of first src table row to read.
                         [Default: $startID]
         -t timeout      Time value for marker generation (sec.)
         -v              Be verbose.
         -w wtFile       Take symbol weights from file.
                         [Default: assign all symbols weight 1]
         -wg cnt         Write to output tables in batches of cnt rows/query.
                         [Default: $WRITE_GROUPSIZE];
";
  exit 1;
}

############################################
# mainline
############################################

my $nextPacketRef;  # data read from db
my $thisPacket;     # copy of packet put into pending queue
my @lastPacket;
my $dst;
my $src;
my $client;
my $bssid;
my $path;
my $marker;

my $nextID;
my $ID_LIMIT;

my $count = 0;

my $useDB = 1;

my $dataFile;
my $weightFile;

my $srcTable;
my $basicOutputTable;
my $extrasOutputTable;
my $synTable;
my $weightTable;

$max_fragno = MAX_FRAGNO;

while ( $ARGV[0] =~ /^-/ ) {
  if ( $ARGV[0] eq '-d' ) {
    shift;            # get rid of '-d'
    $debug = shift;   # read debug level
  }
  elsif ( $ARGV[0] eq '-dp' ) {
    shift;
    $wantDistinctPaths = 1;
  }
  elsif ( $ARGV[0] eq '-C' ) {
    shift;
    $STATE_CACHE_MAX = shift;
  }
  elsif ( $ARGV[0] eq '-l' ) {
    shift;
    $limitCount = shift;
  }
  elsif ( $ARGV[0] eq '-m' ) {
    shift;
    $max_fragno = shift;
  }
  elsif ( $ARGV[0] eq '-e' ) {
    shift;
    my $regexpFile = shift;
    open REGEXPFILE, "<$regexpFile" || die "Couldn't open regular experssion input file $regexpFile";
    $FSA_INPUT = join(' ',<REGEXPFILE>);
    close REGEXPFILE;
  }
  elsif ( $ARGV[0] eq '-s' ) {
    shift;
    $startID = shift;
  }
  elsif ( $ARGV[0] eq '-t' ) {
    shift;
    $MARKER_TIMEOUTVALUE = shift;
  }
  elsif ( $ARGV[0] eq '-nw' ) {
    shift;
    $wantAPWarnings = 0;
  }
  elsif ( $ARGV[0] eq '-v' ) {
    shift;
    setVerbose(1);
  }
  elsif ( $ARGV[0] eq '-R' ) {
    shift;
    $USE_RTS_CHECK = 1;
  }
  elsif ( $ARGV[0] eq '-rg' ) {
    shift;
    $READ_GROUPSIZE = shift;
  }
  elsif ( $ARGV[0] eq '-wg' ) {
    shift;
    $WRITE_GROUPSIZE = shift;
  }
  elsif ( $ARGV[0] eq '-di' ) {
    shift;
    $DUMP_INTERVAL = shift;
  }
  elsif ( $ARGV[0] eq '-pi' ) {
    shift;
    $PRINT_INTERVAL = shift;
  }
  elsif ( $ARGV[0] =~ /^-dat/ ) {
    shift;
    $dataFile = shift;
    undef $useDB;
  }
  elsif ( $ARGV[0] eq '-w' ) {
    shift;
    $weightFile = shift;
  }
  else {
    usage();
  }
}

if ( defined($dataFile) ) {
  usage($0) if  $#ARGV!=0;
}
else {
  usage($0) if ($#ARGV < 0) or ($#ARGV > 2);

}

$srcTable = shift;
$basicOutputTable = shift;
$extrasOutputTable = shift;


for ( my $i=0; $i<=MAX_FRAGNO; $i++ ) {
  push @symList, ( "FRAG_$i", "FRAG_$i". 'R', "DATA_" . ($i+1), "DATA_" . ($i+1) . 'R' );
}
push @symList, ( 'DATA_MAX', 'DATA_MAXR' );

for ( my $i=0; $i<=MAX_FRAGNO; $i++ ) {
  $symNameToSubtypeMap{"FRAG_$i"} = 'DATA';
  $symNameToSubtypeMap{"FRAG_$i" . 'R'} = 'DATA';
  $symNameToSubtypeMap{'DATA_' . ($i+1)} = 'DATA';
  $symNameToSubtypeMap{'DATA_' . ($i+1). 'R'} = 'DATA';
}
$symNameToSubtypeMap{'DATA_MAX'} = 'DATA';
$symNameToSubtypeMap{'DATA_MAXR'} = 'DATA';


dbconnect();

my $query;
my $conn;
my $sth;

if (defined($useDB)) {
  my %squeezedName = ( sahara => 'sah', chihuahuan => 'chi', sonoran => 'son', kalahari => 'kal', mojave => 'moj' );
  if ( !defined($basicOutputTable) ) {
    $basicOutputTable = $srcTable;
    $basicOutputTable =~ s/^[^_]+_/processed_/;
    foreach ( keys %squeezedName ) {   $basicOutputTable =~ s/_$_\_/_$squeezedName{$_}_/; }
  }
  $weightTable = $basicOutputTable . '_SymWts';
}
else {
  undef $weightTable;  # don't write it, no matter what
}

initLog( $basicOutputTable, $dataFile );
initSubtypeMap();
initSymTable( \$weightFile, \$weightTable );
initFieldData( $srcTable );

expandDataRegExp( \$FSA_INPUT, $max_fragno);

if ( !defined($useDB) ) {
  open(TESTDATAFILE, "<$dataFile") || die "Couldn't open test data file $srcTable";
  $ID_LIMIT = $limitCount;
  $BROADCAST_NODE = 0;  # might be used in data file
  $UNKNOWN_NODE = 100;
  foreach (1..10) { $clientData[$_]->[CLIENT_NAME] = "testAP$_"; $clientData[$_]->[CLIENT_AP_TYPE] = AP_INFRASTRUCTURE; }
  $clientData[$UNKNOWN_NODE]->[CLIENT_NAME] = '';  # make the array 100 long...
}
else {
  $extrasOutputTable = $basicOutputTable . '_extras' if !defined($extrasOutputTable);
  $synTable = $basicOutputTable . '_SynthPkts';

  Log ALWAYS, "Start: ",  now(), "
SymWgt Defs:  $weightFile
Source Table: $srcTable
Basic Out:    $basicOutputTable
Extras Out:   $extrasOutputTable
Synth Out:    $synTable
SymWgt Out:   $weightTable
Timeout:      $MARKER_TIMEOUTVALUE

Regular expression used:
$FSA_INPUT
";

  initNodeMap( $srcTable );
# 6/14/2006: We no longer use a static table of APs.  Instead, add/remove them dynamically.
#  initAPs( $srcTable );
  initOutputTables( $srcTable, $basicOutputTable, $extrasOutputTable, $synTable );

  # prepare lock query , used for speed (we hope)
  $lockQuery = "LOCK TABLES $basicOutputTable WRITE,
                      $extrasOutputTable WRITE,
                      $synTable WRITE";

  # finally, prepare to fetch the packets from the db table
  $query = "SELECT COUNT(*) from $srcTable";
  $sth = doQuery(\$query);
  ($ID_LIMIT) = $sth->fetchrow_array;

  $query = "SELECT * FROM $srcTable
               WHERE zErr is null AND id>=? AND id<?";
  $conn = getConnection();
  $sth = prepareQuery( $conn, \$query );
}

# summarize key parameters

Log ALWAYS, "APs:\n";
for ( my $node=0; $node<=$#clientData; $node++ ) {
  next unless $clientData[$node]->[CLIENT_AP_TYPE];
  my $type = $clientData[$node]->[CLIENT_AP_TYPE] == AP_ADHOC ? 'ad hoc' : 'ap';
  Log ALWAYS, sprintf("   $node\t%6s\t$clientData[$node]->[CLIENT_NAME]\n", $type);
}
Log ALWAYS, "Broadcast node: $BROADCAST_NODE\n";
Log ALWAYS, "Unknown node: $UNKNOWN_NODE\n";
Log ALWAYS, "Not associated node: $NOTASSOCIATED_NODE\n";

initClientData();

# now build the fsa that drives everything

($FSA, $START_STATE, $ACCEPT_STATE, $WILDCARD_STATE) =
  constructFSA( \%SYM, \@symNameMap, \@symWeight, $FSA_INPUT, ALWAYS, DEBUG_WARN, DEBUG_FULL, $debug);

$currentCachedStateSet->{undef} = [[$START_STATE, PARITY_ANY, 0, []]];
$currentCachedStateSet->{wildcard} = [[$WILDCARD_STATE, PARITY_ANY, 0, []]];
$numCurrentCachedStates = 2;
$cssTimestamp = gettimeofday;
$cssPacketNum = $startID;

mirrorPrint( formatCacheStats() );

for ( ; $startID<=$ID_LIMIT and $count < $limitCount; $startID += $READ_GROUPSIZE ) {

  $sth->execute($startID, $startID+$READ_GROUPSIZE) if $useDB;
  last if !defined($useDB) && eof(TESTDATAFILE);

  while ( ($useDB && ($nextPacketRef = $sth->fetchrow_arrayref)) ||
          ($nextPacketRef = fetchTestPacket(*TESTDATAFILE))
         ) {

    $count++;
    if ( $count%100000 == 0 ) {
      mirrorPrint( "$count packets read\n" );
      mirrorPrint( "Pending packet queue size: $#pendingPacketList\n" );
      mirrorPrint( "$NEXT_PATHSEQUENCE_ID distinct paths\n") if $wantDistinctPaths;
    }
    last if $count > $limitCount;

    if ($count%$DUMP_INTERVAL == 0) {
      if ( $count % ($DUMP_INTERVAL*$PRINT_INTERVAL) == 0 ) {
	dumpResults( $count, $useDB, 1 ) ;
      } else {
	dumpResults( $count, $useDB, 0 ) ;
      }
    }

    $CURRENT_PACKET_NUM = $nextPacketRef->[$ID_FIELD];

    $thisPacket = [ @{$nextPacketRef} ];  # make a copy
    $#{$thisPacket} = $NUM_ALL_FIELDS;    # force record to correct length
    $thisPacket->[$PENDING_FIELD] = 0;    # takeFSAStep will mark it as 1, if it is consumed in a transition

    $currentTime = $thisPacket->[$TIME_FIELD];
    $dst = $thisPacket->[$A1_DST_FIELD];
    $src = $thisPacket->[$A2_SRC_FIELD];

    $client = classifyPacket( $thisPacket, $lastPacket[$dst] );

    if ( $debug >= DEBUG_FULL ) {
      mirrorPrint( "\n--- Packet $thisPacket->[$ID_FIELD]: $thisPacket->[$TIME_FIELD] ",
		   $subtypeMap[$thisPacket->[$SUBTYPE_FIELD]],
		   " [Retry = $thisPacket->[$RETRY_FIELD] Seqno = $thisPacket->[$SEQNO_FIELD]]\n",
		   "Src = $src [", stateIDList($src), "] Dst = $dst [", stateIDList($dst), "]\n");
      mirrorPrint( "client = $client   parse symbol: $symNameMap[$thisPacket->[$PARSESYMBOL_FIELD]]   ",
		   "dir: $thisPacket->[$DIR_FIELD]  clientField: $thisPacket->[$CLIENT_FIELD] ",
		   "partnerField: $thisPacket->[$PARTNER_FIELD]\n");
    }

    if ( $client != $BROADCAST_NODE && $client != $UNKNOWN_NODE ) {
      # see if an ongoing conversation involving $client should be terminated
      $marker = generateMarker( $thisPacket );
      if ( $marker != $MARKER_NONE && defined($clientData[$client]->[CLIENT_STATESET]) ) {
	takeFSAStep( $client, $marker, undef );
	terminateConversation( $client, $thisPacket->[$ID_FIELD] );
	resetClientSeqnos($thisPacket);
	scanPendingPackets();
      }
    }

    # deal with the packet we just read
    if ( $client == $BROADCAST_NODE ||
	 $client == $UNKNOWN_NODE ||
	 $thisPacket->[$PARSESYMBOL_FIELD] == $SYM_IGNORED ) {
      $clientData[$client]->[CLIENT_STATESET] = $currentCachedStateSet->{wildcard};
      takeFSAStep( $client, $SYM_WILDCARD, $thisPacket );
      terminateConversation( $client, $thisPacket->[$ID_FIELD] );
      # deciding not to bother with scanPendingPackets() here...
      $thisPacket->[$CONVID_FIELD] = $CONVERSATION_ID++;
    }
    else {
      if (! takeFSAStep( $client, $thisPacket->[$PARSESYMBOL_FIELD], $thisPacket ) ) {
	# the symbol wasn't valid for any current state.
	if ( $debug >= DEBUG_FULL ) {
	  mirrorPrint( "INFO: No next state for client $client at packet $thisPacket->[$ID_FIELD] (",
		       $symNameMap[$thisPacket->[$PARSESYMBOL_FIELD]], ")\n",
		       "\t[", stateIDList($client), "]\n");
	}

	takeFSAStep( $client, $MARKER_SYM, undef );  # force termination of this conversation (hopefully)
	terminateConversation( $client, $thisPacket->[$ID_FIELD] );
	resetClientSeqnos($thisPacket);
	scanPendingPackets();

	# now try again
	takeFSAStep( $client, $thisPacket->[$PARSESYMBOL_FIELD], $thisPacket ) ||
	  die "INTERNAL ERROR: Replay of packet $thisPacket->[$ID_FIELD] still didn't find a transition\n" .
	    "   client = $client [" . stateIDList($client) . ']';
      }

      $thisPacket->[$CONVID_FIELD] = $clientData[$client]->[CLIENT_CURRENT_CONV];

    }
    if ( $debug >= DEBUG_FULL ) {
      mirrorPrint( "After processing: Src = $src [", stateIDList($src), "] Dst = $dst [", stateIDList($dst), "]\n" );
    }

    push @pendingPacketList, ( $thisPacket );
    $lastPacket[$thisPacket->[$A2_SRC_FIELD]] = $thisPacket;  # hopefully, src field is filled in by now

    # make sure the pending list is scanned reasonably requently

    scanPendingPackets() if $currentTime - $lastScanTime > $SCAN_TIMEOUT;
  }
}
$sth->finish() if $useDB;

$conn->disconnect if $useDB;

# terminate any still live connections

for ( $client=0; $client<=$#clientData; $client++ ) {
    if ( defined($clientData[$client]->[CLIENT_STATESET]) ) {
	takeFSAStep( $client, $MARKER_SYM, undef );
	terminateConversation( $client, $count++ );
    }
}

# now try to flush any remaining packets
$currentTime = 1000*1000*1000*500;
scanPendingPackets();
pushOutput() if $useDB;

dumpResults( $count, $useDB, 1 );

mirrorPrint( formatCacheStats() );

dbdisconnect();


