#!/usr/bin/perl

# Tries to compute statistics on a trace file (db table) 
# conditioned on number of clients with packets to send.

use strict;
require 'localutils.pl';

my $verbose;
my $fromRecord = -1;
my $toRecord = -1;
my $interval = 0;
my $start = -1;
my $stop = -1;

my $numContenders = 0;

use constant { DIFS => 50, SLOT_TIME => 20, CW_MIN => 31};
my $RANDOM_WAIT_TIME = DIFS + rand(CW_MIN) * SLOT_TIME;
my $PrintContenders = 0;

# statistics are kept in three arrays: one for time-based,
# one for packet-based, and one for interval-based  measures.
# In each, the first level is the client, and the second the number of contenders.
# The time-based stats are:
#   $timeStats[client]->[nContenders]->[idle|busy|deferring]
# The packet-based stats are high dimensionality:
#   $packetStats[client]->[nContenders]->[data pkt|not data]->[dir]->[bcast|notBcast]->[nav|notNav]...
#                            ->[retry|original]->[succeed|fail]->{rate}->[bytes|packets]
#                                                                ^^^^^^
# The interval stats are:
#   $intervalStats[client]->[postPktType]->{prePktXmitRate}->{postPktXmitRate}->[prePktRetry]->[postPktRetry]->...
#                       [(pre)Dir]->[cnt|totalTime]
# Note: output table contains [cnt|avg. time], for convenience
#
#
# One might be tempted to exploit the fact that DATA and NAV can't both be true,
# with the current definitions.  Don't -- it won't buy much, and fragmented
# tranmissions have data packets covered by a nav.

# time-based stats and symbolic constants
my @timeStats = ();
my $IDLE_TIME              = 0;
my $DEFER_TIME             = 1;
my $DEFER_TO_INACTIVE_TIME = 2;
my $BUSY_TIME              = 3;
my $NAV_BUSY_TIME          = 4;
my $NUM_TIMES = 4;

# packet-based stats and symbolic constants
my @packetStats = ();

my $NOT_DATA = 0;
my $IS_DATA  = 1;
my $isData;  # set for each packet

my $UP = 0;
my $DOWN = 1;
my $SIDEWAYS = 2;

my $NOT_NAV = 0;
my $IS_NAV  = 1;
my $isNav;

my $NOT_RETRY = 0;
#my $IS_RETRY  = 1;  # deprecated -- retry produced by lostPackets is now a count 

my $NOT_HEARD = 0;
my $HEARD     = 1;
my $NO_HEARD_INFO = 2;

my $BY_BYTES  = 0;
my $BY_PACKET = 1;

# interval stats
my @intervalStats = ();

use constant {
  INTVL_COUNT   => 0,
  INTVL_LENGTH  => 1,
  INTVL_SUCCEED => 2,
  INTVL_DEFER   => 3,
};

# a helper, to keep track of all the rates we've seen
my %seenRate;

# client 0 is used for stats when we don't know who the client is
my $UNKNOWN_CLIENT = 0;

# client NO_CLIENT is used to hold idleness when there are no contenders.
# it'll be set to something one beyond the range of real client id's in
# the body of the code, then translated to -1 when creating the output tables.
my $NO_CLIENT;

# we map from the input data client numbers to a new client UIDs, so that
# we can do a coarse distinction of client sessions.  (We're want a single
# "client" to represent a relatively fixed situation, e.g., basic connectivity
# parameters.  Actual clients move around from time to time.  We hope that
# the timeout approximates when those changes occur.)

my $MILLION = 1000*1000;
my $SESSION_TIMEOUT_MINUTES = 10;   # set this variable to change timeout
my $SESSION_TIMEOUT = $SESSION_TIMEOUT_MINUTES * 60 * $MILLION;  # don't monkey with this one

# clients for which we have last seen a non-retry packet
# are kept in @firstClient.  We select a random wait time in U(0,31)*10
# and countdown on airtime idleness before removing them as contenders.
my @firstClient = ();
my $FIRSTCLIENT_ID = 0;
my $FIRSTCLIENT_DELAY = 1;
my $FIRSTCLIENT_ENTRY = 2;
my $FIRSTCLIENT_CTYPE = 3;

# vector of state information for each client (indexed by client id in source data):
# - 
# - state:  
# - last activation time: 
# - last deactivation time: 

my @clientData;  # indexed by (actual, source data) client, with fields

my $CLIENT_STATE             = 0;   # active, actve on first queue, or idle
my $CLIENT_LAST_ACTIVATION   = 1;   # when we last added client to the active list
my $CLIENT_SESSION_ID        = 2;   # current virtual client uid
my $CLIENT_SESSION_START     = 3;   # start time of this virtual client uid
my $CLIENT_SESSION_STOP      = 4;   # stop time of this virtual client uid
my $CLIENT_INTVL_STOP_RECORD = 5;   # ref to an interval record

my $IDLE_STATE = 0;
my $ACTIVE_STATE = 1;
my $ON_IDLE_LIST_STATE = 2;

use constant {
  INTVLREC_TIME      => 0,
  INTVLREC_RETRY     => 1,
  INTVLREC_XMIT_RATE => 2,
  INTVLREC_CONVID    => 3,
  INTVLREC_TYPE      => 4,
  INTVLREC_SUCCEED   => 5,
  INTVLREC_DEFERTIME => 6,
};

my $nextSessionID = 1;  # counter to generate virtual client UIDs

# list of clients in active state
my @activeClients;

# these are used for debugging, because I'm worried that we'll forget to
# deal with some way by which a client can be considered active for eternity
my @activeTimeHist;
my $maxActiveTime = 0;

# table/query handling...
my @outputTable;
my @outputTableSuffix = ('_Time', '_Packet', '_Client', '_Interval');

my $TIME_TABLE     = 0;
my $PACKET_TABLE   = 1;
my $CLIENT_TABLE   = 2;
my $INTERVAL_TABLE = 3;
my $NUM_TABLE      = 3;

my $WAY_BIG = $MILLION * $MILLION * $MILLION;

## variables for authoritative num contenders computation
my $truthTable;
my @kickOutTime;
my @scheduledTimes;

###################################################
# insertValues()
###################################################

sub insertValues {
  my ($table, $vals) = @_;

  $outputTable[$table]->{insertQuery} .= $$vals;
  $outputTable[$table]->{insertCount}++;
  $outputTable[$table]->{insertID}++;

  if ($outputTable[$table]->{insertCount} % 1000 == 0 ) {

    my $lockStr = "LOCK TABLES $outputTable[$table]->{name} WRITE";
    doQuery( \$lockStr );

    chop $outputTable[$table]->{insertQuery};
    doQuery( \$outputTable[$table]->{insertQuery} );
    $outputTable[$table]->{insertQuery} = $outputTable[$table]->{insertQueryRoot};
    $outputTable[$table]->{insertCount} = 0;

    $lockStr = 'UNLOCK TABLES';
    doQuery( \$lockStr );
  }
}

###################################################
# makeTableName()
###################################################

sub makeTableName {
  my ($rootName, $genericSuffix, $tblIndex) = @_;

  # WE FORCE A SUFFIX FOR THIS TEST VERSION
  #$genericSuffix = \'TEST';

  # trying to be careful to avoid wiping out the source table, in case something goes wrong...
  my $table = $$rootName;
  $table .= '_' . $$genericSuffix if defined($$genericSuffix);
  unless ( $table =~ s/^[^_]+/contention/ ) {
    mirrorPrint( "CAN'T create a new name for output table:  '$table' derived from '$$rootName'\n");
    return undef;
  }
  $table .= $outputTableSuffix[$tblIndex];

  $outputTable[$tblIndex]->{name} = $table;
  $outputTable[$tblIndex]->{insertQueryRoot} = "INSERT INTO $table VALUES";

  return $table;
}

###################################################
# createOutputTables()
###################################################

sub createOutputTables {
  my ($rootTableName, $suffix ) = @_;

  my $timeTable = makeTableName( $rootTableName, $suffix, $TIME_TABLE );
  my $packetTable = makeTableName( $rootTableName, $suffix, $PACKET_TABLE );
  my $clientTable = makeTableName( $rootTableName, $suffix, $CLIENT_TABLE );
  my $intervalTable = makeTableName( $rootTableName, $suffix, $INTERVAL_TABLE );

  my $query = "DROP TABLE IF EXISTS $timeTable, $packetTable, $clientTable, $intervalTable";
  doQuery( \$query );

  ##########################
  # first the time stats
  ##########################

  $query = 
"CREATE TABLE $timeTable ( id              int NOT NULL UNIQUE,
                           client          int,
                           nContenders     tinyint UNSIGNED,
                           idle            bigint,
                           defer           bigint,
                           deferToInactive bigint,
                           busy            bigint,
                           navBusy         bigint,
                           PRIMARY KEY(id) )";
  doQuery( \$query );
  mirrorPrint("Created $timeTable\n");

  #########################
  # now packet stats
  #   $packetStats[client]->[nContenders]->[data pkt|not data]->[dir]->[nav|notNav]
  #                         ->[retry|original]->[succeed|fail]->{rate}->[bytes|packets]
  #########################

  $query = 
"CREATE TABLE $packetTable ( id          int NOT NULL UNIQUE,
                           client      int UNSIGNED,
                           nContenders tinyint UNSIGNED,
                           isData      tinyint UNSIGNED,
                           dir         tinyint UNSIGNED,
                           isNav       tinyint UNSIGNED,
                           isRetry     tinyint UNSIGNED,
                           wasRcvd     tinyint UNSIGNED,
                           rate        float,
                           nBytes      int UNSIGNED,
                           nPkts       int UNSIGNED,
                           PRIMARY KEY(id) )";
  doQuery( \$query );
  mirrorPrint("Created $packetTable\n");

  #########################
  # now interval stats
  #   $intervalStats[client]->[postPktType]->{prePktXmitRate}->{postPktXmitRate}->[prePktRetry+1]->[postPktRetry+1]->...
  #                       [cnt|totalTime]
  #########################

  $query = 
"CREATE TABLE $intervalTable ( id          int NOT NULL UNIQUE,
                           client      int UNSIGNED,
                           postType    tinyint UNSIGNED,
                           preRate     float,
                           preRetry    tinyint UNSIGNED,
                           postRate    float,
                           postRetry   tinyint UNSIGNED,
                           dir         tinyint UNSIGNED,
                           nIntervals  int UNSIGNED,
                           avgInterval float,
                           nSucceed    int UNSIGNED,
                           deferTime   float,
                           PRIMARY KEY(id) )";
  doQuery( \$query );
  mirrorPrint("Created $intervalTable\n");


  ##########################
  # finally, client table
  ##########################

  $query = 
"CREATE TABLE $clientTable ( id              int NOT NULL UNIQUE,
                           vClient         int,
                           rClient         int,
                           fromTime        decimal(18,7),
                           toTime          decimal(18,7),
                           PRIMARY KEY(id) )";
  doQuery( \$query );
  mirrorPrint("Created $clientTable\n");

  ##########################
  # initialize buffering tables
  ##########################

  for ( my $t=0; $t<=$NUM_TABLE; $t++ ) {
    $outputTable[$t]->{insertQuery} = $outputTable[$t]->{insertQueryRoot};
    $outputTable[$t]->{insertCount} = 0;
    $outputTable[$t]->{insertID}    = 1;
  }
}

###################################################
# dumpClientStats()
#   Dumps the counters into two tables,
#   one for per-packet data and one for time-based
###################################################

sub dumpClientStats {
  my ( $client ) = @_;

  my $contenders;
  my $query;

  #########################
  # first the time stats
  #########################

  if ( defined($timeStats[$client]) ) {
    for ($contenders=0; $contenders<=$#{$timeStats[$client]}; $contenders++ ) {
      next if !defined($timeStats[$client]->[$contenders]);
      $query = '(' . $outputTable[$TIME_TABLE]->{insertID} . ',' . $clientData[$client]->[$CLIENT_SESSION_ID] . ",$contenders," .
	join(',',  map {$timeStats[$client]->[$contenders]->[$_] ?
			  int($timeStats[$client]->[$contenders]->[$_]+0.5) : 0}($IDLE_TIME..$NUM_TIMES) ) .
			    '),';
      insertValues( $TIME_TABLE, \$query );
    }
    undef $timeStats[$client];
  }

  #########################
  # now packet stats
  #   $packetStats[client]->[nContenders]->[data pkt|not data]->[dir]->[nav|notNav]
  #                         ->[retry|original]->[succeed|fail]->{rate}->[bytes|packets]
  #########################

  my @rateVec = sort keys %seenRate;

  if ( defined($packetStats[$client]) ) {
    for ($contenders=0; $contenders<=$#{$packetStats[$client]}; $contenders++ ) {
      next if !defined($packetStats[$client]->[$contenders]);
      for (my $isData=$NOT_DATA; $isData<=$IS_DATA; $isData++ ) {
	next if !defined($packetStats[$client]->[$contenders]->[$isData]);
	for (my $dir=$UP; $dir<=$SIDEWAYS; $dir++) {
	  next if !defined($packetStats[$client]->[$contenders]->[$isData]->[$dir]);
	  for (my $isNav=$NOT_NAV; $isNav<=$IS_NAV; $isNav++ ) {
	    next if !defined($packetStats[$client]->[$contenders]->[$isData]->[$dir]->[$isNav]);
	    for (my $retry=$NOT_RETRY; $retry <= $#{$packetStats[$client]->[$contenders]->[$isData]->[$dir]->[$isNav]}; $retry++ ) {
	      next if !defined($packetStats[$client]->[$contenders]->[$isData]->[$dir]->[$isNav]->[$retry]);
	      for (my $succ=$NOT_HEARD; $succ<=$NO_HEARD_INFO; $succ++ ) {
		next if !defined($packetStats[$client]->[$contenders]->[$isData]->[$dir]->[$isNav]->[$retry]->[$succ]);
		foreach my $rate (@rateVec) {
		  next if (!defined($packetStats[$client]->[$contenders]->[$isData]->[$dir]->[$isNav]->[$retry]->[$succ]->{$rate}));
		  #$packetStats[$client]->[$contenders]->[$isData]->[$isNav]->[$retry]->[$succ]->{$rate}->[$by];
		  $query = '(' . $outputTable[$PACKET_TABLE]->{insertID} . ',' . $clientData[$client]->[$CLIENT_SESSION_ID] .
		       ",$contenders,$isData,$dir,$isNav,$retry,$succ,$rate," .
			 join(',',map{$packetStats[$client]->[$contenders]->[$isData]->[$dir]->[$isNav]->[$retry]->[$succ]->{$rate}->[$_]?
					$packetStats[$client]->[$contenders]->[$isData]->[$dir]->[$isNav]->[$retry]->[$succ]->{$rate}->[$_]:0}
			      ($BY_BYTES..$BY_PACKET) ) .
				'),';
		  insertValues( $PACKET_TABLE, \$query );
		}
	      }
	    }
	  }
	}
      }
    }
    undef $packetStats[$client];
  }

  ############################################################
  # Interval statistics
  ############################################################

#   $intervalStats[client]->{prePktXmitRate}->{postPktXmitRate}->[prePktRetry+1]->[postPktRetry+1]->...
#                       [(pre)dir]->[cnt|totalTime]

  my $p3;
  my $p4;
  my $p5;
  if ( defined($intervalStats[$client]) ) {
    for (my $type=0; $type<=$#{$intervalStats[$client]}; $type++ ) {
      next if !defined($intervalStats[$client]->[$type]);
      foreach my $preRate (sort keys %{$intervalStats[$client]->[$type]}) {
	foreach my $postRate (sort keys %{$intervalStats[$client]->[$type]->{$preRate}}) {
	  for (my $preRetry=0; $preRetry <= $#{$intervalStats[$client]->[$type]->{$preRate}->{$postRate}}; $preRetry++ ) {
	    $p3 = $intervalStats[$client]->[$type]->{$preRate}->{$postRate}->[$preRetry];
	    next if !defined($p3);
	    for (my $postRetry=0; $postRetry <= $#{$p3}; $postRetry++ ) {
	      $p4 = $intervalStats[$client]->[$type]->{$preRate}->{$postRate}->[$preRetry]->[$postRetry];
	      next if !defined($p4);
	      for (my $dir=0; $dir<=$#{$p4}; $dir++ ) {
		next if !defined($p4->[$dir]);
		$p5 = $intervalStats[$client]->[$type]->{$preRate}->{$postRate}->[$preRetry]->[$postRetry]->[$dir];
		next unless defined($p5);
		$query = '(' . $outputTable[$INTERVAL_TABLE]->{insertID} . 
		             ",$client, $type, $preRate, $preRetry, $postRate, $postRetry, $dir, $p5->[INTVL_COUNT]," .
		             ($p5->[INTVL_LENGTH]/$p5->[INTVL_COUNT]) . ', ' .
			     ($p5->[INTVL_SUCCEED]?$p5->[INTVL_SUCCEED]:0) . ', ' .
			     ($p5->[INTVL_DEFER]?$p5->[INTVL_DEFER]:0) .
		       '),';
		insertValues( $INTERVAL_TABLE, \$query );
	      }
	    }
	  }
	}
      }
    }
  }
  undef $intervalStats[$client];
  undef $clientData[$client]->[$CLIENT_INTVL_STOP_RECORD];

  #########################
  # Now data about the client
  #########################

  my $start = $clientData[$client]->[$CLIENT_SESSION_START] / $MILLION;
  my $stop = $clientData[$client]->[$CLIENT_SESSION_STOP] / $MILLION;
  $query = "($outputTable[$CLIENT_TABLE]->{insertID},$clientData[$client]->[$CLIENT_SESSION_ID],$client,$start,$stop),";
  insertValues( $CLIENT_TABLE, \$query );
  undef $clientData[$client]->[$CLIENT_SESSION_ID];
}

###################################################
# deactivate() - remove client from firstClient list
###################################################

sub deactivate {
  my ($now) = @_;

  my $client = $firstClient[0]->[$FIRSTCLIENT_ID];

  my $diff = $firstClient[0]->[$FIRSTCLIENT_ENTRY] - $now;
  mirrorPrint("deactivating $client: $diff $firstClient[0]->[$FIRSTCLIENT_ENTRY] $firstClient[0]->[$FIRSTCLIENT_CTYPE]\n") if ($verbose);

  shift @firstClient;  # remove this guy
  $numContenders--;
  $clientData[$client]->[$CLIENT_STATE] = $IDLE_STATE;

  # keep some stats to try to point out egregious failures to
  # terminate a client's activity
  my $activeTime = $clientData[$client]->[$CLIENT_LAST_ACTIVATION] - $now;
  $maxActiveTime = $activeTime if $maxActiveTime < $activeTime;
  $activeTime > 10 ?  $activeTimeHist[int(log($activeTime)/log(10.0))]++ : $activeTimeHist[0]++;

  # remove the guy from the list of active clients
  my $activeIndex;
  for  ($activeIndex=0; $activeIndex<= $#activeClients; $activeIndex++ ) {
    last if $activeClients[$activeIndex] == $client;
  }
  if ($activeIndex <= $#activeClients) {
    splice(@activeClients, $activeIndex, 1);
  }
  else {
    mirrorPrint( "ERROR -- couldn't find firstlist client $client on active list!\n" );
  }
}

###############################################
# getKickoutTime() - get time when the client's current packet was scheduled
##############################################

sub getKickoutTime {
    my ($client, $timenow) = @_;

    if (!defined(@{$scheduledTimes[$client]}) or $scheduledTimes[$client]->[0] > $timenow) {
	my $query = "SELECT zSched*$MILLION FROM $truthTable WHERE aFrameTime < $timenow/$MILLION 
                       AND zSched IS NOT NULL 
                       AND cA2_Src = $client
                       ORDER BY aFrameTime DESC LIMIT 1";
	my $sth = doQuery(\$query);
	($scheduledTimes[$client]->[0]) = $sth->fetchrow_array();
    }
    
    die "ERROR: kickout time not found for $client / $timenow\n" if (!defined($scheduledTimes[$client]->[0]));

    my $schedTime = $scheduledTimes[$client]->[0];
    mirrorPrint(sprintf("schedtime = $schedTime at $timenow; diff = %d\n", $timenow - $schedTime)) if ($verbose);
    return $schedTime;
}

############################################
# convertTime()
############################################

my @months = ( qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec) );
sub convertTime {
    my ($time) = @_;
    my $secFrac = sprintf "%0.5f", $time - int($time);
    $secFrac =~ s/^0+//;
    my ($sec,$min,$hour,$mday,$mon,$year) = localtime($time);
    $year += 1900;
    $mon++;
    $time = sprintf "$months[$mon-1] $mday, $year %02d:%02d:%02d", $hour, $min, int($sec);
    $time .=  $secFrac if $secFrac > 0;
}

###################################
# usage()
###################################

sub usage {
    print "
Usage: ./computeContention.pl [-n count | -r fromRecord toRecord] [-V] table_name
Options:
   -i from to    Process records in this time interval.
   -n count      Process count records (from the beginning of the table).
   -r from to    Process records from to to.
   -s suffix     Append suffix to output table names.
   -auth: get #contenders from the truthtable (inferred from table_name)
   -pc           print the number of contender at each packet 
   -rwt <usecs>  random wait time to use before the original packet <difs+rand(cw_min)*slot_time>

   -v            Verbose mode (debugging).

If no record range is given, the entire table is processed.
";
}

###################################################
# mainline starts here
###################################################

my $client;
my $effectiveClient;

my $pkt;    # has the following fields (after the SELECT query)
my $Iid        = 0;
my $Itime      = 1; # Itime now refers to frameTime (we don't rely on MACTime past the merging process)
my $Idest      = 2; # values of Idest/Isrc must be consecutive, with Idest first!
my $Isrc       = 3;
my $Isize      = 4;
my $Irate      = 5;
my $Iretry     = 6;
my $Iseqno     = 7;
my $Isubtype   = 8;
my $Idir       = 9;
my $Ircvd      = 10;
my $Iconvid    = 11;

my $Iindex;

my $ACK = 29;
my $CTS = 28;

my $recnum=0;

my $TIMEOUT = 0.05 * $MILLION;  # in usec.
#my $TIMEOUT = 0.5 * $MILLION;  # in usec.

my $idleTime = 0;
my $remainingIdleTime;
my $idleTimeCorrection;
my $airTime;
my $lastTime;
my $endTime;
my $deferType;
my $busyType;
my $totalOverlap = 0;
my $maxOverlap = 0;
my $suffix;
my $type;
my $stopRec;

my $query;
my $sth;

my $CHUNKSIZE = 1000;

while ( $ARGV[0] =~ /^-/ ) {
    if ( $ARGV[0] eq '-v' ) {
	shift;
	$verbose = 1;
	setVerbose($verbose);
    }
    elsif ( $ARGV[0] eq '-n' ) {
	shift;
	$fromRecord = 1;
	$toRecord = shift;
    }
    elsif ($ARGV[0] eq "-r" ) {
	shift;
	$fromRecord = shift;
	$toRecord = shift;
    }
    elsif ($ARGV[0] eq "-i" ) {
	shift;
	$interval = 1;
	$start = shift;
	$stop = shift;
    }
    elsif ($ARGV[0] eq "-s" ) {
	shift;
	$suffix = shift;
    }
    elsif ($ARGV[0] eq "-auth" ) {
	shift;
	$truthTable = 1;
    }
    elsif ($ARGV[0] eq "-pc") {
	shift;
	$PrintContenders = 1;
    }
    elsif ($ARGV[0] eq "-rwt") {
	shift;
	$RANDOM_WAIT_TIME = shift;
    }
    else {
	usage();
	exit 2;
    }
}

if ($#ARGV != 0 ) {
    usage();
    exit 1;
}

my $table = shift;
if ($truthTable) {
    die "unknown table_name format $table\n" if ($table !~ m/[a-z]+_(.+)x([_\d]+|all)_i0_c0/);
    $truthTable = "raw_$1xall_i0_c0";
}

my $logName = "contention_" . $table . ".log.$suffix"; #contention_processed_sahara_sonoran_i0_c1.log
open (LOGFILE, ">$logName") || die "Couldn't open log file $logName";
setLogFile(*LOGFILE);

dbconnect();

# start working on what range of records to read

if ($interval == 1) {
    # get unix time of start and stop in order to get the id range corresponding to this interval,
    # and set fromeRecord and toRecord.
    my $uStart = toUnixTime($start);
    mirrorPrint ("start time: ", $uStart, "\n") if $verbose;
    $query = "SELECT id FROM $table WHERE aFrameTime>=$uStart LIMIT 1";
    $sth = doQuery(\$query);
    $fromRecord = $sth->fetchrow_array;
    mirrorPrint ("from rec: ", $fromRecord, "\n") if $verbose;

    my $uStop = toUnixTime($stop);
    mirrorPrint ("stop time: ", $uStop, "\n") if $verbose;
    $query = "SELECT MAX(id) FROM $table WHERE aFrameTime<=$uStop LIMIT 1";
    $sth = doQuery(\$query);
    $toRecord = $sth->fetchrow_array;
    mirrorPrint ("to rec: ", $toRecord, "\n") if $verbose;
}

elsif ( $fromRecord < 0 ) {
    $query = "SELECT COUNT(id) FROM $table";
    $sth = doQuery( \$query );
    $fromRecord = 1;
    $toRecord = $sth->fetchrow_array;
}

mirrorPrint( "\nReading table '$table', records ", commify($fromRecord), " - ", commify($toRecord), ".\n" );
my $stopRec = $toRecord;
my $startRec = $toRecord - $CHUNKSIZE;

$query = "SELECT aFrameTime from $table where id=$stopRec";
# $query = "SELECT aFrameTime, bMACTime from $table where id=$stopRec";
$sth = doQuery( \$query );
($endTime) = $sth->fetchrow_array;
#($endTime, $oldTime) = $sth->fetchrow_array;
$lastTime = $endTime * $MILLION; #convert from sec to microsec

# figure out how many clients are in the data, and then set $NO_CLIENT

$query = "SELECT MAX(id) FROM $table" . "_MACIndex";
$sth = doQuery( \$query );
($NO_CLIENT) = $sth->fetchrow_array;
$sth->finish();
$NO_CLIENT++;  # create a fake client UID to hold DELAY_TIME data when there are no contenders

# set up client initial data so they all get a new virtual client UID the first time we see them
for ( my $cl=0; $cl<=$NO_CLIENT; $cl++ ) {
  $clientData[$cl]->[$CLIENT_SESSION_START] = $WAY_BIG;
  $clientData[$cl]->[$CLIENT_STATE] = $IDLE_STATE;
}

$clientData[$UNKNOWN_CLIENT]->[$CLIENT_SESSION_ID] = 0;
$clientData[$UNKNOWN_CLIENT]->[$CLIENT_SESSION_STOP ] = $lastTime;

$clientData[$NO_CLIENT]->[$CLIENT_SESSION_ID] = -1;
$clientData[$NO_CLIENT]->[$CLIENT_SESSION_STOP ] = $lastTime;

# things look good.  create the tables (which will wipe them out if
# the exist already).

createOutputTables( \$table, \$suffix );

##############################################################
# start of main loop
##############################################################

do {
  # read table backwards
  $startRec = $fromRecord if $startRec < $fromRecord;
  $query = "SELECT id, aFrameTime, cA1_Dest, cA2_Src, bSize, bRate, cRetry, cSeqno, cSubtype, _zDir, _zWasrcvd, _zConvId
               FROM $table
               WHERE id>=$startRec and id<=$stopRec
               ORDER BY 1 DESC";

  my $sth1 = doQuery( \$query );

  while ( $pkt = $sth1->fetchrow_arrayref ) {

      printf("%.6f numcontenders: $pkt->[$Iid] $numContenders\n", $pkt->[$Itime]) 
	  if ($PrintContenders);

    $recnum++;
    print "$pkt->[$Iid]\n" if $pkt->[$Iid]%100000 == 0;

    mirrorPrint( "$recnum: $numContenders $pkt->[$Iid] $pkt->[$Itime] $pkt->[$Idest]<-$pkt->[$Isrc] ",
		 "sz=$pkt->[$Isize] rt=$pkt->[$Irate] rtry=$pkt->[$Iretry] seq=$pkt->[$Iseqno] typ=$pkt->[$Isubtype] ",
 		 "dir=$pkt->[$Idir] rcvd=$pkt->[$Ircvd]\n",
		 "contenders(): $numContenders\n" ) if $verbose;

    if ($numContenders != $#activeClients+1 ) {
      mirrorPrint( "numContenders/activeClients mismatch: $numContenders vs $#activeClients\n" );
      exit 1;
    }

    # convert this packet's time to microseconds
    $pkt->[$Itime] *= $MILLION;

    ######################################################################
    # work on idle period
    ######################################################################

    # compute how much idle air time we just passed
    $idleTime = $lastTime - $pkt->[$Itime];

    # It's possible to see overlapping transmissions in the data, because it's
    # the merged observations of many monitors.  We deal with this in the
    # simplest possible way (almost as simple as just ignoring it).

    if ( $idleTime < 0 ) {
      $maxOverlap = -$idleTime if -$idleTime > $maxOverlap && $recnum>1;
      mirrorPrint( "Record # $recnum (id = $pkt->[$Iid]): Neg idletime -- $idleTime\n" )
	if $recnum>1 and $verbose;
      # $lastTime is sitting at the beginning (in forward time) of the next packet's xmit.
      # This packet's completion time is sometime after that, apparently.
      # Just keep track of how much overlap there is.
      $totalOverlap += -$idleTime;
      $idleTime = 0;
    }

    # okay, on to processing this packet

    ## fix the firstclient_delay for all the first clients that would be expiring in this round
    ## from there on, the original loop below will take care of things
    if ($truthTable) {
	my $modified = 0;
	for (my $i=0; $i<=$#firstClient; $i++) {
	    my $clientId = $firstClient[$i]->[$FIRSTCLIENT_ID];
	    if (defined($kickOutTime[$clientId]) and $pkt->[$Itime] <= $kickOutTime[$clientId]) {
		mirrorPrint(sprintf("will kick deactivate $clientId type=$firstClient[$i]->[$FIRSTCLIENT_CTYPE]; kotime = $kickOutTime[$clientId]. overshoot = %f\n", $lastTime - $kickOutTime[$clientId])) if ($verbose);

		$firstClient[$i]->[$FIRSTCLIENT_DELAY] = $lastTime - $kickOutTime[$clientId];
		undef $kickOutTime[$clientId];
		$modified = 1;
	    }
	}
	@firstClient = sort {$a->[$FIRSTCLIENT_DELAY] <=> $b->[$FIRSTCLIENT_DELAY]} @firstClient 
	    if ($modified);
    }


    # we break the total idle time up into segements, with at least 1 client
    # leaving the idle set each segment, except possibly for the last
    $remainingIdleTime = $idleTime;
    while ( $remainingIdleTime > 0 ) {

	if ( @firstClient && $firstClient[0]->[$FIRSTCLIENT_DELAY] <= $remainingIdleTime ) {
	    $idleTimeCorrection = $firstClient[0]->[$FIRSTCLIENT_DELAY];
	    mirrorPrint( "\tClient ", $firstClient[0]->[$FIRSTCLIENT_ID], ": ",
			 $firstClient[0]->[$FIRSTCLIENT_DELAY], " --> \n" ) if $verbose;
	}
	else {
	    $idleTimeCorrection = $remainingIdleTime;
	}
	$remainingIdleTime -= $idleTimeCorrection;
	$lastTime -= $idleTimeCorrection;

	######################################################################################
	# update statistics for all active clients
	#   Note: we don't keep aggregate statistics -- we derive them from per-client
	#         information once done.
	######################################################################################

	# all we have at this point is idleness.  To make the total times stored
	# add up to elapsed time, we store idleness that occurs when there are no
	# contenders as $UNKNOWN_CLIENT idleness (with 0 contenders)

	if ( $numContenders == 0 ) {
	    $timeStats[$NO_CLIENT]->[0]->[$IDLE_TIME] += $idleTimeCorrection;
	}
	else {
	    for (my $cl=0; $cl<=$#activeClients; $cl++ ){
		$timeStats[$activeClients[$cl]]->[$numContenders]->[$IDLE_TIME] += $idleTimeCorrection;
	    }
	}

      ###########################################
      # have any "first clients" timed out?
      ###########################################

      for (my $i=0; $i<=$#firstClient; $i++){
	$firstClient[$i]->[$FIRSTCLIENT_DELAY] -= $idleTimeCorrection;
      }

      while ( @firstClient && $firstClient[0]->[$FIRSTCLIENT_DELAY] <= 0 ) {
	  ##printf(STDERR "deactivating1 %d\n", $firstClient[0]->[$FIRSTCLIENT_ID]); 
	  deactivate( $lastTime);
      }

      mirrorPrint( "*** $numContenders  $remainingIdleTime $#{@firstClient}\n" ) if $verbose;
    }


    #############################################################
    # is this a new session for either src or dest of this pkt?
    #############################################################

    for ($Iindex=$Idest; $Iindex<=$Isrc; $Iindex++ ) {
      $client = $pkt->[$Iindex];
      next if !defined($client);
      if ( $clientData[$client]->[$CLIENT_SESSION_START] - $pkt->[$Itime] > $SESSION_TIMEOUT ) {
	# first record info on the session we've just detected has ended
	dumpClientStats( $client ) if defined($clientData[$client]->[$CLIENT_SESSION_ID]);

	# now start the new session
	$clientData[$client]->[$CLIENT_SESSION_ID] = $nextSessionID++;
	$clientData[$client]->[$CLIENT_SESSION_STOP] = $pkt->[$Itime];
	# this next assignment shouldn't be required.  However, the data
	# contains records where a client sends a message to itself (client 00:00:00:00:00:00).
	# If we wait for the update to CLIENT_SESSION_START, just after this loop, we'll
	# dump two records for the self-sending client, one correct, and one with the next stop
	# time just assigned (leading to negative session times).
	$clientData[$client]->[$CLIENT_SESSION_START] = $pkt->[$Itime];
	undef $clientData[$client]->[$CLIENT_INTVL_STOP_RECORD];
      }
    }

    #############################################################
    # now work on busy period covered by this packet
    #############################################################

    # 144 bit PLCP preamble + 48 bit PLCP header, sent at 1Mbps, plus
    # 802.11 frame, at xmit rate
    $airTime = 192 +$pkt->[$Isize]*8.0/$pkt->[$Irate];

    $lastTime = $pkt->[$Itime] - $airTime;
    ##$lastTime = $pkt->[$Itime];


    $client = $pkt->[$Isrc]; # if client is undef, it's effectively 0, which is $UNKNOWN_CLIENT
    $isNav = ($pkt->[$Isubtype] == $ACK) || ($pkt->[$Isubtype]==$CTS);
    $isData = int($pkt->[$Isubtype]/16) == 2;

    $clientData[$client]->[$CLIENT_SESSION_START] = $lastTime;
    $clientData[$pkt->[$Idest]]->[$CLIENT_SESSION_START] = $lastTime;

    # Figure out if some client needs to be added to the active list,
    # or perhaps to the firstClients list. The 'client' is usually the
    # sender, but is the receiver for ACKs and CTSs.
    # (Some ACKs/CTSs are "unmatched" -- we (probably) won't see
    # a packet from the client being responded to.  It'll just
    # time out, though (and there must have been one, even if we didn't see it).

      $effectiveClient = $isNav ? $pkt->[$Idest] : $client;

    # keep track of how much time during an interval each contending client spent defering

    foreach (@activeClients) {
      next if $_ == $effectiveClient;
      $stopRec = $clientData[$client]->[$CLIENT_INTVL_STOP_RECORD];
      next unless defined($stopRec);
      $stopRec->[INTVLREC_DEFERTIME] += $airTime;
    }

    # need to activate the client?

      if ( $clientData[$effectiveClient]->[$CLIENT_STATE] == $IDLE_STATE  ) {	  # activate the client
	  $clientData[$effectiveClient]->[$CLIENT_STATE] = $ON_IDLE_LIST_STATE;
	  push @activeClients, $effectiveClient;
	  $numContenders++;
      }
      else {  # remove the client from the list; we'll it back right after this.
	  for (my $cl=0; $cl<=$#firstClient; $cl++ ) {
	      if ( $firstClient[$cl]->[$FIRSTCLIENT_ID] == $effectiveClient ) {
		  splice(@firstClient, $cl, 1);
		  undef $kickOutTime[$effectiveClient];
		  last;
	      }
	  }
      }
      
      $clientData[$effectiveClient]->[$CLIENT_STATE] = $ON_IDLE_LIST_STATE;
      if ($truthTable and $pkt->[$Isubtype] >= 27) {  ##i.e. RTS, CTS, ACK, DATA; we don't see anything else >= 27
	  $remainingIdleTime = $WAY_BIG;
	  mirrorPrint("inserting kick client $effectiveClient typ=$pkt->[$Isubtype] ") if ($verbose);
	  $kickOutTime[$effectiveClient] = getKickoutTime($effectiveClient, $pkt->[$Itime]);
      }
      else {
	  if ($pkt->[$Iretry] > 0 || $pkt->[$Isubtype] == 28 || $pkt->[$Isubtype] == 29) {
	      $remainingIdleTime = $TIMEOUT;
	  }
	  else {
	      # if this is a non-retry packet, make sure the effective client ends up on the first client list.
	      # (If the packet is an ACK/CTS, we should see the packet being ACK'ed before the first
	      # client timeout expires, so there's no harm done.  I hope.)
	      # THIS CODE (ROUGHLY) DUPLICATED IN EPILOGUE... (not subroutined due to paranoia about speed)

	      #$remainingIdleTime = rand(300) + 10;
	      $remainingIdleTime = $RANDOM_WAIT_TIME;
	  }
	  mirrorPrint("inserting regular client $effectiveClient typ=$pkt->[$Isubtype] $remainingIdleTime\n") if ($verbose);
      }
      push @firstClient, [$effectiveClient, $remainingIdleTime, $lastTime, $pkt->[$Isubtype]];
      @firstClient = sort {$a->[$FIRSTCLIENT_DELAY] <=> $b->[$FIRSTCLIENT_DELAY]} @firstClient;
 
      
#     if ( $clientData[$effectiveClient]->[$CLIENT_STATE] == $IDLE_STATE  ) {
#       # activate the client
#       $clientData[$effectiveClient]->[$CLIENT_STATE] = $ACTIVE_STATE;
#       push @activeClients, $effectiveClient;
#       $numContenders++;

#       # THIS CODE (ROUGHLY) DUPLICATED JUST BELOW ---
#       # This is the test -- will setting a timeout on all active clients affect the shape of
#       # the #contenders distribution?
#       $clientData[$effectiveClient]->[$CLIENT_STATE] = $ON_IDLE_LIST_STATE;
#       if ($truthTable and $pkt->[$Isubtype] >= 27) {  ##i.e. RTS, CTS, ACK, DATA; we don't see anything else >= 27
# 	  $remainingIdleTime = $WAY_BIG;
# 	  mirrorPrint("inserting kick-retry client $effectiveClient typ=$pkt->[$Isubtype] ") if ($verbose);
# 	  $kickOutTime[$effectiveClient] = getKickoutTime($effectiveClient, $pkt->[$Itime]);
#       }
#       else {
# 	  $remainingIdleTime = $TIMEOUT;
# 	  mirrorPrint("inserting regular-retry client $effectiveClient typ=$pkt->[$Isubtype] $remainingIdleTime\n") if ($verbose);
#       }
#       push @firstClient, [$effectiveClient, $remainingIdleTime, $lastTime, $pkt->[$Isubtype]];
#       @firstClient = sort {$a->[$FIRSTCLIENT_DELAY] <=> $b->[$FIRSTCLIENT_DELAY]} @firstClient;
#     }

#     # if this is a non-retry packet, make sure the effective client ends up on the first client list.
#     # (If the packet is an ACK/CTS, we should see the packet being ACK'ed before the first
#     # client timeout expires, so there's no harm done.  I hope.)
#     # Note:  we might not have seen the original, non-retry transmission of a packet,
#     #        so make sure not to put the client on the list twice

#     if ( $pkt->[$Iretry] == 0 ) {
#       if ( $clientData[$effectiveClient]->[$CLIENT_STATE] == $ON_IDLE_LIST_STATE ) {
# 	# client is already on the first list.  Oh, well, update it's idleness timeout
# 	# (by removing it from the list and letting it fall into default code)
# 	for (my $cl=0; $cl<=$#firstClient; $cl++ ) {
# 	  if ( $firstClient[$cl]->[$FIRSTCLIENT_ID] == $effectiveClient ) {
# 	      splice(@firstClient, $cl, 1);
# 	      undef $kickOutTime[$effectiveClient];
# 	      last;
# 	  }
# 	}
#       }
#       # THIS CODE (ROUGHLY) DUPLICATED IN EPILOGUE... (not subroutined due to paranoia about speed)
#       $clientData[$effectiveClient]->[$CLIENT_STATE] = $ON_IDLE_LIST_STATE;
#       if ($truthTable and $pkt->[$Isubtype] >= 27) {  ##i.e. RTS, CTS, ACK, DATA; we don't see anything else >= 27
# 	  $remainingIdleTime = $WAY_BIG;
# 	  mirrorPrint("inserting kick client $effectiveClient typ=$pkt->[$Isubtype] ") if ($verbose);
# 	  $kickOutTime[$effectiveClient] = getKickoutTime($effectiveClient, $pkt->[$Itime]);
#       }
#       else {
# 	  #$remainingIdleTime = rand(300) + 10;
# 	  $remainingIdleTime = DIFS + rand(CW_MIN) * SLOT_TIME;
# 	  mirrorPrint("inserting regular client $effectiveClient typ=$pkt->[$Isubtype] $remainingIdleTime\n") if ($verbose);
#       }
#       push @firstClient, [$effectiveClient, $remainingIdleTime, $lastTime, $pkt->[$Isubtype]];
#       @firstClient = sort {$a->[$FIRSTCLIENT_DELAY] <=> $b->[$FIRSTCLIENT_DELAY]} @firstClient;
#     }


    ###################################################
    # update packet statistics.
    #   Have to wait until here because #contenders
    #   might change due to this packet.
    ###################################################

    $packetStats[$client]->[$numContenders]->[$isData]->[$pkt->[$Idir]]->[$isNav]
       ->[$pkt->[$Iretry]]->[$pkt->[$Ircvd]]->{$pkt->[$Irate]}->[$BY_BYTES] += $pkt->[$Isize];
    $packetStats[$client]->[$numContenders]->[$isData]->[$pkt->[$Idir]]->[$isNav]
       ->[$pkt->[$Iretry]]->[$pkt->[$Ircvd]]->{$pkt->[$Irate]}->[$BY_PACKET]++;
    $seenRate{$pkt->[$Irate]} = 1;  # remember that we've seen this rate

    ###################################################
    # update time statistics.  We attribute busy time
    # to the actual sender, even for ACK/CTS.
    ###################################################

    $busyType = $isNav ? $NAV_BUSY_TIME : $BUSY_TIME;
    $timeStats[$client]->[$numContenders]->[$busyType] += $airTime;

    # we have to know if client who got the busy time attributed to it is
    # in the active set or not.  Otherwise, we can't figure out total times
    # from this data;

    $deferType = ($clientData[$effectiveClient]->[$CLIENT_STATE] == $IDLE_STATE) ? $DEFER_TO_INACTIVE_TIME : $DEFER_TIME;
    for ( my $cl=0; $cl<=$#activeClients; $cl++ ) {
      next if $activeClients[$cl] == $client;
      $timeStats[$activeClients[$cl]]->[$numContenders]->[$deferType] += $airTime;
    }

    ######################################################################
    # update sender's interval statistics
    ######################################################################

    if ( $pkt->[$Isrc] ) {
      $client = $pkt->[$Isrc];
      $type = int($pkt->[$Isubtype]/16);
      $stopRec = $clientData[$client]->[$CLIENT_INTVL_STOP_RECORD];
      if ( defined($stopRec) && ($stopRec->[INTVLREC_CONVID]==$pkt->[$Iconvid]) ) {
	my $intvlEntry = $intervalStats[$client]->[$type]->{$pkt->[$Irate]}->{$stopRec->[INTVLREC_XMIT_RATE]}
	       ->[$pkt->[$Iretry]]->[$stopRec->[INTVLREC_RETRY]]->[$pkt->[$Idir]];
	if ( !defined($intvlEntry) ) {
	  $intvlEntry = [];
	  $intervalStats[$client]->[$type]->{$pkt->[$Irate]}->{$stopRec->[INTVLREC_XMIT_RATE]}
	             ->[$pkt->[$Iretry]]->[$stopRec->[INTVLREC_RETRY]]->[$pkt->[$Idir]] = $intvlEntry;
	}
	$intvlEntry->[INTVL_COUNT]++;
	$intvlEntry->[INTVL_SUCCEED]++ if $stopRec->[INTVLREC_SUCCEED];
	$intvlEntry->[INTVL_LENGTH] += $stopRec->[INTVLREC_TIME] - $pkt->[$Itime];
	$intvlEntry->[INTVL_DEFER] = $stopRec->[INTVLREC_DEFERTIME];
	#mirrorPrint("$pkt->[$Iid]: Had type $stopRec->[INTVLREC_TYPE], now have $type\n")
	#  if ($pkt->[$Iretry] == $stopRec->[INTVLREC_RETRY]-1) && ($type != $stopRec->[INTVLREC_TYPE]);
      }
      $clientData[$client]->[$CLIENT_INTVL_STOP_RECORD] =
	    [$lastTime, $pkt->[$Iretry], $pkt->[$Irate], $pkt->[$Iconvid], $type, $pkt->[$Ircvd], 0 ];
    }

  }

  $stopRec = $startRec - 1;
  $startRec = $stopRec - $CHUNKSIZE;
} while ( $stopRec >= $fromRecord );

##############################################################
# EPILOGUE:  force all clients to terminate their sessions
# (to get us reliable active duration histogram info).
# Then dump all as yet undumped data.
##############################################################

for ( my $cl=0; $cl<=$#clientData; $cl++ ) {
  push @firstClient, [$cl, 0] if $clientData[$cl]->[$CLIENT_STATE] == $ACTIVE_STATE;
}

while ( @firstClient ) {
  deactivate( $lastTime );
}

$clientData[$UNKNOWN_CLIENT]->[$CLIENT_SESSION_START ] = $lastTime;
$clientData[$NO_CLIENT]->[$CLIENT_SESSION_START ] = $lastTime;

for ( my $cl=0; $cl<=$#clientData; $cl++ ) {
  dumpClientStats( $cl ) if defined($clientData[$cl]->[$CLIENT_SESSION_ID]);
}

for ( my $t=0; $t<=$NUM_TABLE; $t++ ) {
  if ( $outputTable[$t]->{insertCount} ) {
    chop $outputTable[$t]->{insertQuery};
    doQuery( \$outputTable[$t]->{insertQuery} );
  }
}


##################################
# print run summary
##################################

$query = "SELECT aFrameTime from $table where id=$fromRecord";
$sth = doQuery( \$query );
$lastTime = $sth->fetchrow_array;
$sth->finish();

dbdisconnect();

mirrorPrint( "\nTrace interval: ", $endTime - $lastTime, "\n" );
mirrorPrint( "\t", convertTime($lastTime), " - ", convertTime($endTime), "\n" );

mirrorPrint( "\nMaximum active time: ", $maxActiveTime/$MILLION, " seconds\n" );
mirrorPrint( "\nMaximum xmit overlap: ", $maxOverlap/$MILLION, " usec.\n" );
mirrorPrint( "Xmit overlap: ", $totalOverlap/$MILLION, " seconds\n");

mirrorPrint("\nClient active duration histogram:\n");
my $format = "%13s";
mirrorPrint( sprintf($format, commify(0)), " - ", sprintf($format,commify(10)), ": $activeTimeHist[$0]\n" );
for (my $i=1; $i<=$#activeTimeHist; $i++ ) {
    mirrorPrint( sprintf($format, commify(exp($i*log(10))) ), " - ",
		 sprintf($format, commify(exp(($i+1)*log(10)))), "  ",
		 $activeTimeHist[$i]?commify($activeTimeHist[$i]):"",
		 "\n" );
}
mirrorPrint("\n");

